/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/

#include <cstring>
#include "bitboard.h"

using namespace std;


int firstbit_lookup[0xFFFF+1];
int lastbit_lookup[0xFFFF+1];
bitboard_t bb_square[MAXPOS];
bitboard_t bb_col[MAXCOL];
bitboard_t bb_row[MAXROW];
bitboard_t bb_white_squares;
bitboard_t bb_black_squares;

square_t rot90_lookup[MAXPOS];

square_t rot45_lookup[MAXPOS];
square_t rot45_shift[MAXPOS];
bitboard_t rot45_and[MAXPOS];

square_t rot315_lookup[MAXPOS];
square_t rot315_shift[MAXPOS];
bitboard_t rot315_and[MAXPOS];

bitboard_t bb_ray[MAXPOS][MAXPOS];
unsigned char ray_type[MAXPOS][MAXPOS];

bitboard_t bb_castling[PLAYERS][3];

bitboard_t bb_offensive_pawn[2][MAXPOS];

char square_last[MAXPOS][MAXPOS];
char square_next[MAXPOS][MAXPOS];



static void InitRotatedBB();
static void InitRay();
  

void InitBB()
{
  firstbit_lookup[0] = -1;
  lastbit_lookup[0] = -1;
  memset( bb_col, 0, sizeof( bb_col ) );
  memset( bb_row, 0, sizeof( bb_row ) );
  bb_white_squares = bb_black_squares = NULLBOARD;
  
  for( int i = 1; i <= 0xFFFF; ++i ) {
    for( int j = 0; j < 16; ++j ) {
      if( ( 1 << j ) & i ) {
	lastbit_lookup[i] = j;
      }
      if( ( 0x8000 >> j ) & i ) {
	firstbit_lookup[i] = 15-j;
      }
    }
  }

  for( int i = 0; i < MAXPOS; ++i ) {
    bb_square[i] = static_cast<bitboard_t>( 1 ) << i;
    bb_col[Column(i)] |= bb_square[i];
    bb_row[Row(i)] |= bb_square[i];
    if( (Column(i)%2) == (Row(i)%2) ) {
      bb_black_squares |= bb_square[i];
    } else {
      bb_white_squares |= bb_square[i];
    }
  }

  InitRotatedBB();
  InitRay();

  bb_castling[WHITE][CASTLING_KING_SIDE] = bb_square[F1] | bb_square[G1];
  bb_castling[BLACK][CASTLING_KING_SIDE] = bb_square[F8] | bb_square[G8];
  bb_castling[WHITE][CASTLING_QUEEN_SIDE] = bb_square[D1] | bb_square[C1] | bb_square[B1];
  bb_castling[BLACK][CASTLING_QUEEN_SIDE] = bb_square[D8] | bb_square[C8] | bb_square[B8];
}

void InitRotatedBB()
{
  // 90 degree
  for( square_t i = 0; i < MAXPOS; ++i ) {
    square_t x, y;
    x = Column( i );
    y = Row( i );
    square_t x2, y2;
    x2 = y;
    y2 = MAXCOL-1 - x;

    rot90_lookup[i] = CR2Square( x2, y2 );
  }

  // 45 and 315 degree
  int dimdiag[MAXCOL*2 - 1];
  int posdiag[MAXCOL*2 - 1];

  dimdiag[0] = 1;
  posdiag[0] = 0;
  for( int i = 1; i < MAXCOL*2-1; ++i ) {
    posdiag[i] = posdiag[i-1] + dimdiag[i-1];
    if( i < MAXCOL ) {
      dimdiag[i] = dimdiag[i-1] + 1;
    } else {
      dimdiag[i] = dimdiag[i-1] - 1;
    }
  }

  for( square_t i = 0; i < MAXPOS; ++i ) {
    square_t x, y;
    x = MAXCOL-1 - Column( i );
    y = Row( i );
    int diag = x + y;
    if( diag >= MAXCOL ) {
      y = Column( i );
    }

    rot45_lookup[i] = posdiag[diag] + y;
    rot45_shift[i] = posdiag[diag];
    rot45_and[i] = ( static_cast<bitboard_t>( 1 ) << dimdiag[diag] ) - 1;

    x = Column( i );
    y = Row( i );
    diag = x + y;
    if( diag >= MAXCOL ) {
      x = MAXCOL-1 - y;
    }
    rot315_lookup[i] = posdiag[diag] + x;
    rot315_shift[i] = posdiag[diag];
    rot315_and[i] = ( static_cast<bitboard_t>( 1 ) << dimdiag[diag] ) - 1;
  }
}

void InitRay()
{
  memset( bb_offensive_pawn, 0, sizeof( bb_offensive_pawn ) );
  
  for( square_t i = 0; i < MAXPOS; i++ ) {
    square_t x1 = Column( i );
    square_t y1 = Row( i );

    for( square_t j = 0; j < i; ++j ) {
      bb_ray[i][j] = bb_ray[j][i];
      ray_type[i][j] = ray_type[j][i];
    }

    // offensive pawn
    // for white...
    for( square_t y2 = y1 + 1; y2 < MAXROW - 1; ++y2 ) {
      if( x1 > 0 ) {
	bb_offensive_pawn[WHITE][i] |= bb_square[CR2Square( x1 - 1, y2 )];
      }
      if( x1 < MAXCOL ) {
	bb_offensive_pawn[WHITE][i] |= bb_square[CR2Square( x1 + 1, y2 )];
      }
    }
    // and for black...
    for( square_t y2 = y1 - 1; y2 > 0; --y2 ) {
      if( x1 > 0 ) {
	bb_offensive_pawn[BLACK][i] |= bb_square[CR2Square( x1 - 1, y2 )];
      }
      if( x1 < MAXCOL ) {
	bb_offensive_pawn[BLACK][i] |= bb_square[CR2Square( x1 + 1, y2 )];
      }
    }
    
    for( square_t j = i; j < MAXPOS; ++j ) {
      ray_type[i][j] = RAY_NULL;
      bb_ray[i][j] = NULLBOARD;
      square_t x2 = Column( j );
      square_t y2 = Row( j );
      if( y1 == y2 ) { // horizontal ray
	for( square_t k = i; k <= j; ++k ) {
	  bb_ray[i][j] |= bb_square[k];
	}
	ray_type[i][j] = RAY_ROOK;
      } else if( x1 == x2 ) { // vertical ray
	for( square_t k = i; k <= j; k += MAXCOL ) {
	  bb_ray[i][j] |= bb_square[k];
	}
	ray_type[i][j] = RAY_ROOK;
      } else {
	square_t deltax = x1 - x2;
	square_t deltay = y1 - y2;
	if( deltax == deltay ) {
	  for( square_t k = i; k <= j; k += MAXCOL + 1 ) {
	    bb_ray[i][j] |= bb_square[k];
	  }
	  ray_type[i][j] = RAY_BISHOP;
	} else if( deltax == -deltay ) {
	  for( square_t k = i; k <= j; k += MAXCOL - 1 ) {
	    bb_ray[i][j] |= bb_square[k];
	  }
	  ray_type[i][j] = RAY_BISHOP;
	}
      }
    }
  }

  // build table for next square of a ray
  for( square_t i = 0; i < MAXPOS; ++i ) {
    square_t x1 = Column( i );
    square_t y1 = Row( i );

    for( square_t j = 0; j < MAXPOS; ++j ) {
      square_t x2 = Column( j );
      square_t y2 = Row( j );

      square_next[i][j] = -1;
      if( i == j ) {
	continue;
      }

      if( y1 == y2 ) {
	// horizontal ray
	if( x1 < x2 ) {
	  if( x2 < MAXCOL-1 ) {
	    square_next[i][j] = static_cast<char>( j + 1 );
	  }
	} else {
	  if( x2 > 0 ) {
	    square_next[i][j] = static_cast<char>( j - 1 );
	  }
	}
      } else if( x1 == x2 ) {
	// vertical ray
	if( y1 < y2 ) {
	  if( y2 < MAXROW-1 ) {
	    square_next[i][j] = static_cast<char>( j + MAXCOL );
	  }
	} else {
	  if( y2 > 0 ) {
	    square_next[i][j] = static_cast<char>( j - MAXCOL );
	  }
	}
      } else {
	square_t deltax = x1 - x2;
	square_t deltay = y1 - y2;

	if( deltax == deltay ) {
	  if( i < j ) {
	    if( y2 < MAXROW-1 && x2 < MAXCOL-1 ) {
	      square_next[i][j] = static_cast<char>( j + MAXCOL + 1 );
	    }
	  } else {
	    if( y2 > 0 && x2 > 0 ) {
	      square_next[i][j] = static_cast<char>( j - MAXCOL - 1 );
	    }
	  }
	} else if( deltax == -deltay ) {
	  if( i < j ) {
	    if( y2 < MAXROW-1 && x2 > 0 ) {
	      square_next[i][j] = static_cast<char>( j + MAXCOL - 1 );
	    }
	  } else {
	    if( y2 > 0 && x2 < MAXCOL-1 ) {
	      square_next[i][j] = static_cast<char>( j - MAXCOL + 1 );
	    }
	  }
	}
      }
    }
  }

  // build table for last square of a ray
  for( square_t i = 0; i < MAXPOS; ++i ) {
    for( square_t j = 0; j < MAXPOS; ++j ) {
      if( ray_type[i][j] != RAY_NULL ) {
	square_last[i][j] = j;
      } else {
	square_last[i][j] = -1;
	continue;
      }

      square_t k = j;
      while( square_next[i][k] != -1 ) {
	k = square_next[i][k];
	square_last[i][j] = k;
      }
    }
  }
}
