/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#ifndef _BITBOARD_H_
#define _BITBOARD_H_

/*
  Bitboard unit.
  Note: there are no move generation functions here.

  Created it 28/06/2004
  Maurizio
*/

#include "smash.h"

typedef uint64_t bitboard_t;
const bitboard_t NULLBOARD = bitboard_t(0);

// lookup tables to find first and last bit in a 16 bit word
extern int firstbit_lookup[0xFFFF+1];
extern int lastbit_lookup[0xFFFF+1];

// bitboard for each square in the board
extern bitboard_t bb_square[MAXPOS];

inline bitboard_t BBSquare( square_t sq )
{
  return bitboard_t(1) << sq;
}


// bitboard for each col/row in the board
extern bitboard_t bb_col[MAXCOL];
extern bitboard_t bb_row[MAXROW];

// bitboard for white and black squares
extern bitboard_t bb_white_squares;
extern bitboard_t bb_black_squares;

// rotated bitboard lookup tables
extern square_t rot90_lookup[MAXPOS];
extern square_t rot45_lookup[MAXPOS];
extern square_t rot45_shift[MAXPOS];
extern bitboard_t rot45_and[MAXPOS];
extern square_t rot315_lookup[MAXPOS];
extern square_t rot315_shift[MAXPOS];
extern bitboard_t rot315_and[MAXPOS];

// squares involved by castling
extern bitboard_t bb_castling[2][3];

// squares to check if any pawn can attack a square
extern bitboard_t bb_offensive_pawn[2][MAXPOS];

// lookup tables for rays
extern bitboard_t bb_ray[MAXPOS][MAXPOS];
extern unsigned char ray_type[MAXPOS][MAXPOS];

const unsigned char RAY_NULL = 0;
const unsigned char RAY_ROOK = 1;
const unsigned char RAY_BISHOP = 2;

// lookup tables with next and last square of a ray
extern char square_last[MAXPOS][MAXPOS];
extern char square_next[MAXPOS][MAXPOS];

// initialize bitboard lookup tables
void InitBB();

// return first bit set to 1
inline int FirstBit( bitboard_t bb )
{
  if( bb & 0xFFFF ) {
    return firstbit_lookup[bb&0xFFFF];
  } else if( ( bb >> 16 ) & 0xFFFF ) {
    return firstbit_lookup[(bb>>16)&0xFFFF] + 16;
  } else if( ( bb >> 32 ) & 0xFFFF ) {
    return firstbit_lookup[(bb>>32)&0xFFFF] + 32;
  } else {
    return firstbit_lookup[bb>>48] + 48;
  }
}

// return last bit set to 1
inline int LastBit( bitboard_t bb )
{
  if( bb >> 48 ) {
    return lastbit_lookup[bb>>48] + 48;
  } else if( bb >> 32 ) {
    return lastbit_lookup[bb>>32] + 32;
  } else if( bb >> 16 ) {
    return lastbit_lookup[bb>>16] + 16;
  } else {
    return lastbit_lookup[bb];
  }
}

// return how many bits are set to 1
/*
inline int CountBits( bitboard_t bb )
{
  static const uint64_t m1 = 0x5555555555555555ULL;
  static const uint64_t m2 = 0x3333333333333333ULL;
  
  uint32_t n;
  uint64_t a = (uint64_t)bb - ( ((uint64_t)bb>>1) & m1 );
  uint64_t c = ( a & m2 ) + ( (a>>2) & m2 );
  n = ((uint32_t)c) + (uint32_t)(c>>32);
  n = ( n & 0x0F0F0F0F ) + ( ( n >> 4 ) & 0x0F0F0F0F );
  n = ( n & 0xFFFF ) + ( n >> 16 );
  n = ( n & 0xFF ) + ( n >> 8 );
  return ( int )n;
}
*/

inline int CountBits( bitboard_t bb )
{
  uint32_t w = bb >> 32, v = uint32_t(bb);
  v = v - ((v >> 1) & 0x55555555);
  w = w - ((w >> 1) & 0x55555555);
  v = (v & 0x33333333) + ((v >> 2) & 0x33333333);
  w = (w & 0x33333333) + ((w >> 2) & 0x33333333);
  v = (v + (v >> 4)) & 0x0F0F0F0F;
  w = (w + (w >> 4)) & 0x0F0F0F0F;
  v = ((v+w) * 0x01010101) >> 24;
  return v;
}

inline void ToggleBit( bitboard_t &bb, int sq )
{
  //  bb ^= bb_square[sq];
  bb ^= BBSquare( sq );
}

inline void SetBit( bitboard_t &bb, int sq )
{
  //  bb |= bb_square[sq];
  bb |= BBSquare( sq );
}

inline void ResetBit( bitboard_t &bb, int sq )
{
  //  bb &= ~bb_square[sq];
  bb &= ~BBSquare( sq );
}

inline void MoveBit( bitboard_t &bb, int from, int to )
{
  ResetBit( bb, from );
  SetBit( bb, to );
}

#endif
