/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#include <cstdio>
#include <string>
#include <vector>

#include "smash.h"
#include "board.h"
#include "bitboard.h"
#include "movegen.h"

using namespace std;

const char *fen_start_position = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1";


// C O N S T R U C T O R

Board::Board() : side( m_side ), movecount( m_movecount ), en_passant( m_en_passant ), rule50( m_rule50 ),
		 bb_rot45( m_bb_rot45 ), bb_rot315( m_bb_rot315 ), bb_rot90( m_bb_rot90 ),
		 treeply( m_treeply ), tree( m_tree ), hashkey( m_hashkey ), m_tree( MAXPLY )
{
  reset();
}

// reset board to start position
void Board::reset()
{
  loadFen( string( fen_start_position ) );
}

// fill BBs with right values
void Board::fillBB()
{
  m_bb_rot45 = m_bb_rot315 = m_bb_rot90 = NULLBOARD;
  for( int i = 0; i < PLAYERS; i++ ) {
    for( int j = 0; j < PIECES_NUM; j++ ) {
      m_bb_pieces[i][j] = NULLBOARD;
      m_num_pieces[i][j] = 0;
    }
  }
  m_material_score[WHITE] = m_material_score[BLACK] = 0;

  for( square_t sq = 0; sq < MAXPOS; ++sq ) {
    if( m_board[sq] != EMPTY ) {
      piece_t piece = GetPiece( m_board[sq] );
      int player = GetPieceColor( m_board[sq] );
      ++m_num_pieces[player][piece];
      ++m_num_pieces[player][0];
      m_material_score[player] += PIECE_VALUE[piece];
      SetBit( m_bb_pieces[player][piece], sq );
      if( piece == KING ) {
	m_king_pos[player] = sq;
      }
      SetBit( m_bb_pieces[player][EMPTY], sq );
      SetBit( m_bb_rot45, rot45_lookup[sq] );
      SetBit( m_bb_rot90, rot90_lookup[sq] );
      SetBit( m_bb_rot315, rot315_lookup[sq] );
    }
  }
}

// compute zobrist's hash
hash_t Board::setHashKey()
{
  m_hashkey = side == WHITE ? 0ULL : zobrist_key_side;

  for( int i = 0; i < MAXPOS; ++i ) {
    if( m_board[i] != EMPTY ) {
      m_hashkey ^= zobrist_key[GetPieceColor( m_board[i] )][GetPiece( m_board[i] )][i];
    }
  }

  if( en_passant != -1 ) {
    m_hashkey ^= zobrist_key_ep[Column( en_passant )];
  }

  for( int s = 0; s < PLAYERS; ++s ) {
    if( m_castling[s] & CASTLING_KING_SIDE ) {
      m_hashkey ^= zobrist_key_castling[s][0];
    }
    if( m_castling[s] & CASTLING_QUEEN_SIDE ) {
      m_hashkey ^= zobrist_key_castling[s][1];
    }
  }

  return m_hashkey;
}


// C H E C K   M E T H O D S

// check if player attacks specified square with any pieces
// (also pinned ones)
bool Board::attacks( int player, square_t square ) const
{
  // check pawn attacks
  if( pawn_captures[ChangeSide( player )][square] & m_bb_pieces[player][PAWN] ) {
    return true;
  }

  // check knight attacks
  if( knight_moves[square] & m_bb_pieces[player][KNIGHT] ) {
    return true;
  }

  // check horizontal/vertical attacks
  int row = square / MAXCOL;
  int col = rot90_lookup[square] / MAXCOL;
  const bitboard_t mask = ( bitboard_t )0xFF;
  bitboard_t bbp = m_bb_pieces[WHITE][EMPTY] | m_bb_pieces[BLACK][EMPTY];
  bitboard_t bb_attacks;
  bb_attacks = rook_moves[square][0][( int )( ( bbp >> (row*MAXCOL) ) & mask )]
    | rook_moves[square][1][( int )( ( m_bb_rot90 >> (col*MAXCOL) ) & mask )];
  if( bb_attacks & ( m_bb_pieces[player][ROOK] | m_bb_pieces[player][QUEEN] ) ) {
    return true;
  }

  // check diagonal attacks
  bb_attacks = bishop_moves[square][0][( int )( ( m_bb_rot45 >> rot45_shift[square] ) & rot45_and[square] )]
    | bishop_moves[square][1][( int )( ( m_bb_rot315 >> rot315_shift[square] ) & rot315_and[square] )];
  if( bb_attacks & ( m_bb_pieces[player][BISHOP] | m_bb_pieces[player][QUEEN] ) ) {
    return true;
  }

  // check king attack
  if( king_moves[square] & m_bb_pieces[player][KING] ) {
    return true;
  }

  return false;
}

// return all pieces that attack specified square
bitboard_t Board::piecesAttack( int player, square_t square ) const
{
  bitboard_t bb_pieces = NULLBOARD;
  
  // pawn, knight and king are the easiest to find
  bb_pieces = ( pawn_captures[ChangeSide( player )][square] & m_bb_pieces[player][PAWN] )
    | ( knight_moves[square] & m_bb_pieces[player][KNIGHT] )
    | ( king_moves[square] & m_bb_pieces[player][KING] );

  // find horizontal/vertical attacks
  int row = square / MAXCOL;
  int col = rot90_lookup[square] / MAXCOL;
  const bitboard_t mask = ( bitboard_t )0xFF;
  bitboard_t bbp = m_bb_pieces[WHITE][EMPTY] | m_bb_pieces[BLACK][EMPTY];
  bitboard_t bb_attacks;
  bb_attacks = rook_moves[square][0][( int )( ( bbp >> (row*MAXCOL) ) & mask )]
    | rook_moves[square][1][( int )( ( m_bb_rot90 >> (col*MAXCOL) ) & mask )];
  bb_pieces |= bb_attacks & ( m_bb_pieces[player][ROOK] | m_bb_pieces[player][QUEEN] );

  // find diagonal attacks
  bb_attacks = bishop_moves[square][0][( int )( ( m_bb_rot45 >> rot45_shift[square] ) & rot45_and[square] )]
    | bishop_moves[square][1][( int )( ( m_bb_rot315 >> rot315_shift[square] ) & rot315_and[square] )];
  bb_pieces |= bb_attacks & ( m_bb_pieces[player][BISHOP] | m_bb_pieces[player][QUEEN] );

  return bb_pieces;
}

// return pieces that attack square horizontally/vertically
bitboard_t Board::rooksQueensAttack( int player, square_t square ) const
{
  // find horizontal/vertical attacks
  int row = square / MAXCOL;
  int col = rot90_lookup[square] / MAXCOL;
  const bitboard_t mask = ( bitboard_t )0xFF;
  bitboard_t bbp = m_bb_pieces[WHITE][EMPTY] | m_bb_pieces[BLACK][EMPTY];
  bitboard_t bb_attacks;
  bb_attacks = rook_moves[square][0][( int )( ( bbp >> (row*MAXCOL) ) & mask )]
    | rook_moves[square][1][( int )( ( m_bb_rot90 >> (col*MAXCOL) ) & mask )];
  return bb_attacks & ( m_bb_pieces[player][ROOK] | m_bb_pieces[player][QUEEN] );
}

// return pieces that attack square diagonally
bitboard_t Board::bishopsQueensAttack( int player, square_t square ) const
{
  // find diagonal attacks
  bitboard_t bb_attacks = bishop_moves[square][0][( int )( ( m_bb_rot45 >> rot45_shift[square] ) & rot45_and[square] )]
    | bishop_moves[square][1][( int )( ( m_bb_rot315 >> rot315_shift[square] ) & rot315_and[square] )];
  return bb_attacks & ( m_bb_pieces[player][BISHOP] | m_bb_pieces[player][QUEEN] );
}

// check if specified piece is pinned
bool Board::pinned( int player, square_t dst ) const
{
  int xplayer = ChangeSide( player );
  square_t src = m_king_pos[player];

  // first check if there are any pieces between king and square:
  // in this case we can skip pinned check
  if( ( bb_ray[src][dst] & ( m_bb_pieces[WHITE][EMPTY] | m_bb_pieces[BLACK][EMPTY] ) )
      ^ ( BBSquare( src ) | BBSquare( dst ) ) ) {
    return false;
  }

  // pinned check
  switch( ray_type[src][dst] ) {
  case RAY_ROOK:
    if( rooksQueensAttack( xplayer, dst ) & bb_ray[dst][( int )square_last[src][dst]] ) {
      return true;
    }
    break;

  case RAY_BISHOP:
    if( bishopsQueensAttack( xplayer, dst ) & bb_ray[dst][( int )square_last[src][dst]] ) {
      return true;
    }
    break;
  }

  return false;
}

/*
  Check move legality.
    m -- move
  return true if the move is legal

  Assume move is legal in some board position. (es. castling can't be at rank 4)
*/
bool Board::isLegalMove( move_t m ) const
{
  square_t src, dst;
  src = Src( m );
  dst = Dst( m );

  assert( m != NULLMOVE );
  assert( src >= 0 && src < MAXPOS );
  assert( dst >= 0 && dst < MAXPOS );

  if( src == dst ) {
    return false;
  }
  

  // check for piece to move
  
  if( m_board[src] == EMPTY || GetPieceColor( m_board[src] ) != m_side ) {
    return false;
  }


  // check capture
  
  if( m_board[dst] != EMPTY && GetPieceColor( m_board[dst] ) == m_side ) {
    return false;
  }

  piece_t piece = GetPiece( m_board[src] );
  
  if( IsCapture( m ) ) {
    if( dst == en_passant && Capture( m ) != PAWN && piece != PAWN ) {
      return false;
    }
    if( dst != en_passant && GetPiece( m_board[dst] ) != Capture( m ) ) {
      return false;
    }
  }

  // check piece move
  
  int xside = ChangeSide( m_side );
  if( piece == KING ) {
    // if destination is on an adjacent square, the move is legal iff
    // it's not attacked by enemy
    if( king_moves[src] & BBSquare( dst ) ) {
      if( attacks( xside, dst ) ) {
	return false;
      } else {
	return true;
      }
    }

    // if destination square is not on an adjacent square, move must be
    // castling
    static const int castling_row[] = { 0, 7 };
    if( !IsCastling( m ) || Row( src ) != castling_row[m_side] || Row( dst ) != castling_row[m_side]
	|| tree[treeply].checked ) {
      return false;
    }

    assert( abs( src - dst ) == 2 );

    if( dst > src ) {
      if( !( m_castling[m_side] & CASTLING_KING_SIDE ) ) {
	return false;
      }

      if( m_board[src+1] != EMPTY || m_board[src+2] != EMPTY || GetPiece( m_board[dst+1] ) != ROOK
	  || GetPieceColor( m_board[dst+1] ) != m_side || attacks( xside, src+1 ) || attacks( xside, src+2) ) {
	return false;
      }
    } else {
      if( !( m_castling[m_side] & CASTLING_QUEEN_SIDE ) ) {
	return false;
      }

      if( m_board[src-1] != EMPTY || m_board[src-2] != EMPTY || m_board[src-3] != EMPTY
	  || GetPiece( m_board[src-4] ) != ROOK || GetPieceColor( m_board[src-4] ) != m_side
	  || attacks( xside, src-1 ) || attacks( xside, src-2 ) ) {
	return false;
      }
    }

    return true;
  }

  switch( piece ) {
  case PAWN:
    if( IsCapture( m ) ) {
      if( !( pawn_captures[xside][dst] & BBSquare( src ) ) ) {
	return false;
      }
    } else {
      static const int pawn_shift[2] = { 8, -8 };
      int c = IsDoublePawnPush( m ) ? 2 * pawn_shift[m_side] : pawn_shift[m_side];
      if( ( dst - src ) != c ) {
	return false;
      }
    }
    break;

  case BISHOP:
    if( ray_type[src][dst] != RAY_BISHOP ) {
      return false;
    }
    break;
    
  case KNIGHT:
    if( !( knight_moves[src] & BBSquare( dst ) ) ) {
      return false;
    }
    break;

  case ROOK:
    if( ray_type[src][dst] != RAY_ROOK ) {
      return false;
    }
    break;

  case QUEEN:
    if( ray_type[src][dst] == RAY_NULL ) {
      return false;
    }
    break;

  default:
    assert( 0 );
  }

  // there should'n be any pieces between source and destination
  bitboard_t ray = bb_ray[src][dst] & ~BBSquare( src ) & ~BBSquare( dst );
  if( getBBPieces() & ray ) {
    return false;
  }

  /*
    TODO: detect if leave king in check (by the way at this point it's quite
    hard move is not legal)
  */
  
  return true;
}

/*
  Check if move do check
    m -- move
  return true if the move do check
*/
bool Board::isCheckMove( move_t move ) const
{
  // check special move
  if( IsCastling( move ) ) {
    return isCheckMove_Castling( move );
  } else if( IsEnPassant( move ) ) {
    return isCheckMove_EnPassant( move );
  }
  
  const square_t src = Src(move);
  const square_t dst = Dst(move);
  const piece_t piece = !IsPromote( move ) ? GetPiece(m_board[src]) : Promote( move );
  const int xside = ChangeSide( side );

  assert( GetPieceColor( m_board[src] ) == side );

  const square_t enemy_king_pos = m_king_pos[xside];

  // check if this piece do check
  if( piece == PAWN ) {
    if( pawn_captures[side][dst] & BBSquare( enemy_king_pos ) ) {
      return true;
    }
  } else if( piece == KNIGHT ) {
    if( knight_moves[dst] & BBSquare( enemy_king_pos ) ) {
      return true;
    }
  } else {
    switch( ray_type[dst][enemy_king_pos] ) {
    case RAY_ROOK:
      if( piece == ROOK || piece == QUEEN ) {
	if( !( bb_ray[dst][enemy_king_pos] & getBBPieces() & ~BBSquare( src ) & ~BBSquare( dst ) & ~BBSquare( enemy_king_pos ) ) ) {
	  return true;
	}
      }
      break;

    case RAY_BISHOP:
      if( piece == BISHOP || piece == QUEEN ) {
	if( !( bb_ray[dst][enemy_king_pos] & getBBPieces() & ~BBSquare( src ) & ~BBSquare( dst ) & ~BBSquare( enemy_king_pos ) ) ) {
	  return true;
	}
      }
      break;
    }
  }

  // discover hidden check
  // TODO: FIX here
  if( bb_ray[enemy_king_pos][src] & ( getBBPieces() | BBSquare( dst ) ) & ~BBSquare( enemy_king_pos ) & ~BBSquare( src ) ) {
    return false;
  }

  bitboard_t bb;
  
  switch( ray_type[enemy_king_pos][src] ) {
  case RAY_ROOK:
    bb = rooksQueensAttack( side, src ) & bb_ray[src][(int)square_last[enemy_king_pos][src]];
    if( bb ) {
      assert( CountBits( bb ) == 1 );
      square_t hidden_src = FirstBit( bb );
      if( ( bb_ray[hidden_src][src] & ~BBSquare( hidden_src ) & ~BBSquare( src ) & ( getBBPieces() | BBSquare( dst ) ) ) == NULLBOARD ) {
	return true;
      }
    }
    break;
    
  case RAY_BISHOP:
    bb = bishopsQueensAttack( side, src ) & bb_ray[src][(int)square_last[enemy_king_pos][src]];
    if( bb ) {
      assert( CountBits( bb ) == 1 );
      square_t hidden_src = FirstBit( bb );
      if( ( bb_ray[hidden_src][src] & ~BBSquare( hidden_src ) & ~BBSquare( src ) & ( getBBPieces() | BBSquare( dst ) ) ) == NULLBOARD ) {
	return true;
      }
    }
    break;
  }

  return false;
}

// check if castling can do check
// (we just check if rook can do it)
bool Board::isCheckMove_Castling( move_t move ) const
{
  const square_t src = Src( move );
  const square_t dst = Dst( move );

  const square_t rook_dst = src < dst ? dst - 1 : dst + 1;
  const square_t enemy_king_pos = m_king_pos[ChangeSide( side )];
  if( ray_type[enemy_king_pos][rook_dst] == RAY_ROOK
      && !( bb_ray[enemy_king_pos][rook_dst] & getBBPieces() & ~BBSquare( enemy_king_pos ) & ~BBSquare( rook_dst ) ) ) {
    return true;
  }

  return false;
}

// check if en passant capture do check
// 1. check if the pawn do it
// 2. discover hidden check from moved pawn square
// 3. discover hidden check from captured pawn square
bool Board::isCheckMove_EnPassant( move_t move ) const
{
  const square_t src = Src( move );
  const square_t dst = Dst( move );
  const square_t enemy_king_pos = m_king_pos[ChangeSide( side )];
  const square_t captured_src = side == WHITE ? dst - MAXCOL : dst + MAXCOL;

  assert( GetPiece( m_board[src] ) == PAWN );

  // does the pawn do check?
  if( pawn_captures[side][dst] & BBSquare( enemy_king_pos ) ) {
    return true;
  }

  // hidden check from moved pawn square
  if( !( bb_ray[enemy_king_pos][src] & ( getBBPieces() | BBSquare( dst ) )
	 & ~BBSquare( enemy_king_pos ) & ~BBSquare( src ) & ~BBSquare( captured_src ) ) ) {
    switch( ray_type[enemy_king_pos][src] ) {
    case RAY_ROOK:
      if( ( rooksQueensAttack( side, src )
	    & bb_ray[src][(int)square_last[enemy_king_pos][src]] ) ) {
	return true;
      }
      break;

    case RAY_BISHOP:
      if( ( bishopsQueensAttack( side, src )
	    & bb_ray[src][(int)square_last[enemy_king_pos][src]] ) ) {
	return true;
      }
      break;
    }
  }

  // hidden check from captured square
  if( !( bb_ray[enemy_king_pos][captured_src] & ( getBBPieces() | BBSquare( dst ) )
	 & ~BBSquare( enemy_king_pos ) & ~BBSquare( src ) & ~BBSquare( captured_src ) ) ) {
    switch( ray_type[enemy_king_pos][captured_src] ) {
    case RAY_ROOK:
      if( ( rooksQueensAttack( side, captured_src )
	    & bb_ray[captured_src][(int)square_last[enemy_king_pos][captured_src]] ) ) {
	return true;
      }
      break;

    case RAY_BISHOP:
      if( ( bishopsQueensAttack( side, captured_src )
	    & bb_ray[captured_src][(int)square_last[enemy_king_pos][captured_src]] ) ) {
	return true;
      }
      break;
    }
  }

  return false;
}


/*
  Count position repetitions
    maxrep -- maximum number of repetition that we want to check
              (>1)
    maxply -- maximum number of ply to check from last
  return number of repetition
*/
int Board::repetitions( int maxrep, int maxply ) const
{
  assert( maxrep > 1 );
  
  int count = 1;
  int i = treeply;
  while( count < maxrep && maxply >= 2 && !m_tree[i].irreversible && i >= 2 ) {
    i -= 2;
    if( m_tree[i].hashkey == hashkey ) {
      ++count;
    }
    maxply -= 2;
  }

  return count;
}

// P L A Y   M O V E   M E T H O D S

// make a move
void Board::makeMove( move_t move )
{
  static const int enpassant_shift[2] = { -MAXCOL, MAXCOL };
  static const int king_side_rook[2] = { H1, H8 };
  static const int queen_side_rook[2] = { A1, A8 };
  static const piece_t piece_color_mask[2] = { WHITE_PIECE, BLACK_PIECE };

  bool irreversible = false;
  
  // save current move in the game tree
  m_tree[m_treeply].move = move;

  assert( move == NULLMOVE || m_board[Src( move )] != EMPTY );
  assert( move == NULLMOVE || GetPieceColor( m_board[Src( move )] ) == m_side );
  assert( move == NULLMOVE || m_board[Dst( move )] == EMPTY || GetPieceColor( m_board[Dst( move )] ) == ChangeSide( m_side ) );

  if( move != NULLMOVE ) {
    square_t from, to;
    from = Src( move );
    to = Dst( move );
    piece_t piece = GetPiece( m_board[from] );
  
    if( piece == ROOK ) {
      if( from == king_side_rook[m_side] ) {
	if( m_castling[m_side] & CASTLING_KING_SIDE ) {
	  m_castling[m_side] ^= CASTLING_KING_SIDE;
	  m_hashkey ^= zobrist_key_castling[m_side][0];
	}
      } else if( from == queen_side_rook[m_side] ) {
	if( m_castling[m_side] & CASTLING_QUEEN_SIDE ) {
	  m_castling[m_side] ^= CASTLING_QUEEN_SIDE;
	  m_hashkey ^= zobrist_key_castling[m_side][1];
	}
      }
    } else if( piece == PAWN ) {
      m_rule50 = -1;
      irreversible = true;
    }

    // capture move
    if( IsCapture( move ) ) {
      irreversible = true;
    
      piece_t cpt = Capture( move );
      int xside = ChangeSide( m_side );
      --m_num_pieces[xside][cpt];
      --m_num_pieces[xside][EMPTY];
      m_material_score[xside] -= PIECE_VALUE[cpt];
      square_t dst = to;
      if( IsEnPassant( move ) ) {
	dst = to + enpassant_shift[m_side];
	m_board[dst] = EMPTY;
      }
      resetBB( xside, cpt, dst );
      m_hashkey ^= zobrist_key[xside][cpt][dst];
      if( cpt == ROOK && dst == king_side_rook[xside] ) {
	if( m_castling[xside] & CASTLING_KING_SIDE ) {
	  m_castling[xside] ^= CASTLING_KING_SIDE;
	  m_hashkey ^= zobrist_key_castling[xside][0];
	}
      } else if( cpt == ROOK && dst == queen_side_rook[xside] ) {
	if( m_castling[xside] & CASTLING_QUEEN_SIDE ) {
	  m_castling[xside] ^= CASTLING_QUEEN_SIDE;
	  m_hashkey ^= zobrist_key_castling[xside][1];
	}
      }
      m_rule50 = -1;
    }
    
    movePiece( m_side, piece, from, to );
    m_hashkey ^= zobrist_key[m_side][piece][from];
    
    // promote pawn
    if( IsPromote( move ) ) {
      piece_t pr = Promote( move );
      m_board[to] = pr | piece_color_mask[m_side];
      --m_num_pieces[m_side][PAWN];
      ResetBit( m_bb_pieces[m_side][PAWN], to );
      ++m_num_pieces[m_side][pr];
      m_material_score[m_side] += PIECE_VALUE[pr] - PIECE_VALUE[PAWN];
      SetBit( m_bb_pieces[m_side][pr], to );
      m_hashkey ^= zobrist_key[m_side][pr][to];
    } else {
      m_hashkey ^= zobrist_key[m_side][piece][to];
    }
    
    // castling
    if( IsCastling( move ) ) {
      irreversible = true;
      
      square_t rook_src, rook_dst;
      // queen side castling
      if( to < from ) {
	assert( m_castling[m_side] & CASTLING_QUEEN_SIDE );
	rook_src = to - 2;
	rook_dst = from - 1;
      } else {
	// king side castling
	assert( m_castling[m_side] & CASTLING_KING_SIDE );
	rook_src = to + 1;
	rook_dst = from + 1;
      }

      assert( GetPiece( m_board[rook_src] ) == ROOK && GetPieceColor( m_board[rook_src] ) == m_side );
      assert( m_board[rook_dst] == EMPTY );
      movePiece( m_side, ROOK, rook_src, rook_dst );
      m_hashkey ^= zobrist_key[m_side][ROOK][rook_src];
      m_hashkey ^= zobrist_key[m_side][ROOK][rook_dst];
    }
    if( piece == KING ) {
      m_king_pos[m_side] = to;
      if( m_castling[m_side] & CASTLING_KING_SIDE ) {
	irreversible = true;
	m_hashkey ^= zobrist_key_castling[m_side][0];
      }
      if( m_castling[m_side] & CASTLING_QUEEN_SIDE ) {
	irreversible = true;
	m_hashkey ^= zobrist_key_castling[m_side][1];
      }
      m_castling[m_side] = NO_CASTLING;
    }
    
    
    // en passant...
    
    if( en_passant != -1 ) {
      m_hashkey ^= zobrist_key_ep[Column( en_passant )];
    }
    
    // set en passant square if it's a double pawn push move
    if( IsDoublePawnPush( move ) ) {
      m_en_passant = to + enpassant_shift[m_side];
      m_hashkey ^= zobrist_key_ep[Column( m_en_passant )];
    } else {
      m_en_passant = -1;
    }

    assert( m_en_passant == -1 || ( m_en_passant >= MAXCOL && m_en_passant <= (MAXPOS - MAXCOL) ) );
  } else {
    if( m_en_passant != -1 ) {
      m_hashkey ^= zobrist_key_ep[Column( en_passant )];
      m_en_passant = -1;
    }
  }
  
  // save information for current status
  ++m_rule50;
  ++m_movecount;
  ++m_treeply;
  
  m_tree[m_treeply].rule50 = m_rule50;
  m_tree[m_treeply].ep = m_en_passant;
  m_tree[m_treeply].castling[WHITE] = m_castling[WHITE];
  m_tree[m_treeply].castling[BLACK] = m_castling[BLACK];
  m_tree[m_treeply].irreversible = irreversible;
  
  // and finally... change side
  m_side = ChangeSide( m_side );
  m_hashkey ^= zobrist_key_side;
  m_tree[m_treeply].hashkey = m_hashkey;

  // for debug purpose, if you have any doubt about hashkey :)
  //  assert( setHashKey() == m_tree[m_treeply].hashkey );
}

// undo move and go back to old position
move_t Board::undoMove()
{
  static const piece_t piece_color_mask[2] = { WHITE_PIECE, BLACK_PIECE };
  assert( m_treeply > 0 );

  TransitionRec &tr( m_tree[--m_treeply] );

  int xside = ChangeSide( side );
  if( tr.move != NULLMOVE ) {
    square_t from, to;
    from = Src( tr.move );
    to = Dst( tr.move );
    piece_t piece = GetPiece( m_board[to] );
    m_board[from] = m_board[to];
    movePiece( xside, piece, to, from );
    
    if( piece == KING ) {
      m_king_pos[xside] = from;
      
      // castling move
      if( IsCastling( tr.move ) ) {
	if( from < to ) {
	  // king side castling
	  movePiece( xside, ROOK, from+1, to+1 );
	} else {
	  // queen side castling
	  movePiece( xside, ROOK, from-1, to-2 );
	}
      }
    }
    
    // put back captured piece
    if( IsCapture( tr.move ) ) {
      square_t square = to;
      if( IsEnPassant( tr.move ) ) {
	if( xside == WHITE ) {
	  square -= MAXCOL;
	} else {
	  square += MAXCOL;
	}
      }
      m_board[square] = Capture( tr.move ) | piece_color_mask[m_side];
      setBB( m_side, Capture( tr.move ), square );
      piece_t cpt = Capture( tr.move );
      ++m_num_pieces[m_side][cpt];
      ++m_num_pieces[m_side][EMPTY];
      m_material_score[m_side] += PIECE_VALUE[cpt];
    }
    
    // restore promoted pawn
    if( IsPromote( tr.move ) ) {
      --m_num_pieces[xside][piece];
      ResetBit( m_bb_pieces[xside][piece], from );
      ++m_num_pieces[xside][PAWN];
      m_material_score[xside] += PIECE_VALUE[PAWN] - PIECE_VALUE[piece];
      SetBit( m_bb_pieces[xside][PAWN], from );
      m_board[from] = PAWN | piece_color_mask[xside];
    }
  }
  
  // restore previous status
  m_en_passant = tr.ep;
  m_rule50 = tr.rule50;
  m_castling[0] = tr.castling[0];
  m_castling[1] = tr.castling[1];
  m_hashkey = tr.hashkey;
  --m_movecount;

  m_side = xside;
  return tr.move;
}


// F E N   S T R I N G   M A N A G E M E N T   M E T H O D S

// load board from a FEN string
bool Board::loadFen( const string &fen )
{
  m_tree.clear();
  memset( m_board, 0, sizeof( m_board ) );

  char c;
  int id = 0;

  // board status
  for( int i = 7; i >= 0; --i ) {
    int j = 0;
    do {
      c = fen[id++];
      if( isdigit( c ) ) {
	j += c - '0';
      } else {
	piece_t piece;
	switch( c ) {
	case 'K':
	  piece = WHITE_KING;
	  break;
	case 'Q':
	  piece = WHITE_QUEEN;
	  break;
	case 'R':
	  piece = WHITE_ROOK;
	  break;
	case 'N':
	  piece = WHITE_KNIGHT;
	  break;
	case 'B':
	  piece = WHITE_BISHOP;
	  break;
	case 'P':
	  piece = WHITE_PAWN;
	  break;

	case 'k':
	  piece = BLACK_KING;
	  break;
	case 'q':
	  piece = BLACK_QUEEN;
	  break;
	case 'r':
	  piece = BLACK_ROOK;
	  break;
	case 'n':
	  piece = BLACK_KNIGHT;
	  break;
	case 'b':
	  piece = BLACK_BISHOP;
	  break;
	case 'p':
	  piece = BLACK_PAWN;
	  break;

	default:
	  //throw new FENException( "Unknown symbol in FEN string" );
	  assert( 0 );
	}

	m_board[i*MAXCOL+j] = piece;
	++j;
      }
    } while( j < MAXROW );
    c = fen[id++];
  }

  // side to move
  c = fen[id++];
  if( c == 'w' ) {
    m_side = WHITE;
  } else {
    m_side = BLACK;
  }

  // castling status
  c = fen[id++];
  m_castling[WHITE] = m_castling[BLACK] = NO_CASTLING;

  do {
    c = fen[id++];
    switch( c ) {
    case 'K':
      m_castling[WHITE] |= CASTLING_KING_SIDE;
      break;
    case 'Q':
      m_castling[WHITE] |= CASTLING_QUEEN_SIDE;
      break;
    case 'k':
      m_castling[BLACK] |= CASTLING_KING_SIDE;
      break;
    case 'q':
      m_castling[BLACK] |= CASTLING_QUEEN_SIDE;
      break;
    }
  } while( c != ' ' );

  // en passant
  c = fen[id++];
  if( c >= 'a' && c <= 'h' ) {
    m_en_passant = c - 'a';
    c = fen[id++];
    if( c >= '1' && c <= '8' ) {
      m_en_passant += ( c - '1' ) * MAXCOL;
    } else {
      m_en_passant = -1;
    }
  } else {
    m_en_passant = -1;
  }

  fillBB();
  setHashKey();
  sscanf( &fen[id], "%d %d", &m_rule50, &m_movecount );
  m_treeply = 0;

  m_tree[0].move = 0;
  m_tree[0].ep = m_en_passant;
  m_tree[0].rule50 = m_rule50;
  m_tree[0].checked = isInCheck();
  m_tree[0].castling[WHITE] = m_castling[WHITE];
  m_tree[0].castling[BLACK] = m_castling[BLACK];
  m_tree[0].irreversible = true;
  m_tree[0].hashkey = m_hashkey;

  return true;
}

// return fen string of the board
string Board::fen() const
{
  char str[100] = {0};
  int p = 0;
  int empties = 0;

  // board status
  for( int i = 0; i < MAXPOS; ++i ) {
    if( i > 0 && ( i % MAXCOL ) == 0 ) {
      if( empties ) {
	str[p++] = empties + '0';
	empties = 0;
      }
      str[p++] = '/';
    }

    square_t row = Row( i );
    square_t col = Column( i );
    square_t sq = CR2Square( col, MAXROW - row - 1 );
    assert( sq >= 0 );

    char w = 0;
    switch( m_board[sq] ) {
    case BLACK_BISHOP: w = 'b'; break;
    case BLACK_KNIGHT: w = 'n'; break;
    case BLACK_ROOK: w = 'r'; break;
    case BLACK_QUEEN: w = 'q'; break;
    case BLACK_KING: w = 'k'; break;
    case BLACK_PAWN: w = 'p'; break;

    case WHITE_BISHOP: w = 'B'; break;
    case WHITE_KNIGHT: w = 'N'; break;
    case WHITE_ROOK: w = 'R'; break;
    case WHITE_QUEEN: w = 'Q'; break;
    case WHITE_KING: w = 'K'; break;
    case WHITE_PAWN: w = 'P'; break;
    }

    assert( m_board[sq] == EMPTY || w != 0 );

    if( w ) {
      if( empties ) {
	str[p++] = empties + '0';
	empties = 0;
      }
      str[p++] = w;
    } else {
      empties++;
    }
  }

  if( empties ) {
    str[p++] = empties + '0';
  }
  str[p++] = ' ';

  // side to move
  if( m_side == WHITE ) {
    str[p++] = 'w';
  } else {
    str[p++] = 'b';
  }

  str[p++] = ' ';


  // castling
  bool written = false;
  if( m_castling[WHITE] & CASTLING_KING_SIDE ) {
    str[p++] = 'K';
    written = true;
  }
  if( m_castling[WHITE] & CASTLING_QUEEN_SIDE ) {
    str[p++] = 'Q';
    written = true;
  }
  if( m_castling[BLACK] & CASTLING_KING_SIDE ) {
    str[p++] = 'k';
    written = true;
  }
  if( m_castling[BLACK] & CASTLING_QUEEN_SIDE ) {
    str[p++] = 'q';
    written = true;
  }

  if( !written ) {
    str[p++] = '-';
  }

  str[p++] = ' ';

  // en passant
  if( m_en_passant != -1 ) {
    str[p++] = square2str[m_en_passant][0];
    str[p++] = square2str[m_en_passant][1];
  } else {
    str[p++] = '-';
  }

  snprintf( &str[p], sizeof( str ) - p, " %d %d", rule50, movecount );

  return string( str );
}


// O P E R A T O R S

ostream &operator<<( ostream &out, const Board &board )
{
  const char *sepv = "*****************************************\n";
  const char *seph = "*";

  out << "   ";
  for( int i = 0; i < MAXCOL; i++ ) {
    out << "   " << char( 'A' + i ) << " ";
  }
  out << endl;

  for( int i = MAXROW-1; i >= 0; --i ) {
    out << "   " << sepv;
    out << ' ' << char( '1' + i ) << ' ';
    for( int j = 0; j < MAXCOL; j++ ) {
      out << seph;
      square_t dst = CR2Square( j, i );
      if( board[dst] == EMPTY ) {
	out << "    ";
      } else {
	out << ' ' << piece2str[board[dst]] << ' ';
      }
    }
    out << seph << ' ' << char( '1' + i ) << endl;
  }
  out << "   " << sepv;

  out << "   ";
  for( int i = 0; i < MAXCOL; i++ ) {
    out << "   " << char( 'A' + i ) << " ";
  }
  out << endl;

  // counter
  out << (board.movecount+1)/2 << " ";

  // is king in check?
  if( board.isInCheck() ) {
    out << "<<CHECK>> ";
  }
  // side to move
  if( board.side == WHITE ) {
    out << "White to move";
  } else {
    out << "Black to move";
  }

  
  out << " -- ";

  // castling status
  bool written = false;
  if( board.getCastlingStatus( WHITE ) & CASTLING_KING_SIDE ) {
    out << "K";
    written = true;
  }
  if( board.getCastlingStatus( WHITE ) & CASTLING_QUEEN_SIDE ) {
    out << "Q";
    written = true;
  }
  if( board.getCastlingStatus( BLACK ) & CASTLING_KING_SIDE ) {
    out << "k";
    written = true;
  }
  if( board.getCastlingStatus( BLACK ) & CASTLING_QUEEN_SIDE ) {
    out << "q";
    written = true;
  }
  if( !written ) {
    out << "no castling available";
    written = true;
  }
  out << endl;

  return out;
}

// ------------------ TEST SUITE ----------------
#ifdef TEST
#include "search.h"

static void TestIsMoveCheck();
static bool TestIsMoveCheck_Perft( int depth, int ply );

void TestBoard()
{
  TestIsMoveCheck();
}

static void TestIsMoveCheck()
{
  Board &board( thread_data[0].board );
  
  cout << endl << ">>------------------------------ isMoveCheck Test -----------------------------<<" << endl;
  cout << endl << "NOTE: this test can require some minutes..." << endl;

  // TODO: do like a perft
  const struct {
    char *fen;
    int depth;
  } suite[] = {
    { "k7/8/8/8/r2pP2K/8/8/8 b - e3 0 1", 3 },
    { "rnbqkb1r/ppppp1pp/7n/4Pp2/8/8/PPPP1PPP/RNBQKBNR w KQkq f6 0 3", 5 },
    { "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1", 5 },
    { "r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq -", 5 },
    { "8/3K4/2p5/p2b2r1/5k2/8/8/1q6 b - 1 67", 3 },
    { "8/7p/p5pb/4k3/P1pPn3/8/P5PP/1rB2RK1 b - d3 0 28", 6 },
    { "8/2p5/3p4/KP5r/1R3p1k/8/4P1P1/8 w - -", 6 }
  };

  int good, total;
  good = total = 0;
  
  for( unsigned int i = 0; i < sizeof(suite)/sizeof(suite[0]); ++i ) {
    cout << "Test n." << i << " FEN: " << suite[i].fen << " DEPTH: " << suite[i].depth << "... ";
    board.loadFen( suite[i].fen );
    if( TestIsMoveCheck_Perft( suite[i].depth, 0 ) ) {
      ++good;
      cout << "Good :-) ";
    } else {
      cout << " Bad :-( ";
    }
    ++total;
    cout << " Score: " << good << "/" << total << endl;
  }
}

static bool TestIsMoveCheck_Perft( int depth, int ply )
{
  ThreadData &td( thread_data[0] );
  Board &board( td.board );
  
  bool check = board.setCheckStatus();

  if( !check ) {
    GenCaptures( board, ply, td.moves_set );
    GenNonCaptures( board, ply, td.moves_set );
  } else {
    GenEvasions( board, ply, td.moves_set );
  }

  for( move_t *m = td.moves_set[ply]; m < td.moves_set[ply+1]; ++m ) {
    bool do_check = board.isCheckMove( *m );
    bool error = false;
    bool exit = false;
    board.makeMove( *m );
    if( check || board.isLegal() ) {
      if( do_check != board.setCheckStatus() ) {
	error = true;
      }
      
      if( !error && depth > 1 ) {
	exit = !TestIsMoveCheck_Perft( depth-1, ply+1 );
      }
    }
    board.undoMove();

    if( error ) {
      cout << endl << "ERROR at FEN: " << board.fen() << " MOVE: " << Move2Str( *m ) << " isCheckMove() = " << do_check << " ";
    }
    
    if( error || exit ) {
      return false;
    }
  }

  return true;
}
#endif
