/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#ifndef _BOARD_H_
#define _BOARD_H_

/*
  Board class

  Created in 30/06/2004
  Maurizio
*/

#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <cassert>

#include "smash.h"
#include "bitboard.h"
#include "move.h"
#include "eval.h"
#include "gametree.h"
#include "zobrist.h"

class Board
{
 public:
  Board();
  ~Board() {}

  void reset();
  bool loadFen( const std::string &fen );
  std::string fen() const;

  bool attacks( int player, square_t square ) const;
  bitboard_t piecesAttack( int player, square_t square ) const;
  bitboard_t rooksQueensAttack( int player, square_t square ) const;
  bitboard_t bishopsQueensAttack( int player, square_t square ) const;
  bool pinned( int player, square_t square ) const;
  inline int getCastlingStatus( int player ) const;
  inline const bitboard_t &getBBPieces( int player, piece_t piece ) const;
  inline bitboard_t getBBPieces( piece_t piece = EMPTY ) const;
  inline square_t getKingPos( int player ) const;
  inline int getNumPieces( int player, piece_t piece = EMPTY ) const;
  inline int getNumPieces() const;
  inline score_t getMaterial( int player ) const;
  inline score_t getMaterial() const;

  inline int getPhase() const;

  inline bool isLegal() const;
  bool isLegalMove( move_t m ) const;
  bool isCheckMove( move_t m ) const;
  inline bool isInCheck() const;
  inline bool setCheckStatus();
  int repetitions( int maxrep, int maxply ) const;

  void makeMove( move_t move );
  move_t undoMove();

  const int &side;
  const int &movecount;
  const int &en_passant;
  const int &rule50;

  const bitboard_t &bb_rot45, &bb_rot315, &bb_rot90;

  const int &treeply;
  const std::vector<TransitionRec> &tree;

  const hash_t &hashkey;

  inline piece_t operator[]( square_t sq ) const;

 private:
  void fillBB();
  hash_t setHashKey();
  inline void movePiece( int player, piece_t piece, square_t src, square_t dst );
  inline void resetBB( int player, piece_t piece, square_t square );
  inline void setBB( int player, piece_t piece, square_t square );

  bool isCheckMove_Castling( move_t move ) const;
  bool isCheckMove_EnPassant( move_t move ) const;
  
  piece_t m_board[MAXPOS];               // board's squares
  int m_num_pieces[PLAYERS][PIECES_NUM]; // pieces number for both player
  square_t m_en_passant;                 // en passant square
  int m_rule50;                          // rule50 counter
  int m_movecount;                       // move counter
  int m_treeply;                         // game tree ply counter
  int m_side;                            // side to move
  int m_castling[PLAYERS];               // castling status
  score_t m_material_score[PLAYERS];     // material score
  square_t m_king_pos[PLAYERS];          // king position for both players

  bitboard_t m_bb_rot45, m_bb_rot315, m_bb_rot90; // rotated bitboards
  bitboard_t m_bb_pieces[PLAYERS][PIECES_NUM];    // pieces bitboards

  std::vector<TransitionRec> m_tree;     // game history

  hash_t m_hashkey; // board zobrist's hash key
};


// O P E R A T O R S

// print board to output stream
std::ostream &operator<<( std::ostream &out, const Board &board );


// I N L I N E   B O A R D   M E T H O D S

// check board position legality
inline bool Board::isLegal() const
{
  return !attacks( m_side, m_king_pos[ChangeSide( m_side )] );
}

// true if current player is in check
inline bool Board::isInCheck() const
{
  return attacks( ChangeSide( m_side ), m_king_pos[m_side] );
}

// set check status in game tree
inline bool Board::setCheckStatus()
{
  return m_tree[treeply].checked = attacks( ChangeSide( m_side ), m_king_pos[m_side] );
}

// move a piece in the board updating each bb
inline void Board::movePiece( int player, piece_t piece, int from, int to )
{
  m_board[to] = m_board[from];
  m_board[from] = EMPTY;
  MoveBit( m_bb_pieces[player][EMPTY], from, to );
  MoveBit( m_bb_pieces[player][piece], from, to );
  MoveBit( m_bb_rot45, rot45_lookup[from], rot45_lookup[to] );
  MoveBit( m_bb_rot90, rot90_lookup[from], rot90_lookup[to] );
  MoveBit( m_bb_rot315, rot315_lookup[from], rot315_lookup[to] );
}

// delete piece from board
inline void Board::resetBB( int player, piece_t piece, int square )
{
  ResetBit( m_bb_pieces[player][EMPTY], square );
  ResetBit( m_bb_pieces[player][piece], square );
  ResetBit( m_bb_rot45, rot45_lookup[square] );
  ResetBit( m_bb_rot90, rot90_lookup[square] );
  ResetBit( m_bb_rot315, rot315_lookup[square] );
}

// set piece in the board
inline void Board::setBB( int player, piece_t piece, int square )
{
  SetBit( m_bb_pieces[player][EMPTY], square );
  SetBit( m_bb_pieces[player][piece], square );
  SetBit( m_bb_rot45, rot45_lookup[square] );
  SetBit( m_bb_rot90, rot90_lookup[square] );
  SetBit( m_bb_rot315, rot315_lookup[square] );
}

// get castling status
inline int Board::getCastlingStatus( int player ) const
{
  return m_castling[player];
}

// get desired player/piece bitboard 
inline const bitboard_t &Board::getBBPieces( int player, piece_t piece ) const
{
  return m_bb_pieces[player][piece];
}

// get requested piece bitboard (both players)
inline bitboard_t Board::getBBPieces( piece_t piece ) const
{
  return m_bb_pieces[WHITE][piece] | m_bb_pieces[BLACK][piece];
}

// get specified player's king position
inline square_t Board::getKingPos( int player ) const
{
  return m_king_pos[player];
}

// get number of pieces of a player
inline int Board::getNumPieces( int player, piece_t piece ) const
{
  return m_num_pieces[player][piece];
}

// get number of all pieces
inline int Board::getNumPieces() const
{
  return m_num_pieces[WHITE][EMPTY] + m_num_pieces[BLACK][EMPTY];
}

// get material score
inline score_t Board::getMaterial( int player ) const
{
  return m_material_score[player];
}

// get material difference between current player and opponent
inline score_t Board::getMaterial() const
{
  return m_material_score[m_side] - m_material_score[ChangeSide( m_side )];
}

// get game phase (0..256, 0 opening, 256 ending)
inline int Board::getPhase() const
{
  int phase = 4 * ( m_num_pieces[WHITE][QUEEN] + m_num_pieces[BLACK][QUEEN] )
    + 2 * ( m_num_pieces[WHITE][ROOK] + m_num_pieces[BLACK][ROOK] )
    + 1 * ( m_num_pieces[WHITE][BISHOP] + m_num_pieces[BLACK][BISHOP] )
    + 1 * ( m_num_pieces[WHITE][KNIGHT] + m_num_pieces[BLACK][KNIGHT] );

  phase = std::max( 24 - phase, 0 );
  phase = ( 256 * phase + 12 ) / 24;

  return phase;
}

// I N L I N E   O P E R A T O R S

inline piece_t Board::operator[]( square_t sq ) const
{
  assert( sq >= 0 && sq < MAXPOS );
  return m_board[sq];
}


// Test functions...
#ifdef TEST
void TestBoard();
#endif

#endif
