/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
/*
  Evaluation unit

  Created in 16/03/2005
  Maurizio
*/

#include <iostream>
#include "eval.h"
#include "board.h"
#include "smash.h"
#include "movegen.h"

using namespace std;


// #define EVAL_VERBOSE
#ifdef EVAL_VERBOSE
#define DPRINT( s ) printf s
#else
#define DPRINT( s )
#endif

inline const char *GetSideStr( int side )
{
  return side == WHITE ? "WHITE" : "BLACK";
}


/*
  PawnStructure: describe pawn structure with
  useful information like opened files, blocked and
  passed pawns.
*/
struct PawnStructure {
  uint8_t open_file;
  uint8_t semiopen_file[PLAYERS];
  bitboard_t passed_pawn;
  //  bitboard_t blocked_pawn;
};


// F U N C T I O N   D E C L A R A T I O N S

static void EvaluateQueen( const Board &board, int player, score_t &opening, score_t &ending, square_t pos );
static void EvaluateKing( const Board &board, int player, score_t &opening, score_t &ending, square_t pos );
static void EvaluateRook( const Board &board, int side, score_t &opening, score_t &ending,
			  square_t pos, const PawnStructure &pawn_structure );
static void EvaluateKnight( const Board &board, int side, score_t &opening, score_t &ending,
			    square_t pos, const PawnStructure &pawn_structure );
static void EvaluateBishop( const Board &board, int side, score_t &opening, score_t &ending,
			    square_t pos, const PawnStructure &pawn_structure );

static void EvaluatePassedPawn( const Board &board, square_t pos, int prev_passed, score_t &opening, score_t &ending );
static void EvaluatePawns( const Board &board, score_t opening[], score_t ending[], PawnStructure &pawn_structure );


// E V A L U A T I O N   C O N S T A N T S

const int PIECE_VALUE[] = {
  0, PAWN_VALUE, KNIGHT_VALUE, BISHOP_VALUE, ROOK_VALUE, QUEEN_VALUE, KING_VALUE
};


// Piece/square table
const char positional_score[PIECES_NUM][2][MAXPOS] = {
  // empty
  { {0} },
  
  // pawn
  {
    {
      0,  0,  0,  0,  0,  0,  0,  0,
      1,  1,  1,-30,-30,  1,  1,  1,
      0,  0,  0,  4,  4,  0,  0,  0,
      0,  0,  0, 10, 10,  0,  0,  0,
      1,  1,  2,  7,  7,  2,  1,  1,
      4,  4,  4,  8,  8,  4,  4,  4,
      10, 10, 10, 10, 10, 10, 10, 10,
      0,  0,  0,  0,  0,  0,  0,  0
    },
    {
      0,  0,  0,  0,  0,  0,  0,  0,
      1,  1,  1, -4, -4,  1,  1,  1,
      3,  3,  3,  6,  6,  3,  3,  3,
      6,  6,  6,  8,  8,  6,  6,  6,
      12, 12, 12, 12, 12, 12, 12, 12,
      20, 20, 20, 20, 20, 20, 20, 20,
      30, 30, 30, 30, 30, 30, 30, 30,
      0,  0,  0,  0,  0,  0,  0,  0
    }
  },
      
  
  // knight
  {
    {
      -10,-20,  -6,  -6,  -6,  -6, -20, -10,
      -8,   0,   0,   6,   6,   0,   0,  -8,
      -8,   1,   8,   6,   6,   8,   1,  -8,
      -6,   1,   8,  12,  12,   8,   1,  -6,
      -6,   2,  11,  14,  14,  11,   2,  -6,
      -6,   2,   7,  10,  10,   7,   2,  -6,
      -8,  -3,   0,   0,   0,   0,  -3,  -8,
      -100,-8,  -6,  -6,  -6,  -6,  -8, -100
    },
    {
      -10, -8,  -6,  -6,  -6,  -6,  -8, -10,
      -8,   0,   0,   6,   6,   0,   0,  -8,
      -6,   2,   7,  10,  10,   7,   2,  -6,
      -6,   2,  10,  14,  14,  10,   2,  -6,
      -6,   2,  10,  14,  14,  10,   2,  -6,
      -6,   2,   7,  10,  10,   7,   2,  -6,
      -8,  -3,   0,   0,   0,   0,  -3,  -8,
      -10, -8,  -6,  -6,  -6,  -6,  -8, -10
    }
  },
  
  // bishop
  {
    {
      1,  0,-20,  0,  0,-20,  0,  1,
      1,  6,  3,  1,  1,  3,  6,  1,
      2,  5,  6,  8,  8,  6,  5,  2,
      6,  6, 12, 14, 14, 12,  6,  6,
      6,  6, 12, 14, 14, 12,  6,  6,
      2,  5,  4,  6,  6,  4,  5,  2,
      1,  2,  2,  2,  2,  2,  2,  1,
      -1,-2,  0,  0,  0,  0, -2, -1
    },
    {
      1,  1,  0,  0,  0,  0,  0,  1,
      1,  6,  2,  2,  2,  2,  6,  1,
      2,  5, 12,  8,  8, 12,  5,  2,
      3,  8, 10, 15, 15, 10,  8,  3,
      3,  8, 10, 15, 15, 10,  8,  3,
      2,  5, 12,  8,  8, 12,  5,  2,
      1,  6,  2,  2,  2,  2,  6,  1,
      1 , 1,  0,  0,  0,  0,  1,  1
    }
  },
  
  // rooks
  {
    {
      0, 1, 2, 4, 4, 2, 1, 0,
      0, 1, 2, 4, 4, 2, 1, 0,
      0, 1, 2, 4, 4, 2, 1, 0,
      0, 1, 2, 4, 4, 2, 1, 0,
      0, 1, 2, 4, 4, 2, 1, 0,
      0, 1, 2, 4, 4, 2, 1, 0,
      0, 1, 2, 4, 4, 2, 1, 0,
      0, 1, 2, 4, 4, 2, 1, 0
    },
    {
      0, 1, 4, 6, 6, 4, 1, 0,
      0, 1, 4, 6, 6, 4, 1, 0,
      0, 1, 4, 6, 6, 4, 1, 0,
      0, 1, 4, 6, 6, 4, 1, 0,
      0, 1, 4, 6, 6, 4, 1, 0,
      0, 1, 4, 6, 6, 4, 1, 0,
      0, 1, 4, 6, 6, 4, 1, 0,
      0, 1, 4, 6, 6, 4, 1, 0
    }
  },
  
  // queen
  {
    {
      0,  0,  5,  5,  5,  5,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0
    },
    {
      -2,  0,  1,  3,  3,  1,  0,  -2,
      0,   4,  5,  6,  6,  5,  4,   0,
      1,   5,  7,  8,  8,  7,  5,   1,
      3,   6,  8, 10, 10,  8,  6,   3,
      3,   6,  8, 10, 10,  8,  6,   3,
      1,   5,  7,  8,  8,  7,  5,   1,
      0,   4,  5,  6,  6,  5,  4,   0,
      -2,  0,  1,  3,  3,  1,  0,  -2
    }
  },
  
  // king
  {
    {
      10,  20,  -1,  -5,  -5,  -1,  20,  10,
      -4,  -4,  -6,  -8,  -8,  -6,  -4,  -4,
      -5, -10,  -8,  -8,  -8,  -8, -10,  -5,
      -15, -20, -25, -30, -30, -25, -20, -15,
      -20, -20, -40, -60, -60, -40, -20, -20,
      -25, -30, -50, -55, -55, -50, -30, -25,
      -40, -40, -40, -50, -50, -40, -40, -40,
      -60, -55, -50, -50, -50, -50, -55, -60
    },
    {
      -20, -10, -5,  -5,  -5,  -5,  -10, -20,
      -10,  -4,  2,   6,   6,   2,  -4, -10,
      -5,  -3,  10,  20,  20,  10,   2,  -5,
      -5,  -2,  20,  30,  30,  20,   6,  -5,
      -5,  -2,  20,  30,  30,  20,   6,  -5,
      -5,  -3,  10,  20,  20,  10,   2,  -5,
      -10,  -4,  2,   6,   6,   2,  -4, -10,
      -20, -10, -5,  -5,  -5,  -5, -10, -20
    }    
  }
};

// queen bonus/malus
static const score_t QUEEN_EARLY_MOVED_OPENING = -20;

// rook bonus/malus
static const score_t ROOK_TRAPPED = -75;
static const score_t ROOK_BEHIND_PASSED_PAWN_OPENING = 20;
static const score_t ROOK_BEHIND_PASSED_PAWN_ENDING = 40;
static const score_t ROOK_COVER_PASSED_PAWN_OPENING = 10;
static const score_t ROOK_COVER_PASSED_PAWN_ENDING = 30;

// king bonus/malus
static bitboard_t bb_outpost_knight[PLAYERS];
static const score_t KING_CENTER_FILE_OPENING = -5;
static const score_t KING_CASTLE_PAWN_PROTECTION_OPENING = 10;

// light piece bonus/malus
static const score_t LIGHT_PIECE_COVER_PASSED_PAWN_OPENING = 20;
static const score_t LIGHT_PIECE_COVER_PASSED_PAWN_ENDING = 20;

// knight bonus/malus
static const score_t KNIGHT_OUTPOST_OPENING = 20;
static const score_t KNIGHT_OUTPOST_ENDING = 10;

// bishop bonus/malus
static const score_t BISHOP_PAIR_OPENING = 20;
static const score_t BISHOP_PAIR_ENDING = 20;

// PAWN bonus/malus

// passed pawns bonus/malus
static const score_t PASSED_PAWN_OPENING[MAXROW] = { 0, 5, 10, 20, 50, 80, 100, 0 };
static const score_t PASSED_PAWN_ENDING[MAXROW] = { 0, 20, 30, 50, 100, 120, 150, 0 };
static const score_t DOUBLE_PASSED_PAWN_OPENING[MAXROW] = { 0, 5, 7, 15, 20, 25, 30, 0 };
static const score_t DOUBLE_PASSED_PAWN_ENDING[MAXROW] = { 0, 5, 10, 20, 30, 40, 50, 0 };
static const score_t PASSED_PAWN_NEAR_OWN_KING = 10;
static const score_t PASSED_PAWN_NEAR_ENEMY_KING = 5; // a defended pawn can't be captured,
                                                      // so defend a passed pawn
                                                      // is more important

// doubled pawn bonus/malus
static const score_t DOUBLED_PAWN_OPENING = -20;
static const score_t DOUBLED_PAWN_ENDING = -20;

// isolated pawn bonus/malus
static const score_t ISOLATED_PAWN_OPENING[MAXCOL] = { -10, -11, -13, -15, -15, -13, -11, -10 };
static const score_t ISOLATED_PAWN_ENDING[MAXCOL] = { -10, -11, -13, -15, -15, -13, -11, -10 };

// backward pawn bonus/malus
static const score_t BACKWARD_PAWN_OPENING = -8;
static const score_t BACKWARD_PAWN_ENDING = -10;


void InitEval()
{
  bb_outpost_knight[WHITE] =
    bb_square[E4] | bb_square[E5] | bb_square[E6]
    | bb_square[D4] | bb_square[D5] | bb_square[D6]
    | bb_square[C5] | bb_square[C6]
    | bb_square[F5] | bb_square[F6];

  bb_outpost_knight[BLACK] =
    bb_square[E4] | bb_square[E5] | bb_square[E3]
    | bb_square[D4] | bb_square[D5] | bb_square[D3]
    | bb_square[C4] | bb_square[C3]
    | bb_square[F4] | bb_square[F3];
}

// E V A L U A T I O N   F U N C T I O N S

// evaluate board position
score_t Eval( const Board &b )
{
  // compute positional score
  score_t opening[PLAYERS] = {0,0};
  score_t ending[PLAYERS] = {0,0};
  bitboard_t bb_white = b.getBBPieces( WHITE, EMPTY );
  bitboard_t bb_black = b.getBBPieces( BLACK, EMPTY );

  int phase = b.getPhase();
  DPRINT(( "Phase: %d\n", phase ));
  assert( phase >= 0 && phase <= 256 );

  // evaluate pawns
  PawnStructure pawn_structure;
  EvaluatePawns( b, opening, ending, pawn_structure );


  // evaluate white pieces
  while( bb_white ) {
    square_t pos = FirstBit( bb_white );
    ToggleBit( bb_white, pos );

    // PST score
    piece_t piece = GetPiece( b[pos] );
    opening[WHITE] += positional_score[piece][OPENING][pos];
    ending[WHITE] += positional_score[piece][ENDING][pos];

    // evaluate each piece
    switch( piece ) {
    case KNIGHT: EvaluateKnight( b, WHITE, opening[WHITE], ending[WHITE], pos, pawn_structure ); break;
    case BISHOP: EvaluateBishop( b, WHITE, opening[WHITE], ending[WHITE], pos, pawn_structure ); break;
    case ROOK: EvaluateRook( b, WHITE, opening[WHITE], ending[WHITE], pos, pawn_structure ); break;
    case QUEEN: EvaluateQueen( b, WHITE, opening[WHITE], ending[WHITE], pos ); break;
    case KING: EvaluateKing( b, WHITE, opening[WHITE], ending[WHITE], pos ); break;
    }
  }

  DPRINT(( "Opening WHITE: %d  Ending WHITE: %d\n", opening[WHITE], ending[WHITE] ));

  // evaluate black pieces
  while( bb_black ) {
    square_t pos = FirstBit( bb_black );
    ToggleBit( bb_black, pos );

    // PST score (NOTE: black pieces need to transform coords 'cause PST is white oriented)
    piece_t piece = GetPiece( b[pos] );
    square_t rpos = FlipSquare( pos );
    assert( rpos < MAXPOS && rpos >= 0 );
    opening[BLACK] += positional_score[piece][OPENING][rpos];
    ending[BLACK] += positional_score[piece][ENDING][rpos];

    // evaluate each piece
    switch( piece ) {
    case KNIGHT: EvaluateKnight( b, BLACK, opening[BLACK], ending[BLACK], pos, pawn_structure ); break;
    case BISHOP: EvaluateBishop( b, BLACK, opening[BLACK], ending[BLACK], pos, pawn_structure ); break;
    case ROOK: EvaluateRook( b, BLACK, opening[BLACK], ending[BLACK], pos, pawn_structure ); break;
    case QUEEN: EvaluateQueen( b, BLACK, opening[BLACK], ending[BLACK], pos ); break;
    case KING: EvaluateKing( b, BLACK, opening[BLACK], ending[BLACK], pos ); break;
    }
  }

  DPRINT(( "Opening BLACK: %d  Ending BLACK: %d\n", opening[BLACK], ending[BLACK] ));
  DPRINT(( "Material: %d\n", b.getMaterial() ));

  assert( abs( opening[WHITE] - opening[BLACK] ) < SCORE_MATEHI );
  assert( abs( ending[WHITE] - ending[BLACK] ) < SCORE_MATEHI );

  score_t ret = MergeScore( opening[b.side] - opening[ChangeSide(b.side)],
			    ending[b.side] - ending[ChangeSide(b.side)], phase );

  // add material score
  ret += b.getMaterial();
  assert( abs( ret ) < SCORE_MATEHI );

  return ret;
}


/*
  Evaluate Queen strength.
  1. add opening malus if queen moved early to board center (before light pieces develop)
*/
static void EvaluateQueen( const Board &board, int side, score_t &opening, score_t &ending, square_t pos )
{
  assert( GetPiece( board[pos] ) == QUEEN );
  assert( GetPieceColor( board[pos] ) == side );
  
  static const int original_rank[PLAYERS] = { RANK_1, RANK_8 };
  if( Row( pos ) != original_rank[side] ) {
    bitboard_t bb_develop = NULLBOARD;
    switch( side ) {
    case WHITE: bb_develop = BBSquare( B1 ) | BBSquare( C1 ) | BBSquare( F1 ) | BBSquare( G1 ); break;
    case BLACK: bb_develop = BBSquare( B8 ) | BBSquare( C8 ) | BBSquare( F8 ) | BBSquare( G8 ); break;
    default: assert( 0 );
    }

    // assume that develop squares are occupied just by own light pieces
    // (eg. knight in b1 and g1)
    if( bb_develop & ( board.getBBPieces( side, KNIGHT ) | board.getBBPieces( side, BISHOP ) ) ) {
      DPRINT(( "EvaluateQueen: %s %s QUEEN_EARLY_MOVED_OPENING\n", square2str[pos], GetSideStr( side ) ));
      opening += QUEEN_EARLY_MOVED_OPENING;
    }
  }
}

/*
  Evaluate King safety (not threat)
  1. add opening malus for king in center file (iff other side queen is alive)
  2. add opening bonus for protected castle
*/
static void EvaluateKing( const Board &board, int side, score_t &opening, score_t &ending, square_t pos )
{
  assert( GetPiece( board[pos] ) == KING );
  assert( GetPieceColor( board[pos] ) == side );
  
  int file = Column( pos );
  int rank = Row( pos );
  const int original_rank[] = { RANK_1, RANK_8 };

  // king in center file
  // NOTE: file's interval is [D,F] and not [C,F] 'cause we can't penalize too much long castle
  //       it's [D,F] and not [D,E] because we don't want moves like e1f1
  // TODO: use [D,E] file's interval, F square should be penalized by protected castle
  if( file >= FILE_D && file <= FILE_F  ) {
    if( board.getNumPieces( ChangeSide( side ), QUEEN ) > 0 ) {
      DPRINT(( "EvaluateKing: %s %s KING_CENTER_FILE_OPENING\n", square2str[pos], GetSideStr( side ) ));
      opening += KING_CENTER_FILE_OPENING;
    }
    
  } else if( rank == original_rank[side] ) {
    // check protected castle
    int maxfile = min( file + 1, MAXCOL-1 );
    const int rank_step = side == WHITE ? 1 : -1;
    int pawn_rank = Row( pos ) + rank_step;
    piece_t our_pawn = side == WHITE ? WHITE_PAWN : BLACK_PAWN;
    int king_file = file;

    // check for pawns close to king
    for( file = max( 0, file - 1 ); file <= maxfile; ++file ) {
      square_t sq = CR2Square( file, pawn_rank );
      if( board[sq] == our_pawn ) {
	opening += KING_CASTLE_PAWN_PROTECTION_OPENING;
	DPRINT(( "EvaluateKing: %s %s KING_CASTLE_PAWN_PROTECTION_OPENING %s\n",
		 square2str[pos], GetSideStr( side ), square2str[sq] ));
      } else if( file == king_file ) {
	// fianchetto.. check if we have a pawn in fianchetto and the right bishop
	// TODO: now assume bishop is not trapped and can protect king, we should
	// check if bishop is in the right diagonal or could be.
	sq = CR2Square( file, pawn_rank + rank_step );
	bitboard_t bb_color = bb_white_squares & BBSquare( sq ) ? bb_black_squares : bb_white_squares;
	if( board[sq] == our_pawn
	    && bb_color & board.getBBPieces( side, BISHOP ) ) {	
	  DPRINT(( "EvaluateKing: %s %s KING_CASTLE_PAWN_PROTECTION_OPENING (FIANCHETTO) %s\n",
		   square2str[pos], GetSideStr( side ), square2str[sq] ));
	  opening += KING_CASTLE_PAWN_PROTECTION_OPENING;
	}
      }
    }
  }
}

/*
  Evaluate rook strength.
  1. malus for trapped rook
  2. bonus for rook behind passed pawn
  TODO: avoid malus for trapped rook at the end: we should consider mobility instead
*/   
static void EvaluateRook( const Board &board, int side, score_t &opening, score_t &ending,
			  square_t pos, const PawnStructure &pawn_structure )
{
  assert( GetPiece( board[pos] ) == ROOK );
  assert( GetPieceColor( board[pos] ) == side );
  
  if( side == WHITE ) {
    
    if( ( pos == A1 || pos == B1 )
	&& board.getKingPos( WHITE ) <= D1 && board.getKingPos( WHITE ) > pos
	&& ( board[B2] == WHITE_PAWN || board[B2] == WHITE_PAWN )
	&& ( board[A2] == WHITE_PAWN || board[A3] == WHITE_PAWN ) ) {
      DPRINT(( "EvaluateRook: %s %s ROOK_TRAPPED a1/b1\n", square2str[pos], GetSideStr( side ) ));
      opening += ROOK_TRAPPED;
      ending += ROOK_TRAPPED;
    } else if( ( pos == G1 || pos == H1 )
	       && board.getKingPos( WHITE ) >= F1 && board.getKingPos( WHITE ) < pos
	       && ( board[G2] == WHITE_PAWN || board[G3] == WHITE_PAWN )
	       && ( board[H2] == WHITE_PAWN || board[H3] == WHITE_PAWN ) ) {
      DPRINT(( "EvaluateRook: %s %s ROOK_TRAPPED g1/h1\n", square2str[pos], GetSideStr( side ) ));
      opening += ROOK_TRAPPED;
      ending += ROOK_TRAPPED;
    }

  } else {
    
    if( ( pos == A8 || pos == B8 )
	&& board.getKingPos( BLACK ) <= D8 && board.getKingPos( BLACK ) > pos
	&& ( board[B7] == BLACK_PAWN || board[B6] == BLACK_PAWN )
	&& ( board[A7] == BLACK_PAWN || board[A6] == BLACK_PAWN ) ) {
      DPRINT(( "EvaluateRook: %s %s ROOK_TRAPPED a8/b8\n", square2str[pos], GetSideStr( side ) ));
      opening += ROOK_TRAPPED;
      ending += ROOK_TRAPPED;
    } else if( ( pos == G8 || pos == H8 )
	       && board.getKingPos( BLACK ) >= F8 && board.getKingPos( BLACK ) < pos
	       && ( board[G7] == BLACK_PAWN || board[G6] == BLACK_PAWN )
	       && ( board[H7] == BLACK_PAWN || board[H6] == BLACK_PAWN ) ) {
      DPRINT(( "EvaluateRook: %s %s ROOK_TRAPPED g8/h8\n", square2str[pos], GetSideStr( side ) ));
      opening += ROOK_TRAPPED;
      ending += ROOK_TRAPPED;
    }
    
  } // else (side == BLACK)


  // rook behind passed pawn bonus
  
  int col, row;
  col = Column( pos );
  row = Row( pos );
  bitboard_t bb;
  bb = pawn_structure.passed_pawn & bb_col[col];

  while( bb ) {
    square_t pawn_pos = FirstBit( bb );
    ToggleBit( bb, pawn_pos );

    if( GetPieceColor( board[pawn_pos] ) == side ) {
      if( !( bb_ray[pos][pawn_pos] & board.getBBPieces() & ~BBSquare( pawn_pos ) & ~BBSquare( pos ) ) ) {
	if( side == WHITE ) {
	  if( pawn_pos > pos ) {
	    DPRINT(( "EvaluateRook: %s %s ROOK_BEHIND_PASSED_PAWN %s\n",
		     square2str[pos], GetSideStr( side ), square2str[pawn_pos] ));
	    opening += ROOK_BEHIND_PASSED_PAWN_OPENING;
	    ending += ROOK_BEHIND_PASSED_PAWN_ENDING;
	  }
	} else {
	  if( pawn_pos < pos ) {
	    DPRINT(( "EvaluateRook: %s %s ROOK_BEHIND_PASSED_PAWN %s\n",
		     square2str[pos], GetSideStr( side ), square2str[pawn_pos] ));
	    opening += ROOK_BEHIND_PASSED_PAWN_OPENING;
	    ending += ROOK_BEHIND_PASSED_PAWN_ENDING;
	  }
	}
      }
    } else {
      // check for defensive position against enemy passed pawn

      // cover only 1 pawn per file
      if( !( bb_ray[pos][pawn_pos] & board.getBBPieces( ChangeSide( side ), PAWN ) & ~BBSquare( pawn_pos ) ) ) {
	if( side == WHITE ) {
	  if( pawn_pos > pos ) {
	    DPRINT(( "EvaluateRook: %s %s ROOK_COVER_PASSED_PAWN %s\n",
		     square2str[pos], GetSideStr( side ), square2str[pawn_pos] ));
	    opening += ROOK_COVER_PASSED_PAWN_OPENING;
	    ending += ROOK_COVER_PASSED_PAWN_ENDING;
	  }
	} else {
	  if( pawn_pos < pos ) {
	    DPRINT(( "EvaluateRook: %s %s ROOK_COVER_PASSED_PAWN %s\n",
		     square2str[pos], GetSideStr( side ), square2str[pawn_pos] ));
	    opening += ROOK_COVER_PASSED_PAWN_OPENING;
	    ending += ROOK_COVER_PASSED_PAWN_ENDING;
	  }
	}
      }
    }
  }  
}

/*
  Evaluate knight strenght.
  1. bonus if behind an enemy passed pawn
  2. bonus if in good central position defended and unattackable by enemy pawns
*/
static void EvaluateKnight( const Board &board, int side, score_t &opening, score_t &ending,
			    square_t pos, const PawnStructure &pawn_structure )
{
  assert( GetPiece( board[pos] ) == KNIGHT );
  assert( GetPieceColor( board[pos] ) == side );
  
  // bonus for knight stopping enemy passed pawn
  int rank_step = side == WHITE ? MAXCOL : -MAXCOL;
  piece_t piece_toward = (pos >= MAXCOL && pos < MAXPOS-MAXCOL) ? board[pos + rank_step] : EMPTY;
  if( GetPiece( piece_toward ) == PAWN
      && GetPieceColor( piece_toward ) == ChangeSide( side )
      && ( pawn_structure.passed_pawn & BBSquare( pos + rank_step ) ) ) {
    DPRINT(( "EvaluateKnight: %s %s LIGHT_PIECE_COVER_PASSED_PAWN %s\n",
	     square2str[pos], GetSideStr( side ), square2str[pos+rank_step] ));
    opening += LIGHT_PIECE_COVER_PASSED_PAWN_OPENING;
    ending += LIGHT_PIECE_COVER_PASSED_PAWN_ENDING;
  }

  // outpost knight
  if( ( BBSquare( pos ) & bb_outpost_knight[side] )
      && !( bb_offensive_pawn[side][pos] & board.getBBPieces( ChangeSide( side ), PAWN ) ) ) {
    int defenders = CountBits( pawn_captures[ChangeSide( side )][pos] & board.getBBPieces( side, PAWN ) );
    DPRINT(( "EvaluateKnight: %s %s KNIGHT_OUTPOST with %d defenders\n",
	     square2str[pos], GetSideStr( side ), defenders ));
    opening += defenders * KNIGHT_OUTPOST_OPENING;
    ending += defenders * KNIGHT_OUTPOST_ENDING;
  }
}

/*
  Evaluate bishop strenght.
  1. bonus if behind an enemy passed pawn
  2. bonus for bishop pair
*/
static void EvaluateBishop( const Board &board, int side, score_t &opening, score_t &ending,
			    square_t pos, const PawnStructure &pawn_structure )
{
  assert( GetPiece( board[pos] ) == BISHOP );
  assert( GetPieceColor( board[pos] ) == side );
  
  // bonus for bishop stopping enemy passed pawn
  int rank_step = side == WHITE ? MAXCOL : -MAXCOL;
  piece_t piece_toward = (pos >= MAXCOL && pos < MAXPOS-MAXCOL) ? board[pos + rank_step] : EMPTY;
  if( GetPiece( piece_toward ) == PAWN
      && GetPieceColor( piece_toward ) == ChangeSide( side )
      && ( pawn_structure.passed_pawn & BBSquare( pos + rank_step ) ) ) {
    DPRINT(( "EvaluateBishop: %s %s LIGHT_PIECE_COVER_PASSED_PAWN %s\n",
	     square2str[pos], GetSideStr( side ), square2str[pos+rank_step] ));
    opening += LIGHT_PIECE_COVER_PASSED_PAWN_OPENING;
    ending += LIGHT_PIECE_COVER_PASSED_PAWN_ENDING;
  }

  // bonus for bishop pair
  // HACK: to compute once this bonus we check if this bishop is over a white square
  // NOTE: this is not a good evaluation if we have more than one bishop of the same colour!!
  if( BBSquare( pos ) & bb_white_squares && board.getBBPieces( side, BISHOP ) & bb_black_squares ) {
    DPRINT(( "EvaluateBishop: %s %s BISHOP_PAIR\n", square2str[pos], GetSideStr( side ) ));
    opening += BISHOP_PAIR_OPENING;
    ending += BISHOP_PAIR_ENDING;
  }
}

/*
  Evaluate pawns structure
  1. passed pawns
  2. doubled pawn malus
  3. isolated pawn malus
  4. backward pawn malus

  TODO: optimize!!!
*/
static void EvaluatePawns( const Board &board, score_t opening[], score_t ending[], PawnStructure &pawn_structure )
{
  bitboard_t bb_pawns[PLAYERS] = { board.getBBPieces( WHITE, PAWN ), board.getBBPieces( BLACK, PAWN ) };

  // initialize pawn structure information
  pawn_structure.open_file = pawn_structure.semiopen_file[WHITE] = pawn_structure.semiopen_file[BLACK] = 0;
  pawn_structure.passed_pawn = NULLBOARD;
  
  // pawn structure in file
  // each element of array represent one file with a mailbox system (first and last
  // element don't exist and contain convenience value)
  int file_pawn_count[PLAYERS][MAXCOL+2] = { {0}, {0} };
  square_t pawn_forwards[PLAYERS][MAXCOL+2] = { {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
						{MAXPOS, MAXPOS, MAXPOS, MAXPOS, MAXPOS,
						 MAXPOS, MAXPOS, MAXPOS, MAXPOS, MAXPOS} };
  square_t pawn_backwards[PLAYERS][MAXCOL+2] = { {MAXPOS, MAXPOS, MAXPOS, MAXPOS, MAXPOS,
						  MAXPOS, MAXPOS, MAXPOS, MAXPOS, MAXPOS},
						 {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1} };

  // fill arrays
  for( int side = 0; side < PLAYERS; ++side ) {
    bitboard_t bb = bb_pawns[side];
    while( bb ) {
      square_t pos = FirstBit( bb );
      ToggleBit( bb, pos );
      int file = Column( pos ) + 1;
      assert( file >= 1 && file <= 8 );

      ++file_pawn_count[side][file];
      switch( side ) {
      case WHITE:
	if( pawn_forwards[side][file] < pos ) pawn_forwards[side][file] = pos;
	if( pawn_backwards[side][file] > pos ) pawn_backwards[side][file] = pos;
	break;

      case BLACK:
	if( pawn_forwards[side][file] > pos ) pawn_forwards[side][file] = pos;
	if( pawn_backwards[side][file] < pos ) pawn_backwards[side][file] = pos;
	break;

      default:
	assert( 0 );
      }
    }
  }

  int prev_passed[2] = { -1, -1 }; // keep track for rank of previous pawn if it's passed
  
  // evaluate each pawns
  for( int file = FILE_A+1; file <= FILE_H+1; ++file ) {
    // WHITE

    // passed pawn
    square_t pos = pawn_forwards[WHITE][file];
    if( pos > pawn_backwards[BLACK][file]
	&& Row( pos ) >= Row( pawn_backwards[BLACK][file-1] )
	&& Row( pos ) >= Row( pawn_backwards[BLACK][file+1] ) ) {
      assert( pos >= 0 && pos < MAXPOS );
      pawn_structure.passed_pawn |= BBSquare( pos );
      EvaluatePassedPawn( board, pos, prev_passed[WHITE], opening[WHITE], ending[WHITE] );
      prev_passed[WHITE] = Row( pos );
    } else {
      prev_passed[WHITE] = -1;
    }

    // doubled pawn malus
    if( file_pawn_count[WHITE][file] >= 2 ) {
      assert( file_pawn_count[WHITE][file] <= 6 );
      DPRINT(( "Double Pawn: %s WHITE File: %c Count: %d\n",
	       square2str[pos], square2str[pos][0], file_pawn_count[WHITE][file] ));
      opening[WHITE] += DOUBLED_PAWN_OPENING * ( file_pawn_count[WHITE][file] - 1 );
      ending[WHITE] += DOUBLED_PAWN_ENDING * ( file_pawn_count[WHITE][file] - 1 );
    }

    pos = pawn_backwards[WHITE][file];
    // isolated pawn malus
    if( file_pawn_count[WHITE][file-1] == 0
	&& file_pawn_count[WHITE][file+1] == 0
	&& file_pawn_count[WHITE][file] > 0 ) {
      assert( pos >= 0 && pos < MAXPOS );
      DPRINT(( "Isolated Pawn: %s WHITE File: %c\n",
	       square2str[pos], square2str[pos][0] ));
      opening[WHITE] += ISOLATED_PAWN_OPENING[file-1];
      ending[WHITE] += ISOLATED_PAWN_ENDING[file-1];
      // backward pawn malus
    } else if( Row( min( pawn_backwards[WHITE][file-1], pawn_backwards[WHITE][file+1] ) ) > Row( pos ) ) {
      assert( pos >= 0 && pos < MAXPOS );
      DPRINT(( "Backward Pawn: %s WHITE File: %c\n", square2str[pos], square2str[pos][0] ));
      opening[WHITE] += BACKWARD_PAWN_OPENING;
      ending[WHITE] += BACKWARD_PAWN_ENDING;
    }

    // BLACK

    // passed pawn
    pos = pawn_forwards[BLACK][file];
    if( pos < pawn_backwards[WHITE][file]
	&& Row( pos ) <= Row( pawn_backwards[WHITE][file-1] )
	&& Row( pos ) <= Row( pawn_backwards[WHITE][file+1] ) ) {
      assert( pos >= 0 && pos < MAXPOS );
      pawn_structure.passed_pawn |= BBSquare( pos );
      EvaluatePassedPawn( board, pos, prev_passed[BLACK], opening[BLACK], ending[BLACK] );
      prev_passed[BLACK] = Row( pos );
    } else {
      prev_passed[BLACK] = -1;
    }

    // doubled pawn malus
    if( file_pawn_count[BLACK][file] >= 2 ) {
      assert( file_pawn_count[BLACK][file] <= 6 );
      DPRINT(( "Double Pawn: %s BLACK File: %c Count: %d\n",
	       square2str[pos], square2str[pos][0], file_pawn_count[BLACK][file] ));
      opening[BLACK] += DOUBLED_PAWN_OPENING * ( file_pawn_count[BLACK][file] - 1 );
      ending[BLACK] += DOUBLED_PAWN_ENDING * ( file_pawn_count[BLACK][file] - 1 );
    }

    // isolated pawn malus
    pos = pawn_backwards[BLACK][file];
    if( file_pawn_count[BLACK][file-1] == 0
	&& file_pawn_count[BLACK][file+1] == 0
	&& file_pawn_count[BLACK][file] > 0 ) {
      assert( pos >= 0 && pos < MAXPOS );
      DPRINT(( "Isolated Pawn: %s BLACK File: %c\n",
	       square2str[pos], square2str[pos][0] ));
      opening[BLACK] += ISOLATED_PAWN_OPENING[file-1];
      ending[BLACK] += ISOLATED_PAWN_ENDING[file-1];
    } else if( Row( max( pawn_backwards[BLACK][file-1], pawn_backwards[BLACK][file+1] ) ) < Row( pos ) ) {
      assert( pos >= 0 && pos < MAXPOS );
      DPRINT(( "Backward Pawn: %s BLACK File: %c\n", square2str[pos], square2str[pos][0] ));
      opening[BLACK] += BACKWARD_PAWN_OPENING;
      ending[BLACK] += BACKWARD_PAWN_ENDING;
    }

    // pawn structure information: open and semiopen file
    if( file_pawn_count[WHITE][file] == 0 ) {
      if( file_pawn_count[BLACK][file] == 0 ) {
	pawn_structure.open_file = 1 << (file-1);
      } else {
	pawn_structure.semiopen_file[WHITE] = 1 << (file-1);
      }
    } else if( file_pawn_count[BLACK][file] == 0 ) {
      pawn_structure.semiopen_file[BLACK] = 1 << (file-1);
    }
  }
}

/*
  Evaluate passed pawn
  1. add score for passed pawn advance
  2. if previous pawn is passed, give a bonus
  3. give bonus/malus for distance from the kings for the end
*/
static void EvaluatePassedPawn( const Board &board, square_t pos, int prev_passed, score_t &opening, score_t &ending )
{
  int side = GetPieceColor( board[pos] );
  int rank = Row( pos );
  if( side == BLACK ) {
    rank = FlipRow( rank );
    prev_passed = FlipRow( prev_passed );
  }

  // rank related bonus
  DPRINT(( "EvaluatePassedPawn: %s %s Rank: %d\n",
	   square2str[pos], GetSideStr( side ), rank+1 ));
  assert( rank >= RANK_2 && rank < RANK_8 );
  opening += PASSED_PAWN_OPENING[rank];
  ending += PASSED_PAWN_ENDING[rank];

  // passed pawn pair
  if( prev_passed > 0 && abs( prev_passed - rank ) < 2 ) {
    DPRINT(( " DOUBLE_PASSED_PAWN bonus\n" ));
    opening += DOUBLE_PASSED_PAWN_OPENING[min( prev_passed, rank )];
    ending += DOUBLE_PASSED_PAWN_ENDING[min( prev_passed, rank )];
  }

  // distance from the kings
  const int inc = side == WHITE ? 8 : -8;
  DPRINT(( "Distance from side king: %d\n", Distance( board.getKingPos( side ), pos + inc ) ));
  DPRINT(( "Distance from other side king: %d\n", Distance( board.getKingPos( ChangeSide( side ) ), pos + inc ) ));
  ending += PASSED_PAWN_NEAR_OWN_KING * Distance( board.getKingPos( side ), pos + inc );
  ending -= PASSED_PAWN_NEAR_ENEMY_KING * Distance( board.getKingPos( ChangeSide( side ) ), pos + inc );
  
  // TODO: check for unstoppable passed pawn
}
