/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
/*
  History class
  
  Created in 03/02/06
  Maurizio
*/

#include "history.h"
#include "eval.h"
#include "options.h"
#include "smash.h"

#include <algorithm>

using namespace std;

/*
  UCI Options
*/
bool InitializeHistoryHeuristicValuesFromPST = true;
bool UseDifferentialHistoryHeuristic = true;

enum HHScoreT {
  DepthAndDepth,
  DepthAndPly,
  PlyAndPly
} HistoryHeuristicScore;


// reset history values
// Initialize history table with pst values of current game phase
void History::reset( const Board &board )
{
  /* initialize options */
  
  string hhstartvalue = options.getString( "HistoryHeuristicStartValue" );
  if( hhstartvalue == "FromPST" ) {
    InitializeHistoryHeuristicValuesFromPST = true;
  } else if( hhstartvalue == "Zero" ) {
    InitializeHistoryHeuristicValuesFromPST = false;
  } else {
    assert( 0 );
  }

  string hhscore = options.getString( "HistoryHeuristicScore" );
  if( hhscore == "DepthAndDepth" ) {
    HistoryHeuristicScore = DepthAndDepth;
  } else if( hhscore == "DepthAndPly" ) {
    HistoryHeuristicScore = DepthAndPly;
  } else if( hhscore == "PlyAndPly" ) {
    HistoryHeuristicScore = PlyAndPly;
  } else {
    assert( 0 );
  }

  UseDifferentialHistoryHeuristic = options.getBool( "DifferentialHistoryHeuristic" );

  
  /* reset history table */
  
  if( InitializeHistoryHeuristicValuesFromPST ) {
    int phase = board.getPhase();

    // WHITE PIECES
    // NOTE: Assume WHITE_KING is the last piece and WHITE_PAWN the first
    assert( WHITE_KING - WHITE_PAWN + 1 == 6 );
    for( piece_t piece = WHITE_PAWN; piece <= WHITE_KING; ++piece ) {
      for( square_t square = 0; square < MAXPOS; ++square ) {
	htable[piece][square] = MergeScore( positional_score[piece][OPENING][square], positional_score[piece][ENDING][square], phase ) + 100;
	assert( htable[piece][square] < MAX_HISTORY_VALUE && htable[piece][square] >= 0 );
      }
    }
    
    assert( BLACK_KING - BLACK_PAWN + 1 == 6 );
    for( piece_t piece = BLACK_PAWN; piece <= BLACK_KING; ++piece ) {
      for( square_t square = 0; square < MAXPOS; ++square ) {
	int flipped_square = FlipSquare( square );
	assert( flipped_square >= 0 && flipped_square < MAXPOS );
	htable[piece][flipped_square] = htable[GetPiece( piece ) | WHITE_PIECE][square];
	assert( htable[piece][flipped_square] <= MAX_HISTORY_VALUE && htable[piece][flipped_square] >= 0 );
	assert( htable[piece][flipped_square] == MergeScore( positional_score[GetPiece(piece)][OPENING][square],
							     positional_score[GetPiece(piece)][ENDING][square], phase ) + 100 );
      }
    }
  } else {
    memset( htable, 0, sizeof( htable ) );
  }
}

extern int root_depth; // HACK to use root_depth variable in this module, maybe we can declare
                       // it in search.h

// push a good move into history table
void History::push_good( int ply, int depth, const Board &board, move_t move )
{
  assert( depth >= 0 );
  assert( move != NULLMOVE );
  
  score_t score;

  switch( HistoryHeuristicScore ) {
  case DepthAndDepth:
    score = depth * depth;
    break;
    
  case DepthAndPly:
    score = max( 0, root_depth - ply + 2 ) * depth;
    break;
    
  case PlyAndPly:
    score = max( 0, root_depth - ply + 2 );
    score *= score;
    break;

  default:
    assert( 0 );
  }
  assert( score < SCORE_MAX && score >= 0 );

  piece_t piece = board[Src(move)];
  square_t dst = Dst(move);
  assert( htable[piece][dst] >= 0 );
  if( ( htable[piece][dst] += score ) >= MAX_HISTORY_VALUE ) {
    assert( KING - PAWN + 1 == 6 );
    piece_t mask = piece & PLAYER_PIECE_MASK;
    for( piece_t p = PAWN; p < KING; ++p ) {
      for( square_t sq = 0; sq < MAXPOS; ++sq ) {
	htable[p | mask][sq] = ( htable[p|mask][sq] + 1 ) / 2;
      }
    }
  } else {
    assert( htable[piece][dst] >= score );
  }
}

// return history score of the move
score_t History::get( const Board &board, move_t move )
{
  assert( move != NULLMOVE );
  if( UseDifferentialHistoryHeuristic ) {
    return htable[board[Src(move)]][Dst(move)] - htable[board[Src(move)]][Src(move)];
  } else {
    return htable[board[Src(move)]][Dst(move)];
  }
}
