/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#ifndef _KILLER_H_
#define _KILLER_H_

/*
  KillerMove class
  
  Created in 08/01/06
  Maurizio
*/


#include "smash.h"
#include "move.h"
#include <cassert>


// K I L L E R M O V E S T

/*
  This is a template class to manage Killer Moves.
*/

template<int N>
class KillerMovesT
{
public:
  inline KillerMovesT();
  ~KillerMovesT() {}

  inline void push( int ply, move_t move );
  inline move_t get( int ply, int k );
  inline int isKiller( int ply, move_t move );
  inline void clean();

  enum {
    NKILLERS = N
  };
  
private:
  // lists of killer, each list has NKILLERS moves
  move_t m_killers[MAXDEPTH][NKILLERS];
};

// KillerMoves typedef: use this, not KillerMovesT
typedef KillerMovesT<2> KillerMoves;


// I N L I N E   M E T H O D S

template<int N>
inline KillerMovesT<N>::KillerMovesT()
{
  clean();
}

// reset killers list
template<int N>
inline void KillerMovesT<N>::clean()
{
  memset( m_killers, 0, sizeof( m_killers ) );
}

// insert killer in the list of the ply specified rotating it
// (last move inserted into the list is the first killer)
template<int N>
inline void KillerMovesT<N>::push( int ply, move_t move )
{
  assert( ply >= 0 && ply < MAXDEPTH );
  assert( !IsCapture( move ) && !IsPromote( move ) );

  // check before if killer move is already in the list
  for( int i = 0; i < KillerMovesT<N>::NKILLERS; ++i ) {
    if( m_killers[ply][i] == move ) {
      // if so shift other values and put it on the top
      for( int j = i; j > 0; --j ) {
	m_killers[ply][j] = m_killers[ply][j-1];
      }
      m_killers[ply][0] = move;
      return;
    }
  }

  // rotate list and put new move on the top
  for( int i = KillerMovesT<N>::NKILLERS - 1; i > 0 ; --i ) {
    m_killers[ply][i] = m_killers[ply][i-1];
  }
  m_killers[ply][0] = move;
}

// get killer at k position from the list of the specified ply
template<int N>
inline move_t KillerMovesT<N>::get( int ply, int k )
{
  assert( k >= 0 && k < KillerMovesT<N>::NKILLERS );
  return m_killers[ply][k];
}

// check if move is a killer move for the ply
// return the position of it in the list
// from 1 to KillerMovesT<N>::NKILLERS, if the move
// is not a killer, return 0
template<int N>
inline int KillerMovesT<N>::isKiller( int ply, move_t move )
{
  assert( ply >= 0 && ply < MAXDEPTH );
  assert( move != NULLMOVE );
  for( int i = 0; i < KillerMovesT<N>::NKILLERS; ++i ) {
    if( m_killers[ply][i] == ( move & ( MOVE_SQUARE_MASK | ( MOVE_SQUARE_MASK << MOVE_DEST_SHIFT ) | MOVE_CASTLING |
					MOVE_DOUBLE_PAWN_PUSH | MOVE_EN_PASSANT ) ) ) {
      return i + 1;
    }
  }

  return 0;
}


// K I L L E R M A T E

/*
  This is the class to manage killer mate moves
*/

class KillerMate
{
public:
  inline KillerMate();
  ~KillerMate() {}

  inline void push( int ply, move_t move );
  inline move_t get( int ply );
  inline bool isKillerMate( int ply, move_t move );
  inline void clean();

private:
  move_t m_killers[MAXDEPTH];
};

// I N L I N E   M E T H O D S

inline KillerMate::KillerMate()
{
  clean();
}

// insert killer mate move
inline void KillerMate::push( int ply, move_t move )
{
  m_killers[ply] = move;
}

// get killer mate for specified ply
inline move_t KillerMate::get( int ply )
{
  return m_killers[ply];
}

// check if the move is a killer mate move for that ply
inline bool KillerMate::isKillerMate( int ply, move_t move )
{
  assert( move != NULLMOVE );
  return m_killers[ply] == move;
}

// reset killer mate moves for each ply
inline void KillerMate::clean()
{
  memset( m_killers, 0, sizeof( m_killers ) );
}


#endif
