/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
/*
  MoveHandler class

  Created in 28/12/2005
  Maurizio
*/

#include "movehandler.h"
#include "move.h"
#include "movegen.h"
#include "search.h"
#include "killer.h"
#include "history.h"
#include "see.h"

#include <algorithm>

using namespace std;

typedef score_t *score_ptr;


// I N I T I A L I Z A T I O N


// initialize move handler
void InitMoveHandler( ThreadData &td )
{
  td.scores_set[0] = &td.sort_scores[0];
  td.scores_set[1] = &td.sort_scores[0];
}

// if we skip generation in a ply (es. nullmove) we must
// call this function to fix move generator's data structures
// with right values
void FixMoveHandler( ThreadData &td, int ply )
{
  td.moves_set[ply+1] = td.moves_set[ply];
  td.scores_set[ply+1] = td.scores_set[ply]; 
}


// M O V E H A N D L E R   C L A S S

/*
  MoveHandler generate and sort moves by steps:
  1. return hashmove without generate any move
  2. generate and return capture moves
  3. generate and return other moves

  NOTE: if king is in check, steps 2 and 3 are
  combined to: generate and return evasion moves

  With MoveHandler sort is incremental and we can reduce
  the number of moves generated dropping sometime calls
  to GenNonCaptures()
*/

// Contructor
MoveHandler::MoveHandler( ThreadData &td, int ply, move_t hash_move, int filter ) : singlemove( m_singlemove ),
										    m_td( td ),					    
										    m_ply( ply ),
										    m_filter( filter ),
										    m_hash_move( hash_move ),
										    m_singlemove( false )
{
  m_phase = ( m_hash_move != NULLMOVE ) ? MH_PHASE_HASH : MH_PHASE_CAPTURES_GEN;
  // initialize piece hanged table
  memset( m_hanged, 0, sizeof( m_hanged ) );
}


// return next move or NULLMOVE if there are no more
// the move is a pseudo-legal move, so you should check
// legality trying board.makeMove() & board.isLegal().
// If we are in check, nextMove() returns only LEGAL moves
// and set singlemove flag.
move_t MoveHandler::nextMove()
{
  do {
    switch( m_phase ) {
    case MH_PHASE_HASH:
      {
	assert( m_hash_move != NULLMOVE );
      
	// if we are in check we skip generation phase using only
	// GenEvasions(): in this way we can know how many moves
	// we can make (GenEvasions() generate all LEGAL moves)
	if( m_td.board.tree[m_td.board.treeply].checked ) {
	  m_phase = MH_PHASE_OTHERS_GEN;
	  continue;
	}
      
	m_phase = MH_PHASE_CAPTURES_GEN;

	// TODO: now we return hash move ONLY if checkFilter()
	// returns true, should we return bestmoves in Quiesce() too? (non-tactical moves included)
	if( checkFilter( m_hash_move ) ) {
	  // fix: if we skip generation phases in the next ply we haven't
	  // a right value value for scores_set[m_ply+1], so we must
	  // set this in hash phase
	  // fix 2: required also for moves_set[m_ply+1]!!
	  m_td.moves_set[m_ply+1] = m_td.moves_set[m_ply];
	  m_td.scores_set[m_ply+1] = m_td.scores_set[m_ply]; 
	  return m_hash_move;
	}
      }

    case MH_PHASE_CAPTURES_GEN:
      {
	if( !m_td.board.tree[m_td.board.treeply].checked ) {
	  GenCaptures( m_td.board, m_ply, m_td.moves_set );
	  sortCaptures();
	  m_phase = MH_PHASE_CAPTURES;
	} else {
	  m_phase = MH_PHASE_OTHERS_GEN;
	  continue;
	}
      }

    case MH_PHASE_CAPTURES:
      {
	move_t r = findNextMove();
	if( r != NULLMOVE ) {
	  return r;
	}
	m_phase = MH_PHASE_OTHERS_GEN;
      }

    case MH_PHASE_OTHERS_GEN:
      {
	if( !m_td.board.tree[m_td.board.treeply].checked ) {
	  if( m_filter == MH_ALL ) {
	    GenNonCaptures( m_td.board, m_ply, m_td.moves_set );
	    sortNonCaptures();
	  }
	} else {
	  GenEvasions( m_td.board, m_ply, m_td.moves_set );
	  sortEvasions();
	  assert( !m_singlemove );
	  if( m_td.moves_set[m_ply+1] - m_td.moves_set[m_ply] == 1 ) {
	    m_singlemove = true;
	  }
	}
	m_phase = MH_PHASE_OTHERS;
      }

    case MH_PHASE_OTHERS:
      {
	move_t r = findNextMove();
	if( r != NULLMOVE ) {
	  return r;
	}
	m_phase = MH_PHASE_END;
      }

    case MH_PHASE_END:
      {
	return NULLMOVE;
      }

    case MH_PHASE_RESET_HASH:
      {
	assert( 0 );
	assert( m_hash_move != NULLMOVE );
      
	m_phase = MH_PHASE_RESET_MOVES;

	// TODO: now we return hash move ONLY if checkFilter()
	// returns true, should we return bestmoves in Quiesce() too? (non-tactical moves included)
	if( checkFilter( m_hash_move ) ) {
	  return m_hash_move;
	}
      }

    case MH_PHASE_RESET_MOVES:
      {
	assert( 0 );
	if( m == m_td.moves_set[m_ply+1] ) {
	  m_phase = MH_PHASE_END;
	  return NULLMOVE;
	} else {
	  return *m++;
	}
      }
      

    default:
      assert( 0 );
    }

    assert( 0 );
  } while( true );
}

/*
  Reset moves from the first one
  NOTE: Use only at the END, you shouldn't break the regular
  nextMove() loop with this method.
*/
void MoveHandler::reset()
{
  m = m_td.moves_set[m_ply];
  if( m_hash_move != NULLMOVE ) {
    m_phase = MH_PHASE_RESET_HASH;
  } else {
    m_phase = MH_PHASE_RESET_MOVES;
  }
}

const char HANGED = 1;
const char NOHANGED = 2;


// S O R T   M E T H O D S

/*
  order moves using heuristics
  1. HashMove
  2. Captures in MVV/LVA order (equal are sorted by PST value) and promotion
  3. Castling (king side castling is preferred than queen side one)
  4. Passive moves in history heuristic order

  NOTE: hanged piece is moved preferibly
*/

// sort capture and promote moves only
void MoveHandler::sortCaptures()
{
  // aliasing
  Board &board( m_td.board );
  score_ptr &score( m_td.scores_set[m_ply+1] );
  
  score = m_td.scores_set[m_ply];
  assert( m_td.scores_set[m_ply+1] == m_td.scores_set[m_ply] );
  
  int xside = ChangeSide( board.side );

  move_t *mt = m = m_td.moves_set[m_ply];
  int k;

  while( mt != m_td.moves_set[m_ply+1] ) {
    assert( IsCapture( *mt ) || IsPromote( *mt ) );
    
    // skip hash move generated
    if( *mt == m_hash_move ) {
      *mt = *( m_td.moves_set[m_ply+1] - 1 );
      --m_td.moves_set[m_ply+1];
    } else {
      score_t see = SEE( board, *mt );
      // skip no good captures if in tactical/capture mode
      // No Good Captures:
      //   - under queen promote
      //   - SEE negative score
      if( m_filter != MH_ALL ) {
	if( IsPromote( *mt ) && Promote( *mt ) != QUEEN ) {
	  if( !( Promote( *mt ) == KNIGHT && board.isCheckMove( *mt ) ) ) {
	    *mt = *( m_td.moves_set[m_ply+1] - 1 );
	    --m_td.moves_set[m_ply+1];
	    continue;
	  }
	}
	if( !board.isCheckMove( *mt ) && see < 0 ) {
	  *mt = *( m_td.moves_set[m_ply+1] - 1 );
	  --m_td.moves_set[m_ply+1];
	  continue;
	}
      }
      
      square_t src = Src( *mt );
      piece_t piece = GetPiece( board[src] );
      square_t pst_src = board.side == WHITE ? src : FlipSquare( src );
      square_t pst_dst = board.side == WHITE ? Dst( *mt ) : FlipSquare( Dst( *mt ) );
      int phase = board.getPhase() >= 128 ? 1 : 0;

      // killer mate score
      if( m_filter == MH_ALL && m_td.kmate.isKillerMate( m_ply, *mt ) ) {
	*score = SCORE_MAX;
      } else {
	*score = !IsPromote( *mt )
	  ? positional_score[piece][phase][pst_dst] - positional_score[piece][phase][pst_src]
	  : 0;

	/*
	// TODO: we should have a proportional value from history,
	// we don't want too high values!
	*score = !IsPromote( *mt )
	  ? history.get( board, *mt )
	  : 0;
	*/
	
	if( IsCapture( *mt ) ) {
	  *score += see;
	  //	  *score += PIECE_VALUE[Capture( *mt )] - PIECE_VALUE[piece];
	  
	  // EXP: add pst value of captured piece 'cause it won't influence the game
	  //      after the move
	  *score += positional_score[Capture( *mt )][phase][FlipSquare( pst_dst )];
	  if( IsPromote( *mt ) ) {
	    *score += PIECE_VALUE[Promote( *mt )];
	  }
	} else if( IsPromote( *mt ) ) {
	  *score += PIECE_VALUE[Promote( *mt )] - PIECE_VALUE[piece];
	} else {
	  assert( 0 ); // only captures/promotes here!!
	}
	
	// killer move score
	if( m_filter == MH_ALL && ( k = m_td.killers.isKiller( m_ply, *mt ) ) != 0 ) {
	  *score += (KillerMoves::NKILLERS-k-1) * 1024;
	}

	// hanged piece bonus
	if( !m_hanged[src] && piece != KING ) {
	  m_hanged[src] = board.attacks( xside, src ) ? HANGED : NOHANGED;
	}
	*score += m_hanged[src] == HANGED ? 30 : 0;
      }

      ++score;
      ++mt;
    }
  }

  assert( ( m_td.scores_set[m_ply+1] - m_td.scores_set[m_ply] ) == ( m_td.moves_set[m_ply+1] - m_td.moves_set[m_ply] ) );
}

// sort non capture moves
void MoveHandler::sortNonCaptures()
{
  assert( m_td.scores_set[m_ply+1] >= m_td.scores_set[m_ply] );

  // aliasing
  Board &board( m_td.board );
  score_ptr &score( m_td.scores_set[m_ply+1] );
  
  int xside = ChangeSide( board.side );

  move_t *mt = m;
  int k;
  while( mt != m_td.moves_set[m_ply+1] ) {
    assert( !IsCapture( *mt ) && !IsPromote( *mt ) );

    // skip hash move generated
    if( *mt == m_hash_move ) {
      *mt = *( m_td.moves_set[m_ply+1] - 1 );
      --m_td.moves_set[m_ply+1];
    } else if( m_filter == MH_ALL && m_td.kmate.isKillerMate( m_ply, *mt ) ) {
      // killer mate score
      *score++ = SCORE_MAX;
      ++mt;
    } else if( m_filter == MH_ALL && ( k = m_td.killers.isKiller( m_ply, *mt ) ) != 0 ) {
      *score++ = (2*KillerMoves::NKILLERS-k-1) * MAX_HISTORY_VALUE;
      ++mt;
    } else {
      square_t src = Src( *mt );
      piece_t piece = GetPiece( board[src] );
      int phase = board.getPhase() >= 128 ? 1 : 0;

      if( !UseHistoryHeuristicInMoveOrder ) {
	*score = positional_score[piece][phase][Dst( *mt )] - positional_score[piece][phase][Src( *mt )];
      } else {
	*score = m_td.history.get( board, *mt );
      }
      if( IsCastling( *mt ) ) {
	*score = MAX_HISTORY_VALUE;
	// king side castling bonus
	if( src < Dst( *mt ) ) {
	  *score += 50;
	}
      }

      // EXP: killers of two ply before...
      if( m_filter == MH_ALL && m_ply > 2 && ( k = m_td.killers.isKiller( m_ply - 2, *mt ) ) != 0 ) {
	*score += (KillerMoves::NKILLERS-k-1) * MAX_HISTORY_VALUE;
      }
      
      if( !m_hanged[src] && piece != KING ) {
	m_hanged[src] = board.attacks( xside, src ) ? HANGED : NOHANGED;
      }
      *score += m_hanged[src] == HANGED ? 30 : 0;

      ++score;
      ++mt;
    }
  }

  assert( ( m_td.scores_set[m_ply+1] - m_td.scores_set[m_ply] ) == ( m_td.moves_set[m_ply+1] - m_td.moves_set[m_ply] ) );
}

// sort evasion moves
void MoveHandler::sortEvasions()
{
  // aliasing
  Board &board( m_td.board );
  score_ptr &score( m_td.scores_set[m_ply+1] );
  
  score = m_td.scores_set[m_ply];
  int xside = ChangeSide( board.side );

  move_t *mt = m = m_td.moves_set[m_ply];
  while( mt != m_td.moves_set[m_ply+1] ) {
    if( *mt == m_hash_move ) {
      *score++ = SCORE_MAX;
    } else {
      square_t src = Src( *mt );
      piece_t piece = GetPiece( board[src] );
      int phase = board.getPhase() >= 128 ? 1 : 0;

      if( !UseHistoryHeuristicInMoveOrder ) {
	square_t pst_src = board.side == WHITE ? src : FlipSquare( src );
	square_t pst_dst = board.side == WHITE ? Dst( *mt ) : FlipSquare( Dst( *mt ) );
	*score = !IsPromote( *mt )
	  ? positional_score[piece][phase][pst_dst] - positional_score[piece][phase][pst_src]
	  : 0;
      } else {
	*score = m_td.history.get( board, *mt );
      }

      if( IsCapture( *mt ) ) {
	//	*score += PIECE_VALUE[Capture( *mt )] - PIECE_VALUE[piece] + MAX_HISTORY_VALUE + 1000;
	*score += SEE( board, *mt ) + MAX_HISTORY_VALUE + 1000;
	// EXP: add the pst value of the captured piece 'cause it will not influence the game
	//      after the capture
	square_t pst_flipped_dst = board.side == BLACK ? Dst( *mt ) : FlipSquare( Dst( *mt ) );
	*score += positional_score[Capture( *mt )][phase][pst_flipped_dst];
	if( IsPromote( *mt ) ) {
	  *score += PIECE_VALUE[Promote( *mt )];
	}
      } else if( IsPromote( *mt ) ) {
	*score += PIECE_VALUE[Promote( *mt )] - PIECE_VALUE[piece] + MAX_HISTORY_VALUE + 1000;
      }

      if( !m_hanged[src] && piece != KING ) {
	m_hanged[src] = board.attacks( xside, src ) ? HANGED : NOHANGED;
      }
      *score += m_hanged[src] == HANGED ? 30 : 0;
      ++score;
    }
    ++mt;
  }

  assert( ( m_td.scores_set[m_ply+1] - m_td.scores_set[m_ply] ) == ( m_td.moves_set[m_ply+1] - m_td.moves_set[m_ply] ) );
}


// return highest score move or NULLMOVE if there are no more
// moves left
// NOTE: this method should be called internally by nextMove() and only
// after any sort*() methods
move_t MoveHandler::findNextMove()
{
  if( m == m_td.moves_set[m_ply+1] ) {
    return NULLMOVE;
  }

  move_t *mt = m;
  score_t *s = m_td.scores_set[m_ply] + ( m - m_td.moves_set[m_ply] );
  score_t *st = s;
  score_t *bestscore = st;
  move_t *bestmove = mt;

  assert( m < m_td.moves_set[m_ply+1] && m >= m_td.moves_set[m_ply] );
  assert( s < m_td.scores_set[m_ply+1] && s >= m_td.scores_set[m_ply] );
  
  ++st, ++mt;
  while( mt != m_td.moves_set[m_ply+1] ) {
    if( *st > *bestscore ) {
      bestscore = st;
      bestmove = mt;
    }
    ++st, ++mt;
  }

  assert( st == m_td.scores_set[m_ply+1] && mt == m_td.moves_set[m_ply+1] );
  assert( bestscore < m_td.scores_set[m_ply+1] && bestscore >= m_td.scores_set[m_ply] );
  assert( bestmove < m_td.moves_set[m_ply+1] && bestmove >= m );

  // replace first move with bestmove
  swap( *s, *bestscore );
  swap( *m, *bestmove );
  return *m++;
}

// check if the move is accepted by current filter
bool MoveHandler::checkFilter( move_t move )
{
  if( m_filter == MH_ALL ) {
    return true;
  } else if( m_filter == MH_CAPTURES ) {
    if( !m_td.board.tree[m_td.board.treeply].checked ) {
      if( IsPromote( move ) && Promote( move ) != QUEEN ) {
	return false;
      }
      return IsCapture( move ) || IsPromote( move );
    } else {
      return true;
    }
  } else {
    // TODO: tactical moves are also moves that do check
    if( !m_td.board.tree[m_td.board.treeply].checked ) {
      return IsCapture( move ) || IsPromote( move ); // || IsCheck( m_td.board, move )
    } else {
      return true;
    }
  }
}
