/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#include <string>
#include <sstream>
#include <cctype>
#include <map>

#include "protocol.h"
#include "search.h"
#include "movegen.h"
#include "perft.h"
#include "options.h"

using namespace std;


// Command response structures

typedef int ( *command_func )( istringstream &istr, int state );
static map<string, command_func> commands;

// S T A T I C   F U N C T I O N   D E C L A R A T I O N S

// Command response

static void InitCommands();
static int ParseCommand( const string &line, int state );
static int CmdUci( istringstream &istr, int state );
static int CmdUciNewGame( istringstream &istr, int state );
static int CmdDebug( istringstream &istr, int state );
static int CmdIsReady( istringstream &istr, int state );
static int CmdSetOption( istringstream &istr, int state );
static int CmdPosition( istringstream &istr, int state );
static int CmdGo( istringstream &istr, int state );
static int CmdStop( istringstream &istr, int state );
static int CmdPonderHit( istringstream &istr, int state );
static int CmdQuit( istringstream &istr, int state );

static int CmdPerft( istringstream &istr, int state );
static int CmdGenMoves( istringstream &istr, int state );
static int CmdRep( istringstream &istr, int state );
static int CmdEval( istringstream &istr, int state );


// proxy to call methods for each board in every threads
static class BoardProxy
{
public:
  void reset();
  bool loadFen( const std::string &fen );
  bool setCheckStatus();
  int repetitions( int maxrep, int maxply ) const;
  void makeMove( move_t move );
} board;


// F U N C T I O N S

/*
  Main Loop:
    - get input and parse
    - execute command
*/
void MainLoop()
{
  int state = STATE_INPUT_HANDLING;
  bool done = false;

  flags.post = true;
  flags.debug = false;

  InitCommands();

  while( !done ) {
    switch( state ) {
    case STATE_INPUT_HANDLING:
      state = HandleInput( state );
      break;

    case STATE_SEARCH:
    case STATE_PONDER:
      {
	NumThreads = options.getInt( "NumThreads" );
	assert( NumThreads > 0 && NumThreads <= MAX_THREADS );
	if( !board.setCheckStatus() ) {
	  // TODO: maybe it could be better to generate in parallel
	  // due to cache usage
	  for( int i = 0; i < NumThreads; ++i ) {
	    GenCaptures( thread_data[i].board, 0, thread_data[i].moves_set );
	    GenNonCaptures( thread_data[i].board, 0, thread_data[i].moves_set );
	    EraseInvalidMoves( thread_data[i].board, 0, thread_data[i].moves_set );
	  }
	} else {
	  // TODO: same as before
	  for( int i = 0; i < NumThreads; ++i ) {
	    GenEvasions( thread_data[i].board, 0, thread_data[i].moves_set );
	  }
	}
	move_t move = SearchIterate();
	cout << "bestmove " << Move2Str( move ) << endl;
	state = STATE_INPUT_HANDLING;
	break;
      }

    case STATE_SEARCH_STOP:
      state = STATE_INPUT_HANDLING;
      break;

    case STATE_QUIT:
      done = true;
      break;
    }
  }
}


/*
   handle input
     state --  current engine state
   
   return the next engine state
*/
int HandleInput( int state )
{
  if( flags.quit ) {
    return STATE_QUIT;
  }
  
  string cmd;
  getline( cin, cmd );
  if( !cin ) {
    return STATE_QUIT;
  }
  
  int ret = ParseCommand( cmd, state );
  return ret;
}


static int ToLower( int c )
{
  return ::tolower( c );
}

/*
  parse and execute command
    line -- the input command line
    state -- the current engine state
    
  return the next engine state
*/
static int ParseCommand( const string &line, int state )
{
  istringstream istr;
  istr.str( line );

  string cmd;
  istr >> cmd;

  do {
    transform( cmd.begin(), cmd.end(), cmd.begin(), ToLower );
    if( commands.count( cmd ) > 0 ) {
      return (*commands[cmd])( istr, state );
    } else {
      //      log->writeWarning( "Command '%s' unknown.", cmd.c_str() );
      istr >> cmd;
    }
  } while( !cmd.empty() && !!istr );
  
  return state;
}

/*
  initialize command functions lookup table
*/
static void InitCommands()
{
  commands["uci"] = CmdUci;
  commands["ucinewgame"] = CmdUciNewGame;
  commands["debug"] = CmdDebug;
  commands["isready"] = CmdIsReady;
  commands["setoption"] = CmdSetOption;
  commands["position"] = CmdPosition;
  commands["go"] = CmdGo;
  commands["stop"] = CmdStop;
  commands["ponderhit"] = CmdPonderHit;
  commands["quit"] = CmdQuit;

  // extensions
  commands["perft"] = CmdPerft;
  commands["genmoves"] = CmdGenMoves;
  commands["rep"] = CmdRep;
  commands["eval"] = CmdEval;
}

// C O M M A N D   R E S P O N S E   F U N C T I O N

static int CmdUci( istringstream &istr, int state )
{
#if !defined( USESMP )
  cout << "id name smash" << endl;
#else
  cout << "id name Deep Smash" << endl;
#endif
  cout << "id author Maurizio Sambati" << endl;

  options.printDesc( cout );
  
  cout << "uciok" << endl;
  return state;
}

static int CmdUciNewGame( istringstream &istr, int state )
{
  for( int i = 0; i < MAX_THREADS; ++i ) {
    ResetSearch( thread_data[i] );
  }
  ttable.clean();
  // TODO: anything else?
  return state;
}

static int CmdDebug( istringstream &istr, int state )
{
  string on;
  istr >> on;
  if( on == "on" ) {
    flags.debug = true;
  } else {
    flags.debug = false;
  }
  
  return state;
}

static int CmdIsReady( istringstream &istr, int state )
{
  cout << "readyok" << endl;
  
  return state;
}

static int CmdSetOption( istringstream &istr, int state )
{
  string line, name, value;
  getline( istr, line );

  int startname = line.find( "name" ) + 5;
  int endname = line.find( "value", startname );

  name = line.substr( startname, endname - startname - 1 );

  int startvalue = endname + 6;
  value = line.substr( startvalue );

  options.setOption( name, value );
  return state;
}

static int CmdPosition( istringstream &istr, int state )
{
  string tmp, fenstring;
  istr >> tmp;

  // set the start position
  
  if( tmp == "fen" ) {
    istr >> tmp;
    while( tmp != "moves" && !!istr ) {
      fenstring += tmp + " ";
      istr >> tmp;
    }
    board.loadFen( fenstring );
    
  } else if( tmp == "startpos" ) {
    board.reset();
    istr >> tmp;
  } else {
    // TODO: error
    assert( false );
  }

  
  // ... and go to the current position
  
  istr >> tmp;
  while( !tmp.empty() && !!istr ) {
    move_t m = Str2Move( thread_data[0].board, tmp );
    board.makeMove( m );
    istr >> tmp;
  }

  return state;
}


static int CmdGo( istringstream &istr, int state )
{
  string tmp;
  istr >> tmp;

  int wtime, btime, winc, binc, movestogo, fixedtime;
  wtime = btime = winc = binc = movestogo = fixedtime = -1;
  int depth = MAXDEPTH;

  bool ponder, infinite;
  ponder = infinite = false;

  // firstly we parse the line
  
  while( !tmp.empty() && !!istr ) {
    if( tmp == "searchmoves" ) {
      // TODO
    } else if( tmp == "ponder" ) {
      ponder = true;
    } else if( tmp == "infinite" ) {
      infinite = true;
    } else if( tmp == "wtime" ) {
      istr >> wtime;
    } else if( tmp == "btime" ) {
      istr >> btime;
    } else if( tmp == "winc" ) {
      istr >> winc;
    } else if( tmp == "binc" ) {
      istr >> binc;
    } else if( tmp == "movestogo" ) {
      istr >> movestogo;
    } else if( tmp == "depth" ) {
      istr >> depth;
    } else if( tmp == "nodes" ) {
      // TODO
      int dummy;
      istr >> dummy;
    } else if( tmp == "mate" ) {
      // TODO
      int dummy;
      istr >> dummy;
    } else if( tmp == "movetime" ) {
      istr >> fixedtime;
    } else {
      // do nothing...
    }

    istr >> tmp;
  }


  // now we set the time options
  
  timecontrol.wtime = wtime;
  timecontrol.btime = btime;
  timecontrol.winc = winc;
  timecontrol.binc = binc;
  timecontrol.movestogo = movestogo;
  timecontrol.infinite = infinite || ponder;
  timecontrol.fixedtime = fixedtime;
  timecontrol.start();

  flags.max_depth = depth;
  flags.player = thread_data[0].board.side;

  return STATE_SEARCH;
}
  
static int CmdStop( istringstream &istr, int state )
{
  timecontrol.stop();
  if( state == STATE_SEARCH ) {
    return STATE_SEARCH_STOP;
  }
  return state;
}

static int CmdPonderHit( istringstream &istr, int state )
{
  if( state == STATE_PONDER ) {
    return STATE_SEARCH;
  }
  return state;
}

static int CmdQuit( istringstream &istr, int state )
{
  timecontrol.stop();
  flags.quit = true;
  return STATE_QUIT;
}

static int CmdPerft( istringstream &istr, int state )
{
  int depth;
  istr >> depth;
  if( depth > 6 ) {
    cerr << "Warning: this operation could take many time." << endl;
    cerr << "Press CTRL+C to stop." << endl;
  }
  unsigned int start = GetTime();
  uint64_t nodes = Perft( thread_data[0].board, depth );
  unsigned int elapsed = ( GetTime() - start );
  
  cout << "OK Result: " << nodes
       << " Time: " << elapsed/100 << '.' << elapsed%100 << "s." << endl;
  
  return state;
}

static int CmdGenMoves( istringstream &istr, int state )
{
  ThreadData &td( thread_data[0] );
  
  if( td.board.isInCheck() ) {
    GenEvasions( td.board, 0, td.moves_set );
  } else {
    GenCaptures( td.board, 0, td.moves_set );
    GenNonCaptures( td.board, 0, td.moves_set );
  }
  
  cout << "Moves:" << endl;
  int i = 1;
  for( move_t *m = td.moves_set[0]; m != td.moves_set[1]; m++ ) {
    if( i % 4 == 0 ) {
      cout << endl;
    }
    cout << Move2Str( *m ) << "   ";
    i++;
  }
  cout << endl;
  return state;
}

static int CmdRep( istringstream &istr, int state )
{
  cout << "Rep: " << board.repetitions( 3, MAXPLY ) << endl;
  return state;
}

static int CmdEval( istringstream &istr, int state )
{
  score_t score = Eval( thread_data[0].board );
  cout << "Eval: " << score << endl;
  return state;
}


// Board Proxy

void BoardProxy::reset()
{
  for( int i = 0; i < MAX_THREADS; ++i ) {
    thread_data[i].board.reset();
  }
}

bool BoardProxy::loadFen( const std::string &fen )
{
  bool done = true;
  for( int i = 0; i < MAX_THREADS; ++i ) {
    done = thread_data[i].board.loadFen( fen ) && done;
  }
  return done;
}

bool BoardProxy::setCheckStatus()
{
  int num_check = 0;
  for( int i = 0; i < MAX_THREADS; ++i ) {
    num_check += thread_data[i].board.setCheckStatus() ? 1 : 0;
  }
  assert( num_check == 0 || num_check == MAX_THREADS );
  return num_check > 0;
}

int BoardProxy::repetitions( int maxrep, int maxply ) const
{
  return thread_data[0].board.repetitions( maxrep, maxply );
}

void BoardProxy::makeMove( move_t move )
{
  for( int i = 0; i < MAX_THREADS; ++i ) {
    thread_data[i].board.makeMove( move );
  }
}
