/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
/*
  Search functions

  Created in 16/03/2005
  Maurizio
*/

#include <algorithm>

#include "bitboard.h"
#include "search.h"
#include "movegen.h"
#include "movehandler.h"
#include "timecontrol.h"
#include "input.h"
#include "protocol.h"
#include "ttable.h"
#include "options.h"
#include "killer.h"
#include "history.h"
#include "recognizer.h"
#include "smp.h"

using namespace std;


// V A R I A B L E S


// O P T I O N S

static bool UseDeltaPruning = true;
static score_t DeltaPruningMargin = 150;
static bool UseTTInQuiesce = true;
static bool UseTTStoreInQuiesce = true;
bool UseHistoryHeuristicInMoveOrder = true;

static bool UseNullMove = true;
static int NullMoveMinDepth = 2;
static enum {
  NMR_FORMULA, NMR_R2, NMR_R3, NMR_ADAPTIVE
} NullMoveReduction = NMR_FORMULA;

int NumThreads = MAX_THREADS;


// G L O B A L S

// shared data
Status flags;
TimeControl timecontrol;
TTable ttable;
VOLATILE TTable &ttable_search( const_cast<VOLATILE TTable &>( ttable ) );

int root_depth;           // depth at root node
score_t root_score;       // score at root node
const int &max_depth( flags.max_depth );   // max depth for iterative deepening

// thread data
ThreadData thread_data[MAX_THREADS];
static lock_t lock_io;

// useful variables to compute statistics
static int searchtime_start;
static uint64_t searchnodescnt_start;


// F U N C T I O N   D E C L A R A T I O N S

inline score_t EvalMate( int ply );
inline score_t AdjustMate2TT( score_t score, int ply );
inline score_t AdjustMateFromTT( score_t score, int ply );
inline bool IsTTMate( const TTEntry &entry );

inline void CheckUp();

inline bool DoNull( const Board &board );

inline bool NearDraw( const Board &board, int ply );


// ABDADA communication functions
inline static void MakeMove( ThreadData &td, move_t m, int depth, bool pvnode, int ply );
inline static void UndoMove( ThreadData &td, int ply );
inline static void Propagate( ThreadData &current_thread, score_t alpha, int ply );


// F U N C T I O N   D E F I N I T I O N S

void InitSearch()
{
  LockInit( lock_io );
}

void ResetSearch( ThreadData &td )
{
  // reset statistics...
  td.mystat.reset();
}


/*
  return bestmove for root node.
  NOTE: you *must* call GenCaptures/NonCaptures( b, 0 ) with EraseInvalidMoves( b, 0 )
        before this function and check game isn't over (by stalemate or
        mate)
*/
move_t SearchIterate()
{
  // initialize options
  UseDeltaPruning = options.getBool( "DeltaPruning" );
  DeltaPruningMargin = score_t( options.getInt( "DeltaPruningMargin" ) );
  UseHistoryHeuristicInMoveOrder = options.getBool( "HistoryHeuristicInMoveOrder" );

  string TTInQuiesceOpt = options.getString( "TTInQuiesce" );
  if( TTInQuiesceOpt == "Yes" ) {
    UseTTInQuiesce = UseTTStoreInQuiesce = true;
  } else if( TTInQuiesceOpt == "ProbeOnly" ) {
    UseTTInQuiesce = true;
    UseTTStoreInQuiesce = false;
  } else if( TTInQuiesceOpt == "No" ) {
    UseTTInQuiesce = UseTTStoreInQuiesce = false;
  }

  UseNullMove = options.getBool( "NullMove" );
  NullMoveMinDepth = options.getInt( "NullMoveMinDepth" );

  string NullMoveReductionOpt = options.getString( "NullMoveReduction" );
  if( NullMoveReductionOpt == "SmashAdaptive" ) {
    NullMoveReduction = NMR_FORMULA;
  } else if( NullMoveReductionOpt == "StandardAdaptive" ) {
    NullMoveReduction = NMR_ADAPTIVE;
  } else if( NullMoveReductionOpt == "R2" ) {
    NullMoveReduction = NMR_R2;
  } else if( NullMoveReductionOpt == "R3" ) {
    NullMoveReduction = NMR_R3;
  }
  
  ttable.ageing();
  assert( thread_data[0].moves_set[0] != thread_data[0].moves_set[1] );
  assert( NumThreads > 0 && NumThreads <= MAX_THREADS );
  for( int i = 0; i < NumThreads; ++i ) {
    thread_data[i].killers.clean();
    thread_data[i].kmate.clean();
    thread_data[i].history.reset( thread_data[i].board );
    if( !thread_data[i].active ) {
      assert( i != 0 );
      thread_data[i].active = true;
      InitThread( thread_data[i] );
    }
  }

  for( int i = NumThreads; i < MAX_THREADS; ++i ) {
    if( thread_data[i].active ) {
      thread_data[i].active = false;
    }
  }

  ThreadsNotifyAll();
  DestroyInactiveThreads();
  WaitThreadsInitialization( NumThreads - 1 );

  score_t alpha, beta;
  alpha = SCORE_MIN;
  beta = SCORE_MAX;

  if( flags.debug ) {
    cout << "info string Search start" << endl;
    cout << "info string fen: " << thread_data[0].board.fen() << endl;
  }

  score_t bestscore = SCORE_NULL;
  move_t bestmove;
  bool timeout = false;

  ResetThreadsTimeout();
  
  int init_treeply = thread_data[0].board.treeply;
  searchtime_start = GetTime();
  searchnodescnt_start = 0;

  for( int i = 0; i < NumThreads; ++i ) {
    searchnodescnt_start += thread_data[i].mystat.NodesCnt;
  }

  int nmoves = thread_data[0].moves_set[1] - thread_data[0].moves_set[0];
  switch( nmoves ) {
  case 0:
    root_score = SCORE_MIN;
    return NULLMOVE;

  case 1:
    return *thread_data[0].moves_set[0];

  default:
    root_score = SCORE_NULL;
  }

  root_depth = 1;

  uint64_t prev_nodes = 0;
  for( int i = 0; i < NumThreads; ++i ) {
    prev_nodes += thread_data[i].mystat.NodesCnt;
  }
  int prev_iter_nodes = 1;
  float branching_factor = 1.0;
  
  do {
    try {
      bestscore = SearchRoot( alpha, beta, root_depth, bestmove );

      // show branching factor
      uint64_t curr_nodes = 0;
      for( int i = 0; i < NumThreads; ++i ) {
	curr_nodes += thread_data[i].mystat.NodesCnt;
      }
      branching_factor = (float)(curr_nodes - prev_nodes) / (float)prev_iter_nodes;
      prev_iter_nodes = curr_nodes - prev_nodes;
      prev_nodes = curr_nodes;
      if( root_depth > 1 ) {
	cout << "info string branching factor = " << branching_factor << endl;
      }
      
    } catch( TimeException &e ) {
      WaitThreads();
      
      timeout = true;
      // reset board to initial position
      for( int i = 0; i < NumThreads; ++i ) {
	while( init_treeply < thread_data[i].board.treeply ) {
	  thread_data[i].board.undoMove();
	}
      }
    }
    ++root_depth;
  } while( !timeout && root_depth <= max_depth && timecontrol.nextPly() && bestscore < SCORE_MATEHI && bestscore > SCORE_MATELO);
      
  if( flags.debug ) {
    for( int i = 0; i < NumThreads; ++i ) {
      cout << "info string Statistics thread id:" << thread_data[i].thread_id << endl;
      thread_data[i].mystat.dump();
    }
  }

  PauseThreads();

  root_score = bestscore;
  return bestmove;
}

/*
  return bestscore and bestmove for root node at specified depth
  NOTE: this function should be called only by SearchIterate()
*/
score_t SearchRoot( score_t alpha, score_t beta, int depth, move_t &bestmove )
{
  // aliasing
  // NOTE: we should split in this function to parallelize search
  ThreadData &td( thread_data[0] );
    
  score_t root_move_score[MAX_PLY_MOVES];
  assert( td.moves_set[0] != td.moves_set[1] );

  root_move_score[0] = 0;
  bestmove = *td.moves_set[0];

  int i = 0;
  score_t g = alpha;
  int initply = td.board.treeply+1;
  
  for( move_t *m = td.moves_set[0]; m < td.moves_set[1]; ++m, ++i ) {
    // make move for each thread
    for( int tid = 0; tid < NumThreads; ++tid ) {
      thread_data[tid].board.makeMove( *m );
      thread_data[tid].board.setCheckStatus();
    }

    // null window: only if it's not the first move
    if( i != 0 ) {
      ResetThreadTermination();
      
      // start each thread
      for( int tid = 0; tid < NumThreads; ++tid ) {
	thread_data[tid].root_alpha = -alpha - 1;
	thread_data[tid].root_beta = -alpha;
	thread_data[tid].root_depth = depth-1;

	thread_data[tid].comm.bound[0].alpha = alpha;
	thread_data[tid].comm.bound[0].beta = beta;
	thread_data[tid].comm.bound[1].alpha = -alpha - 1;
	thread_data[tid].comm.bound[1].beta = -alpha;
      }
      if( depth > 1 ) {
	StartThreads();
      }
      // start this thread search
      try {
	g = -Search( td, -alpha - 1, -alpha, depth-1, 1 );
	TerminateThreadSearch( g );
      } catch( ThreadSearchTerminate &tst ) {
	g = thread_search_score;
	while( initply < td.board.treeply ) {
	  td.board.undoMove();
	}
      }
      assert( initply == td.board.treeply );


      // wait for each thread but this
      WaitThreads();
    }

    // full window: if it's the first move or we found a better
    // result
    if( i == 0 || g > alpha ) {
      ResetThreadTermination();
      
      // start each thread
      for( int tid = 0; tid < NumThreads; ++tid ) {
	thread_data[tid].root_alpha = -beta;
	thread_data[tid].root_beta = -alpha;
	thread_data[tid].root_depth = depth-1;

	thread_data[tid].comm.bound[0].alpha = alpha;
	thread_data[tid].comm.bound[0].beta = beta;
	thread_data[tid].comm.bound[1].alpha = -beta;
	thread_data[tid].comm.bound[1].beta = -alpha;
      }
      if( depth > 1 ) {
	StartThreads();
      }

      // start this thread search
      try {
	g = -Search( td, -beta, -alpha, depth-1, 1 );
	TerminateThreadSearch( g );
      } catch( ThreadSearchTerminate &tst ) {
	g = thread_search_score;
	while( initply < td.board.treeply ) {
	  td.board.undoMove();
	}
      }
      assert( initply == td.board.treeply );
	

      // wait for each thread but this
      WaitThreads();
    }

    for( int tid = 0; tid < NumThreads; ++tid ) {
      thread_data[tid].board.undoMove();
    }

    // sort moves incrementally
    move_t mt = *m;
    int k = i;
    move_t *m_set = td.moves_set[0];
    while( k > 0 && root_move_score[k-1] < g ) {
      root_move_score[k] = root_move_score[k-1];
      m_set[k] = m_set[k-1];
      --k;
    }
    root_move_score[k] = g;
    m_set[k] = mt;

    // check if this move is the best for now
    if( g > alpha ) {
      alpha = g;
      root_score = g;
      bestmove = mt;
      if( alpha >= beta ) {
	alpha = beta;
	break;
      } else {
	int timediff = max( GetTime() - searchtime_start, 1U );
	uint64_t nodescnt = 0;
	for( int i = 0; i < NumThreads; ++i ) {
	  nodescnt += thread_data[i].mystat.NodesCnt;
	}
	nodescnt -= searchnodescnt_start;
	if( flags.post || flags.debug ) {
	  DisplayPV( cout, root_depth, nodescnt, (nodescnt*100)/timediff, timediff * 10, bestmove );
	}
      }
    }

    if( alpha > SCORE_MATEHI ) {
      break;
    }
    
  }

  return alpha;
}

/*
  return score for the specified depth, bounding alpha and beta scores
*/
score_t Search( ThreadData &td, score_t alpha, score_t beta, int depth, int ply, bool exclusive )
{
  assert( alpha >= SCORE_MIN && beta <= SCORE_MAX && alpha <= beta );
  assert( depth >= 0 && depth < MAXPLY && ply > 0 && (ply + depth) < MAXPLY );
  assert( td.active );
  assert( td.board.isLegal() );

  // if an other thread found the result, exit search
  if( thread_search_terminate ) {
    throw ThreadSearchTerminate();
  }

  // bound alpha & beta by mate scores
  beta = min( beta, -EvalMate( ply+1 ) );
  alpha = max( alpha, EvalMate( ply ) );
  if( alpha >= beta ) {
    return beta;
  }

  // aliasing
  Board &board( td.board );
  SearchStatistics &mystat( td.mystat );
  const bool pvnode = alpha != beta - 1;

  // recognize draw
  // NOTE: depth > 0 in the condition is needed to don't 
  //       check for this in both Search() and Quiesce()
  RecognizerScore recog_result;
  if( depth > 0 && FastDrawRecognize( board, recog_result ) ) {
    switch( recog_result.scoretype ) {
    case SCORE_EXACT:
      ++mystat.NodesCnt;
      ++mystat.LeafNodes;
      return recog_result.score;

    case SCORE_FAILHIGH:
      alpha = max( alpha, recog_result.score );
      if( alpha >= beta ) {
	++mystat.NodesCnt;
	++mystat.LeafNodes;
	return beta;
      }
      break;

    case SCORE_FAILLOW:
      beta = min( beta, recog_result.score );
      if( beta <= alpha ) {
	++mystat.NodesCnt;
	++mystat.LeafNodes;
	return alpha;
      }
      break;

    default:
      assert( 0 );
    }
  }

  // shrink alpha and beta with current values
  td.comm.bound[ply].alpha.maximize( alpha );
  td.comm.bound[ply].beta.minimize( beta );

  // check for repetitions
  if( board.repetitions( 2, ply ) >= 2 ) {
    ++mystat.NodesCnt;
    ++mystat.LeafNodes;
    return SCORE_DRAW;
  } else if( board.repetitions( 3, MAXPLY ) >= 3 ) {
    ++mystat.NodesCnt;
    ++mystat.LeafNodes;
    return SCORE_DRAW;
  }
  
  // if depth <= 0 evaluate position after quiescence search
  if( depth <= 0 ) {
    return Quiesce( td, alpha, beta, ply );
  }
  
  ++mystat.NodesCnt;

  // check up for input and timer
  if( ( mystat.NodesCnt & TIMECHECK_NODES ) == 0 ) {
    CheckUp();
  }

  // if our root score is a mate score in <ply moves, then
  // we cut 'cause we can't do better
  // (I think we can also use >= here)
  if( root_score > SCORE_MAX - ply ) {
    return SCORE_MAX - ply - 1;
  }

  score_t oldalpha = alpha;
  bool do_null = true;

  // probe TT  
  TTEntry ttentry;
  ttentry.bestmove = NULLMOVE;
  ttentry.depth = depth;

  TTLockEntry lockentry( ttable_search );
  bool ttfound = lockentry.acquire( board, ttentry, exclusive );
  assert( !ttfound || ttentry.hashkey == board.hashkey );
  if( ttfound && ttentry.bestmove != NULLMOVE && !board.isLegalMove( ttentry.bestmove ) ) {
    ++mystat.TTBadEntries;
    ttfound = false;
    ttentry.bestmove = NULLMOVE;
  }


  // ABDADA: if it's on evaluation and we want it as exclusive...
  if( ttfound && exclusive && ttentry.score == SCORE_ONEVALUATION ) {
    if( NumThreads == 1 ) {
      cout << "Assertion failure:" << endl;
      cout << "Current depth = " << depth << endl;
      cout << "Entry depth = " << (int)ttentry.depth << endl;
      cout << "NPROC = " << (int)ttentry.nproc << endl;
      cout << "MoveList" << endl;
      for( int i = 0; i < board.treeply; ++i ) {
	cout << Move2Str( board.tree[i].move ) << endl;
      }
    }
    assert( NumThreads > 1 );
    return -SCORE_ONEVALUATION;
  }

  assert( !ttfound || ttentry.score != SCORE_ONEVALUATION );
  
  if( ttfound && !pvnode ) {

    // if is a mate result, change the depth to do the cut
    bool is_mate = IsTTMate( ttentry );
    if( is_mate && ttentry.depth < depth ) {
      ttentry.depth = depth;
    }
    
    if( ttentry.depth >= depth && ( !NearDraw( board, ply ) || ttentry.score == 0 || is_mate )  ) {
      ttentry.score = AdjustMateFromTT( ttentry.score, ply );
      switch( ttentry.scoretype ) {
      case SCORE_FAILHIGH:
	alpha = max( alpha, ttentry.score );
	if( alpha >= beta ) {
	  ++mystat.TTCuts;
	  ++mystat.FailHigh;

	  // add killer move
	  if( ttentry.bestmove != NULLMOVE && !IsCapture( ttentry.bestmove ) &&
	      !IsPromote( ttentry.bestmove ) && !board.tree[board.treeply].checked ) {
	    board.makeMove( ttentry.bestmove );
	    if( !board.setCheckStatus() ) {
	      td.killers.push( ply, ttentry.bestmove );
	      if( !IsCastling( ttentry.bestmove ) ) {
		td.history.push_good( ply, ttentry.depth, board, ttentry.bestmove );
	      }
	    }
	    board.undoMove();
	  }

	  return beta;
	}
	do_null = false; // bestmove is not enough: avoid null move
	break;

      case SCORE_FAILLOW:
	beta = min( beta, ttentry.score );
	if( beta <= alpha ) {
	  ++mystat.TTCuts;
	  ++mystat.FailLow;

	  return alpha;
	}
	break;

      case SCORE_EXACT: {
	++mystat.TTCuts;
	++mystat.ExactScore;

	if( ttentry.score > beta ) {
	  return beta;
	} else if( ttentry.score < alpha ) {
	  return alpha;
	} else {
	  return ttentry.score;
	}
      }

      default:
	assert( 0 );
      }
    }
  }

  // initialize bestmove
  move_t bestmove = NULLMOVE;
  int nmove = 0; // legal move counter
  move_t m;

  bool checked = board.tree[board.treeply].checked;

  // NULLMOVE
  // try to make a nullmove only if:
  // 1. we are not in check
  // 2. previous move isn't a null move
  // 3. we are not in PV line
  // 4. the depth >= 2
  // 5. we aren't going too bad
  // 6. we have sufficient material
  if( UseNullMove && !checked && board.tree[board.treeply-1].move != NULLMOVE && !pvnode
      && do_null
      && depth >= NullMoveMinDepth
      && board.getMaterial() >= beta // TODO: bad condition (try with lazy eval?)
      && DoNull( board )
      ) {

    ++mystat.NullMoves;

    int reduction;
    switch( NullMoveReduction ) {
    case NMR_FORMULA:
      reduction = (depth + 5)/4;
      break;

    case NMR_R2:
      reduction = 2;
      break;

    case NMR_R3:
      reduction = 3;
      break;

    case NMR_ADAPTIVE:
      reduction = depth >= 7 ? 3 : 2;
      break;

    default:
      assert( 0 );
    }

    MakeMove( td, NULLMOVE, depth - reduction - 1, false, ply );
    FixMoveHandler( td, ply );
    
    board.setCheckStatus();     // TODO: must be false!
    assert( !board.tree[board.treeply].checked );
      
    int g = -Search( td, -alpha - 1, -alpha, depth - reduction - 1, ply + 1 );

    UndoMove( td, ply );

    if( g >= beta ) {
      ++mystat.NullCuts;
      ++mystat.FailHigh;
      return beta;
    }

    // TODO: check for threat    
  }

  bool alldone = false;
  int iteration = 0;
  while( iteration < 2 && !alldone ) {
    MoveHandler movehandler( td, ply, ttentry.bestmove );
    ++iteration;
    alldone = true;
    nmove = 0;

    assert( iteration == 1 || NumThreads > 1 );

    // search deeper for each moves
    while( ( m = movehandler.nextMove() ) != NULLMOVE ) {
      MakeMove( td, m, depth, pvnode, ply );

      // skip non legal moves
      // NOTE: we don't check legality if this move is generated by GenEvasions()
      if( !checked && !board.isLegal() ) {
	UndoMove( td, ply );
	continue;
      }
      ++nmove;

      // skip first move in the second iteration
      if( nmove == 1 && iteration == 2 && ttentry.bestmove != NULLMOVE ) {
	UndoMove( td, ply );
	continue;
      }

      int extension = 0;
      bool do_check = false;
    
      /*
	One Reply Extension:
	  if we are in check and we can reply with just
	  one move we extend search depth by one.
	  This is a good thing for 2 reasons:
	  1. The branching factor of this node is 1
	  2. We could be close to a mate
      */
      if( checked && movehandler.singlemove ) {
	assert( nmove == 1 );
	++mystat.OneReplyExt;
	extension = 1;
      }
    
      // set board check status
      if( board.setCheckStatus() ) {
	/*
	  Check Extension:
	    If we check enemy king we extend search by one.
	    In this way, enemy can reply to the move and also
	    we can search deeper for some threat.
	*/
	extension = 1;
	do_check = true;
	++mystat.CheckExt;
      }
      
      assert( extension <= 1 );

      /*
	Pawn ending extension:
          if this move start an ending game with pawns only
	  we extend by 2 ply to be sure about the result
      */
      if( IsCapture( m ) ) {
	if( Capture( m ) != PAWN && board.getNumPieces() == board.getNumPieces( PAWN ) + 2 ) {
	  extension = 2;
	  ++mystat.PawnEndingExt;
	}
      }

      assert( extension <= 2 );

      // Limit search depth
      if( ply > ( root_depth * 2 ) ) {
	extension = 0;
      }


      score_t g;

      bool child_exclusive = iteration == 1 && nmove > 1;

      if( nmove != 1 ) {
	td.comm.bound[ply+1].depth = depth + extension - 1;
	td.comm.bound[ply+1].pvnode = false;
	g = -Search( td, -alpha - 1, -alpha, depth + extension - 1, ply + 1, child_exclusive );
      }

      if( nmove == 1 || ( g > alpha && g < beta ) ) {
	assert( nmove == 1 || pvnode );

	td.comm.bound[ply+1].depth = depth + extension - 1;
	td.comm.bound[ply+1].pvnode = pvnode;
	td.comm.bound[ply+1].alpha = -td.comm.bound[ply].beta.get();
	td.comm.bound[ply+1].beta = -td.comm.bound[ply].alpha.get();
	
	g = -Search( td, -beta, -alpha, depth + extension - 1, ply + 1, child_exclusive );
      }

      UndoMove( td, ply );

      // if this is the first move, we store it as bestmove
      if( nmove == 1 && iteration == 1 ) {
	assert( bestmove == NULLMOVE );
	bestmove = m;
      }

      assert( g != SCORE_NULL );
      if( g == SCORE_ONEVALUATION ) {
	assert( NumThreads > 1 && nmove > 0 && iteration == 1 );
	alldone = false;
      }

      // propagate bounds from previous ply
      td.comm.bound[ply].alpha.maximize( -td.comm.bound[ply-1].beta.get() );
      td.comm.bound[ply].beta.minimize( -td.comm.bound[ply-1].alpha.get() );

      // minimize beta with the shared one
      if( beta > td.comm.bound[ply].beta.get() ) {
	assert( NumThreads > 1 );
	beta = max( alpha, td.comm.bound[ply].beta.get() );
      }

      // maximize shared alpha and check some external thread improved it
      if( alpha < td.comm.bound[ply].alpha.get() ) {
	assert( NumThreads > 1 );
	alpha = min( td.comm.bound[ply].alpha.get(), beta );
	bestmove = NULLMOVE;
      }

      // check if we found a better move
      // and share this result with other threads
      if( g > alpha && alpha < beta ) {
	assert( g > EvalMate( ply ) && g <= -EvalMate( ply+1 ) );
	bestmove = m;
	alpha = g;
	td.comm.bound[ply].alpha.maximize( alpha );
	// propagate result to other threads only if:
	// 1. child is visited in exclusive mode
	// 2. we are not in exclusive mode or we didn't analyze
	//    all previous moves
	if( child_exclusive && (!exclusive||!alldone) ) {
	  Propagate( td, alpha, ply );
	}
      }

      // fail high
      if( alpha >= beta ) {
	// add killer move
	// if another thread found the bestmove, we don't store it
	if( bestmove != NULLMOVE ) {
	  if( !IsCapture( bestmove ) && !IsPromote( bestmove ) && !do_check ) {
	    td.killers.push( ply, bestmove );
	    if( !IsCastling( bestmove ) ) {
	      td.history.push_good( ply, depth + extension, board, bestmove );
	    }
	  }
	}
	// CUT
	alldone = true;
	break;
      }
    }
  }

  ttentry.bestmove = bestmove;
  ttentry.depth = depth;

  assert( bestmove != NULLMOVE || nmove == 0 || NumThreads > 1 );
  assert( alpha != SCORE_NULL && alpha != SCORE_ONEVALUATION );

  // no move == game over
  if( nmove == 0 ) {
    // ttentry.depth = MAXDEPTH; // mate and stalemate score are always exact with any depth
    assert( bestmove == NULLMOVE );

    assert( alpha == oldalpha );
    if( checked ) {
      // checkmate
      alpha = EvalMate( ply );
      ++mystat.LeafNodes;

      // insert killer mate move
      td.kmate.push( ply-1, board.tree[board.treeply-1].move );
    } else {
      // stalemate
      alpha = SCORE_DRAW;
      ++mystat.LeafNodes;
    }
  }

  // set new TT entry and store them
  if( ( alpha > oldalpha && alpha < beta ) || nmove == 0 ) {
    assert( pvnode || nmove == 0 );
    ++mystat.ExactScore;
    ttentry.scoretype = SCORE_EXACT;
    ttentry.score = AdjustMate2TT( alpha, ply );
  } else if( alpha >= beta ) {
    assert( nmove != 0 );
    ++mystat.FailHigh;
    ttentry.scoretype = SCORE_FAILHIGH;
    ttentry.score = AdjustMate2TT( beta, ply );
    alpha = beta;

    switch( nmove ) {
    case 1: ++mystat.CutFirstMove; break;
    case 2: ++mystat.CutSecondMove; break;
    }
  } else {
    assert( nmove != 0 );
    assert( alpha == oldalpha );
    ++mystat.FailLow;
    ttentry.scoretype = SCORE_FAILLOW;
    ttentry.score = AdjustMate2TT( alpha, ply );
  }
  
  if( bestmove != NULLMOVE || nmove == 0 ) {
    assert( ttentry.score != SCORE_NULL && ttentry.score != SCORE_ONEVALUATION );
    lockentry.store( board, ttentry );
  }

  return max( alpha, oldalpha );
}

/*
  compute evaluation of a leaf node avoiding horizon effect
*/
score_t Quiesce( ThreadData &td, score_t alpha, score_t beta, int ply )
{
  assert( alpha >= SCORE_MIN && beta <= SCORE_MAX && alpha <= beta );
  assert( ply > 0 && ply < MAXDEPTH );
  assert( td.active );
  assert( td.board.isLegal() );
  assert( td.board.repetitions( 3, MAXDEPTH ) < 3 );

  if( thread_search_terminate ) {
    throw ThreadSearchTerminate();
  }

  // bound alpha & beta by mate scores
  beta = min( beta, -EvalMate( ply+1 ) );
  alpha = max( alpha, EvalMate( ply ) );
  if( alpha >= beta ) {
    return beta;
  }

  // aliasing
  Board &board( td.board );
  SearchStatistics &mystat( td.mystat );
  const bool pvnode = alpha != beta-1;

  ++mystat.QNodes;
  ++mystat.NodesCnt;
  ++mystat.LeafNodes;

  if( ( mystat.NodesCnt & TIMECHECK_NODES ) == 0 ) {
    CheckUp();
  }


  bool checked = board.tree[board.treeply].checked;
  
  RecognizerScore recog_result;
  
  // recognize draw
  // TODO: add easy stalemate draw
  if( !checked && FastDrawRecognize( board, recog_result ) ) {
    switch( recog_result.scoretype ) {
    case SCORE_EXACT:
      return recog_result.score;

    case SCORE_FAILHIGH:
      alpha = max( alpha, recog_result.score );
      if( alpha >= beta ) {
	return beta;
      }
      break;

    case SCORE_FAILLOW:
      beta = min( beta, recog_result.score );
      if( beta <= alpha ) {
	return alpha;
      }
      break;

    default:
      assert( 0 );
    }
  }

  score_t eval = SCORE_NULL;
  score_t oldalpha = alpha;

  // probe TT
  TTLockEntry lockentry( ttable_search );
  TTEntry ttentry;
  ttentry.depth = 0;
  ttentry.bestmove = NULLMOVE;

  bool ttfound = UseTTInQuiesce && lockentry.acquire( board, ttentry, false );
  if( ttfound && ttentry.bestmove != NULLMOVE && !board.isLegalMove( ttentry.bestmove ) ) {
    cerr << board << endl << "BadMove: " << Move2Str( ttentry.bestmove ) << endl;
    fprintf( stderr, "board.hashkey: %llx, entry.hashkey: %llx\n", board.hashkey, ttentry.hashkey );
    ++mystat.TTBadEntries;
    ttfound = false;
    ttentry.bestmove = NULLMOVE;
  }
  if( ttfound ) {
    ttentry.score = AdjustMateFromTT( ttentry.score, ply );
    if( ttentry.scoretype == SCORE_FAILHIGH ) {
      alpha = max( alpha, ttentry.score );
      if( alpha >= beta ) {
	++mystat.QTTCuts;
	++mystat.QFailHigh;
	return beta;
      }
    } else if( ttentry.scoretype == SCORE_FAILLOW ) {
      beta = min( beta, ttentry.score );
      if( beta <= alpha ) {
	++mystat.QTTCuts;
	++mystat.QFailLow;
	return alpha;
      }
    } else if( ttentry.scoretype == SCORE_EXACT ) {
      ++mystat.QTTCuts;
      ++mystat.QExactScore;
      if( ttentry.score > beta ) {
	return beta;
      } else if( ttentry.score < alpha ) {
	return alpha;
      } else {
	return ttentry.score;
      }
    } else {
      assert( 0 );
    }
  }

  // evaluate statically the board, if not in check
  // do stand pat cut
  if( !checked ) {
    eval = Eval( board );
    alpha = max( alpha, eval );
    if( alpha >= beta ) {
      ++mystat.QStandPat;
      return beta;
    }
  }

  MoveHandler movehandler( td, ply, ttentry.bestmove, MH_CAPTURES );
  
  move_t bestmove = NULLMOVE;
  
  // try each move
  move_t m;
  while( ( m = movehandler.nextMove() ) != NULLMOVE ) {
    assert( checked || IsCapture( m ) || IsPromote( m ) );

    MakeMove( td, m, 0, false, ply );

    // check legality
    // NOTE: the move is surely legal if we have generated evasions
    if( !checked && !board.isLegal() ) {
      UndoMove( td, ply );
      continue;
    }


    bool docheck = board.setCheckStatus();
    bool deltapruning = false;

    // delta pruning euristics:
    //  if we are going too bad (eval < alpha) and with this capture
    //  the score is even bad, cut
    // NOTE: we use a margin called DeltaPruningMargin
    // NOTE: we don't use delta pruning when we are in check or the
    // move checks enemy king, also avoid delta pruning in PV line
    if( UseDeltaPruning && !pvnode && !checked && !docheck ) {
      score_t move_score = 0;
      if( IsCapture( m ) ) {
	move_score += PIECE_VALUE[Capture( m )];
      }
      if( IsPromote( m ) ) {
	move_score += PIECE_VALUE[Promote( m )] - PIECE_VALUE[PAWN];
      }
      if( eval + move_score + DeltaPruningMargin <= alpha && ttentry.bestmove != m ) {
	deltapruning = true;
	++mystat.QDeltaPruning;
      }
    }

    bool draw = false;
    if( checked ) {
      // if we was in check, repetitions could be happen
      if( board.repetitions( 2, ply+1 ) >= 2 ) {
	draw = true;
      } else if( board.repetitions( 3, MAXDEPTH ) >= 3 ) {
	draw = true;
      }
    }

    score_t g = SCORE_NULL;
    if( !deltapruning ) {
      g = draw ? -SCORE_DRAW : -Quiesce( td, -beta, -alpha, ply+1 );
    }
    UndoMove( td, ply );

    // if we don't have a bestmove, this is the bestmove
    if( bestmove == NULLMOVE && !deltapruning ) {
      bestmove = m;
    }

    assert( deltapruning || ( g > EvalMate( ply ) && g <= -EvalMate( ply+1 ) ) );
    // update alpha/beta and bestmove
    if( g > alpha ) {
      assert( !deltapruning );
      alpha = g;
      bestmove = m;
      if( alpha >= beta ) {
	break;
      }
    }
  }

  // if we don't have a good move and we are in check,
  // we are in checkmate, otherwise it's a regular fail low
  ttentry.bestmove = bestmove;
  ttentry.depth = 0;

  if( alpha >= beta ) {
    ++mystat.QFailHigh;
    ttentry.scoretype = SCORE_FAILHIGH;
    ttentry.score = AdjustMate2TT( beta, ply );
    alpha = beta;
  } else if( bestmove == NULLMOVE && checked ) {
    assert( board.isInCheck() );
    ++mystat.QExactScore;
    score_t score = EvalMate( ply );
    ttentry.score = AdjustMate2TT( score, ply );
    ttentry.scoretype = SCORE_EXACT;
    //ttentry.depth = MAXDEPTH;
    if( score < alpha ) {
      score = alpha;
    } else if( score > beta ) {
      score = beta;
    }
    alpha = score;
  } else {
    if( alpha > oldalpha ) {
      ++mystat.QExactScore;
      ttentry.scoretype = SCORE_EXACT;
    } else {
      ++mystat.QFailLow;
      ttentry.scoretype = SCORE_FAILLOW;
    }
    ttentry.score = AdjustMate2TT( max( alpha, oldalpha ), ply );
  }

  if( UseTTInQuiesce && UseTTStoreInQuiesce ) {
    lockentry.store( board, ttentry );
  }

  return max( alpha, oldalpha );
}

// return mate score for "mate in ply"
inline score_t EvalMate( int ply )
{
  assert( ply >= 0 && ply < MAXDEPTH );
  return ply - SCORE_MAX;
}

// transform mate score for TT
inline score_t AdjustMate2TT( score_t score, int ply )
{
  assert( score >= EvalMate( ply ) && score <= -EvalMate( ply+1 ) );
  assert( ply >= 0 && ply < MAXDEPTH );
  if( score > SCORE_MATEHI ) {
    score += ply;
  } else if( score < SCORE_MATELO ) {
    score -= ply;
  }
  assert( score >= SCORE_MIN && score <= SCORE_MAX );
  return score;
}

// transform mate score from TT
inline score_t AdjustMateFromTT( score_t score, int ply )
{
  assert( score >= SCORE_MIN && score <= SCORE_MAX );
  assert( ply >= 0 && ply < MAXDEPTH );
  if( score < SCORE_MATELO ) {
    score += ply;
  } else if( score > SCORE_MATEHI ) {
    score -= ply;
  }
  assert( score >= EvalMate( ply ) && score <= -EvalMate( ply+1 ) );
  return score;
}

// check if is a TT mate score
inline bool IsTTMate( const TTEntry &entry )
{
  if( entry.scoretype == SCORE_EXACT ) {
    if( entry.score <= SCORE_MATELO ) {
      return true;
    } else if( entry.score >= SCORE_MATEHI ) {
      return true;
    }
  }
  return false;
}


// PV manip

// display current PV line
void DisplayPV( std::ostream &out, int depth, int nodescnt, int nps, int totaltime, move_t move )
{
  Board &board( thread_data[0].board );
  
  move_t pv[MAXDEPTH];
  score_t pv_scores[MAXDEPTH];
  hash_t hashkeys[MAXDEPTH];
  int pvlen = 1;

  pv[0] = move;
  pv_scores[0] = root_score;
  assert( board.isLegalMove( move ) );

  board.makeMove( pv[0] );


  // find PV from TT
  while( pvlen < MAXDEPTH ) {
    hashkeys[pvlen] = board.hashkey;

    TTLockEntry lockentry( ttable );
    TTEntry ttentry;
    ttentry.depth = 0;
    if( !lockentry.acquire( board, ttentry, false ) ) {
      break;
    }
    assert( ttentry.score != SCORE_ONEVALUATION );
    if( ttentry.bestmove == NULLMOVE || !board.isLegalMove( ttentry.bestmove ) ) {
      break;
    }
      
    board.makeMove( ttentry.bestmove );
    if( board.repetitions( 2, pvlen + 1 ) == 2 ) {
      board.undoMove();
      break;
    }
    
    pv_scores[pvlen] = ttentry.score;
    pv[pvlen++] = ttentry.bestmove;
  }

  
  // restore initial board position
  for( int i = 0; i < pvlen; ++i ) {
    board.undoMove();
  }
  
  out << "info depth " << depth << " seldepth " << max( depth, pvlen ) << endl;
  out << "info nps " << nps << endl;
  out << "info score cp " << pv_scores[0] << " depth " << depth << " nodes " << nodescnt << " time " << totaltime;
  out << " pv ";
  for( int i = 0; i < pvlen; ++i ) {
    if( i > 0 ) {
      out << " ";
    }
    out << Move2Str( pv[i] );
  }
  out << endl;

  if( flags.debug ) {
    out << "info string nodecnt";
    for( int i = 0; i < NumThreads; ++i ) {
      out << " thread " << thread_data[i].thread_id << ": " << thread_data[i].mystat.NodesCnt;
    }
    out << endl;
  }
}


// Utility functions


// Check input and timer
inline void CheckUp()
{
  {
    Locker locker( lock_io );
    if( InputAvailable() ) {
      HandleInput( STATE_SEARCH );
    }
  }
  
  if( timecontrol.check() || IsThreadTimeout() ) {
    if( flags.debug ) {
      cout << "info string Timeout!" << endl;
    }
    ThreadsTimeout();
    throw TimeException();
  }
}

// check if side to move has sufficient material to make a nullmove
// 1. if stm has queen is ok
// 2. if stm has more than 1 piece other than king is ok
inline bool DoNull( const Board &board )
{
  int side = board.side;
  
  if( board.getNumPieces( side, QUEEN ) > 0 ) {
    return true;
  }

  if( board.getNumPieces( side ) - board.getNumPieces( side, PAWN ) - 1 >= 2 ) {
    return true;
  }

  return false;
}

// check if we are near a draw
// 1. check if the rule50 is high
// 2. TODO: check if we are near draw by repetition
inline bool NearDraw( const Board &board, int ply )
{
  if( board.rule50 > 80 ) {
    return true;
  } else {
    return false;
  }
}


// ABDADA Communication functions

// make move and update bounds
inline static void MakeMove( ThreadData &td, move_t m, int depth, bool pvnode, int ply )
{
  assert( td.active );
  assert( ply >= 0 && ply+1 < MAXPLY );
  Lock( td.comm.bound[ply].lock );
  td.board.makeMove( m );
  td.comm.bound[ply+1].alpha = -td.comm.bound[ply].beta.get();
  td.comm.bound[ply+1].beta = -td.comm.bound[ply].alpha.get();
  td.comm.bound[ply+1].pvnode = pvnode;
  td.comm.bound[ply+1].depth = depth;
  Unlock( td.comm.bound[ply].lock );
}

// unmake move and reset bounds
inline static void UndoMove( ThreadData &td, int ply )
{
  assert( td.active );
  assert( ply >= 0 && ply+1 < MAXPLY );
  
  td.board.undoMove();
#if defined(USESMP)
  td.comm.bound[ply+1].alpha = SCORE_MAX;
  td.comm.bound[ply+1].beta = SCORE_MIN;
#endif
}

// propagate bounds improvement to other threads
inline static void Propagate( ThreadData &current_thread, score_t alpha, int ply )
{
  assert( current_thread.active );
  assert( alpha >= SCORE_MIN && alpha <= SCORE_MAX );
  assert( ply > 0 && ply < MAXDEPTH );

#if defined(USESMP)
  hash_t hashkey = current_thread.board.tree[ply].hashkey;
  for( int i = 0; i < NumThreads; ++i ) {
    assert( thread_data[i].active );
    if( i == current_thread.thread_id
	|| thread_data[i].comm.board.tree[ply].hashkey != hashkey ) {
      continue;
    }

    ThreadData &td( thread_data[i] );

    Locker locker( td.comm.bound[ply].lock );

    // check again after lock that the other thread is searching this node
    // also depth and pvnode must be the same
    if( td.comm.board.tree[ply].hashkey != hashkey
	|| td.comm.bound[ply].depth > current_thread.comm.bound[ply].depth
	|| (td.comm.bound[ply].pvnode && !current_thread.comm.bound[ply].pvnode) ) {
      continue;
    }

    // propagate alpha score
    int p = ply;
    bool done = false;
    while( p < MAXPLY && !done ) {
      if( (p-ply) % 2 == 0 ) {
	done = td.comm.bound[p].alpha.maximize( alpha ) != alpha;
      } else {
	done = td.comm.bound[p].beta.minimize( -alpha ) != -alpha;
      }
      ++p;
    }
  }
#endif
}
