/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
/*
  See function

  Created in 15/02/2006
  Maurizio
*/

#include <string>
#include "see.h"
#include "smash.h"
#include "board.h"
#include "move.h"
#include "timecontrol.h"
#include "bitboard.h"
#include "movegen.h"

using namespace std;


inline bitboard_t XRay( const Board &board, int side, square_t src, square_t dst );
static score_t SEE_EnPassant( const Board &board, const move_t move );

// static exchange evaluation
// return score resulting after captures exchange
// NOTE: it is not *perfect* score, it doesn't consider:
// 1. pinned piece
// 2. double attacks
score_t SEE( const Board &board, const move_t move )
{
  // special: double pawn move
  // check for en passant!
  score_t score_enpassant = SCORE_MAX;
  if( IsDoublePawnPush( move ) ) {
    score_enpassant = SEE_EnPassant( board, move );
  }
  
  score_t score[32] = {0};
  bitboard_t bb_attackers[PLAYERS];

  square_t src = Src(move), dst = Dst(move);

  bb_attackers[WHITE] = board.piecesAttack( WHITE, dst );
  bb_attackers[BLACK] = board.piecesAttack( BLACK, dst );

  int side = GetPieceColor( board[src] );
  int xside = ChangeSide( side );

  // special: en passant
  // add right piece to attackers
  if( IsEnPassant( move ) ) {
    square_t step = side == WHITE ? MAXCOL : -MAXCOL;
    bb_attackers[WHITE] |= XRay( board, WHITE, dst - step, dst - step*2 );
    bb_attackers[BLACK] |= XRay( board, BLACK, dst - step, dst - step*2 );
    score[0] = PAWN_VALUE;
  } else {
    score[0] = PIECE_VALUE[GetPiece(board[dst])];
  }

  // remove first moved piece from attackers list
  // and insert pieces behind it
  bb_attackers[WHITE] |= XRay( board, WHITE, dst, src );
  bb_attackers[BLACK] |= XRay( board, BLACK, dst, src );
  ResetBit( bb_attackers[side], src );

  piece_t curr_piece = GetPiece( board[src] );

  // special: promote
  // promote the pawn
  if( IsPromote( move ) ) {
    assert( curr_piece == PAWN );
    curr_piece = Promote( move );
    score[0] += PIECE_VALUE[curr_piece] - PAWN_VALUE;
  }

  int recaptures = 0;
  int stm = xside;

  bitboard_t bb[] = {
    board.getBBPieces( WHITE, PAWN ) | board.getBBPieces( BLACK, PAWN ),
    board.getBBPieces( WHITE, KNIGHT ) | board.getBBPieces( BLACK, KNIGHT ),
    board.getBBPieces( WHITE, BISHOP ) | board.getBBPieces( BLACK, BISHOP ),
    board.getBBPieces( WHITE, ROOK ) | board.getBBPieces( BLACK, ROOK ),
    board.getBBPieces( WHITE, QUEEN ) | board.getBBPieces( BLACK, QUEEN )
  };

  // SEE, finally :)
  while( bb_attackers[stm] ) {
    if( bb_attackers[stm] & bb[0] ) {
      src = FirstBit( bb_attackers[stm] & bb[0] );
      assert( GetPiece( board[src] ) == PAWN );
    } else if( bb_attackers[stm] & bb[1] ) {
      src = FirstBit( bb_attackers[stm] & bb[1] );
      assert( GetPiece( board[src] ) == KNIGHT );
    } else if( bb_attackers[stm] & bb[2] ) {
      src = FirstBit( bb_attackers[stm] & bb[2] );
      assert( GetPiece( board[src] ) == BISHOP );
    } else if( bb_attackers[stm] & bb[3] ) {
      src = FirstBit( bb_attackers[stm] & bb[3] );
      assert( GetPiece( board[src] ) == ROOK );
    } else if( bb_attackers[stm] & bb[4] ) {
      src = FirstBit( bb_attackers[stm] & bb[4] );
      assert( GetPiece( board[src] ) == QUEEN );
    } else if( !bb_attackers[ChangeSide( stm )] ) {
      // king... only if there are no enemy attackers
      src = FirstBit( bb_attackers[stm] );
      assert( GetPiece( board[src] ) == KING );
    } else {
      break;
    }

    // update attackers bb
    ToggleBit( bb_attackers[stm], src );
    bb_attackers[WHITE] |= XRay( board, WHITE, dst, src );
    bb_attackers[BLACK] |= XRay( board, BLACK, dst, src );

    // update score
    ++recaptures;
    if( stm == side ) {
      score[recaptures] = score[recaptures-1] + PIECE_VALUE[curr_piece];
    } else {
      score[recaptures] = score[recaptures-1] - PIECE_VALUE[curr_piece];
    }

    // update current piece and side to move
    curr_piece = GetPiece( board[src] );
    stm = ChangeSide( stm );
  }

  // minimax
  for( int i = recaptures; i > 0; --i ) {
    if( i % 2 ) {
      score[i-1] = min( score[i], score[i-1] );
    } else {
      score[i-1] = max( score[i], score[i-1] );
    }
  }

  return min( score[0], score_enpassant );
}

static score_t SEE_EnPassant( const Board &board, const move_t move )
{
  score_t score[32] = {0};
  bitboard_t bb_attackers[PLAYERS];

  square_t src = Src(move), dst = Dst(move);

  int side = GetPieceColor( board[src] );
  int xside = ChangeSide( side );

  square_t step = side == WHITE ? MAXCOL : -MAXCOL;
  dst -= step;

  // special: first move must be a pawn capture
  if( !( board.getBBPieces( xside, PAWN ) & pawn_captures[side][dst] ) ) {
    return 0;
  }
  bb_attackers[WHITE] = board.piecesAttack( WHITE, dst );
  bb_attackers[BLACK] = board.piecesAttack( BLACK, dst );

  // remove first moved piece from attackers list
  // and insert pieces behind it
  bb_attackers[WHITE] |= XRay( board, WHITE, dst, src );
  bb_attackers[BLACK] |= XRay( board, BLACK, dst, src );
  ResetBit( bb_attackers[side], src );

  bb_attackers[WHITE] |= XRay( board, WHITE, src, dst );
  bb_attackers[BLACK] |= XRay( board, BLACK, src, dst );

  piece_t curr_piece = GetPiece( board[src] );

  int recaptures = 0;
  int stm = xside;

  bitboard_t bb[] = {
    board.getBBPieces( WHITE, PAWN ) | board.getBBPieces( BLACK, PAWN ),
    board.getBBPieces( WHITE, KNIGHT ) | board.getBBPieces( BLACK, KNIGHT ),
    board.getBBPieces( WHITE, BISHOP ) | board.getBBPieces( BLACK, BISHOP ),
    board.getBBPieces( WHITE, ROOK ) | board.getBBPieces( BLACK, ROOK ),
    board.getBBPieces( WHITE, QUEEN ) | board.getBBPieces( BLACK, QUEEN )
  };


  // SEE, finally :)
  while( bb_attackers[stm] ) {
    if( bb_attackers[stm] & bb[0] ) {
      src = FirstBit( bb_attackers[stm] & bb[0] );
      assert( GetPiece( board[src] ) == PAWN );
    } else if( bb_attackers[stm] & bb[1] ) {
      src = FirstBit( bb_attackers[stm] & bb[1] );
      assert( GetPiece( board[src] ) == KNIGHT );
    } else if( bb_attackers[stm] & bb[2] ) {
      src = FirstBit( bb_attackers[stm] & bb[2] );
      assert( GetPiece( board[src] ) == BISHOP );
    } else if( bb_attackers[stm] & bb[3] ) {
      src = FirstBit( bb_attackers[stm] & bb[3] );
      assert( GetPiece( board[src] ) == ROOK );
    } else if( bb_attackers[stm] & bb[4] ) {
      src = FirstBit( bb_attackers[stm] & bb[4] );
      assert( GetPiece( board[src] ) == QUEEN );
    } else if( !bb_attackers[ChangeSide( stm )] ) {
      // king... only if there are no enemy attackers
      src = FirstBit( bb_attackers[stm] );
      assert( GetPiece( board[src] ) == KING );
    } else {
      break;
    }

    // update attackers bb
    ToggleBit( bb_attackers[stm], src );
    bb_attackers[WHITE] |= XRay( board, WHITE, dst, src );
    bb_attackers[BLACK] |= XRay( board, BLACK, dst, src );

    // update score
    ++recaptures;
    if( stm == side ) {
      score[recaptures] = score[recaptures-1] + PIECE_VALUE[curr_piece];
    } else {
      score[recaptures] = score[recaptures-1] - PIECE_VALUE[curr_piece];
    }

    // update current piece and side to move
    curr_piece = GetPiece( board[src] );
    stm = ChangeSide( stm );
  }

  // minimax
  for( int i = recaptures; i > 0; --i ) {
    if( i % 2 ) {
      score[i-1] = min( score[i], score[i-1] );
    } else {
      score[i-1] = max( score[i], score[i-1] );
    }
  }

  return score[0];
}
      
      
// find attackers behind last moved piece
inline bitboard_t XRay( const Board &board, int side, square_t src, square_t dst )
{
  switch( ray_type[src][dst] ) {
  case RAY_ROOK:
    return board.rooksQueensAttack( side, dst ) & bb_ray[dst][(int)square_last[src][dst]];

  case RAY_BISHOP:
    return board.bishopsQueensAttack( side, dst ) & bb_ray[dst][(int)square_last[src][dst]];

  default:
    return NULLBOARD;
  }
}


#ifdef TEST

#include "search.h"

string FlipFen( const string &fen );
move_t FlipMove( move_t move );

void TestSEE()
{
  cout << endl << ">>--------------------------------- SEE Test ---------------------------------<<" << endl;

  struct {
    const char *fen;
    const char *move;
    score_t score;
  } suite[] = {
    { "7k/8/3Q4/8/2p5/8/3P3K/3r4 w - - 0 1", "d2d4", -PAWN_VALUE },
    { "7k/8/3Q4/5B2/2p5/8/3P3K/3r4 w - - 0 1", "d2d4", 0 },
    { "7k/4r3/8/4nK2/5P2/8/8/8 w - - 0 1", "f4e5", KNIGHT_VALUE },
    { "7k/4r3/8/4nK2/5P2/2b5/8/8 w - - 0 1", "f4e5", KNIGHT_VALUE - PAWN_VALUE },
    { "7k/4r3/8/2Q1nK2/5P2/2b5/8/8 w - - 0 1", "f4e5", KNIGHT_VALUE - PAWN_VALUE },
    { "7k/4r3/2N5/4nK2/5P2/2b5/8/8 w - - 0 1", "f4e5", KNIGHT_VALUE },
    { "7k/4r3/2N5/4nK2/5P2/2b5/1q6/8 w - - 0 1", "f4e5", KNIGHT_VALUE - PAWN_VALUE + BISHOP_VALUE - KNIGHT_VALUE },
    { "k7/4r3/2N5/4nK2/5P2/2b3B1/1q6/8 w - - 0 1", "f4e5", KNIGHT_VALUE }
  };

  const int SPEED_TEST_CASE_ID = 7;
  const int SPEED_TEST_REP = 1000000;

  unsigned int good, total;
  good = total = 0;

  for( unsigned int i = 0; i < sizeof( suite )/sizeof( suite[0] ); ++i ) {
    cout << ">> Testing position n." << (i+1) << endl;
    cout << "FEN: " << suite[i].fen << " MOVE: " << suite[i].move << " SCORE: " << suite[i].score << endl;
    board.loadFen( suite[i].fen );
    move_t move = Str2Move( board, suite[i].move );

    score_t score = SEE( board, move );

    int ok = 0;
    
    cout << "SEE SCORE: " << score;
    if( score == suite[i].score ) {
      ok |= 1;
    }

    // flipped board...
    board.loadFen( FlipFen( suite[i].fen ) );
    move = FlipMove( move );

    score_t score_flipped = SEE( board, move );

    if( score == score_flipped ) {
      ok |= 2;
    }
    
    if( ok == 3 ) {
      cout << " Good :-) ";
      ++good;
    } else {
      cout << "  Bad :-( ";
      if( (ok & 1) == 0 ) {
	cout << " [Wrong score] ";
      }
      if( (ok & 2) == 0 ) {
	cout << " [Flipped score different than original] ";
      }
    }
    ++total;
    cout << good << "/" << total << endl;
  }

  cout << "Speed test..." << endl;
  board.loadFen( suite[SPEED_TEST_CASE_ID].fen );
  move_t move = Str2Move( board, suite[SPEED_TEST_CASE_ID].move );
  int time_start = GetTime();
  for( int i = 0; i < SPEED_TEST_REP; ++i ) {
    SEE( board, move );
  }
  int time_diff = GetTime() - time_start;
  cout << "Time to repeat " << SPEED_TEST_REP << " times SEE(): "
       << (time_diff/100) << "." << ((time_diff/10)%10) << (time_diff%10) << "s" << endl;
  cout << "Average time is: " << ((time_diff/SPEED_TEST_REP)/100) << "."
       << (((time_diff/SPEED_TEST_REP)/10)%10) << ((time_diff/SPEED_TEST_REP)%10) << "s" << endl;
}
#endif
