/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#include <string>

#include "smash.h"
#include "bitboard.h"
#include "movegen.h"
#include "movehandler.h"
#include "search.h"
#include "options.h"

#include <cstdarg>

using namespace std;

static bool ResizeHash( Options &opt, const string &name, const string &value );

void Initialize()
{
  InitBB();
  InitMoveGen();

  // initialize search and move handler for each thread and start them
  // initialization
  InitSearch();
  for( int i = 0; i < MAX_THREADS; ++i ) {
    thread_data[i].thread_id = i;
    InitMoveHandler( thread_data[i] );
    ResetSearch( thread_data[i] );
    thread_data[i].active = i == 0;
  }
  InitPauseThreads();

  // initialize zobrist key and reset board for each thread
  InitializeZobristKey();
  for( int i = 0; i < MAX_THREADS; ++i ) {
    thread_data[i].board.reset();
  }

  // initialize evaluation function
  InitEval();

  // initialize options
  options.addOption( "Hash", "spin", "0", 0, 100, ResizeHash );
  
  options.addOption( "DeltaPruning", "check", "true", 0 );
  options.addOption( "DeltaPruningMargin", "spin", "150", 0, 500, 0 );
  
  options.addOption( "TTInQuiesce", "combo", "No", 0 );
  options.addVar( "TTInQuiesce", "Yes" );
  options.addVar( "TTInQuiesce", "ProbeOnly" );
  options.addVar( "TTInQuiesce", "No" );

  options.addOption( "HistoryHeuristicInMoveOrder", "check", "true", 0 );
  
  options.addOption( "HistoryHeuristicStartValue", "combo", "FromPST", 0 );
  options.addVar( "HistoryHeuristicStartValue", "FromPST" );
  options.addVar( "HistoryHeuristicStartValue", "Zero" );

  options.addOption( "HistoryHeuristicScore", "combo", "DepthAndPly", 0 );
  options.addVar( "HistoryHeuristicScore", "DepthAndDepth" );
  options.addVar( "HistoryHeuristicScore", "DepthAndPly" );
  options.addVar( "HistoryHeuristicScore", "PlyAndPly" );

  options.addOption( "DifferentialHistoryHeuristic", "check", "true", 0 );

  options.addOption( "NullMove", "check", "true", 0 );
  options.addOption( "NullMoveMinDepth", "spin", "2", 0, 64, 0 );
  options.addOption( "NullMoveReduction", "combo", "SmashAdaptive", 0 );
  options.addVar( "NullMoveReduction", "SmashAdaptive" );
  options.addVar( "NullMoveReduction", "StandardAdaptive" );
  options.addVar( "NullMoveReduction", "R2" );
  options.addVar( "NullMoveReduction", "R3" );

  options.addOption( "NumThreads", "spin", "1", 1, MAX_THREADS, 0 );

  flags.quit = false;
}

void Destroy()
{
  for( int i = 0; i < MAX_THREADS; ++i ) {
    thread_data[i].active = false;
  }
  ThreadsNotifyAll();
  DestroyInactiveThreads();
  WaitThreadsInitialization( 0 );

  flags.quit = true;
}


static bool ResizeHash( Options &opt, const string &name, const string &value )
{
  int megabytes = atoi( value.c_str() );
  ttable.resize( megabytes * 1024 * 1024 );
  return true;
}
