/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#ifndef _SMASH_H_
#define _SMASH_H_

/*
  smash global constants, types and some functions

  Created in 28/06/2004
  Maurizio
*/

#include <stdint.h>
#include <algorithm>
#include <cmath>
#include <string>

typedef unsigned long long uint64_t;
//typedef unsigned long uint32_t;

// max depth reacheable by search
const int MAXDEPTH = 100;

// max number of ply in a match
const int MAXPLY = 500;

//////////////////////////////////////////////
// Board square type and constants

typedef int square_t;

// board size constants
const int MAXROW = 8;
const int MAXCOL = 8;
const int MAXPOS = MAXROW*MAXCOL;

// board square constants
enum {
  A1, B1, C1, D1, E1, F1, G1, H1,
  A2, B2, C2, D2, E2, F2, G2, H2,
  A3, B3, C3, D3, E3, F3, G3, H3,
  A4, B4, C4, D4, E4, F4, G4, H4,
  A5, B5, C5, D5, E5, F5, G5, H5,
  A6, B6, C6, D6, E6, F6, G6, H6,
  A7, B7, C7, D7, E7, F7, G7, H7,
  A8, B8, C8, D8, E8, F8, G8, H8
};

enum {
  FILE_A, FILE_B, FILE_C, FILE_D,
  FILE_E, FILE_F, FILE_G, FILE_H
};

enum {
  RANK_1, RANK_2, RANK_3, RANK_4,
  RANK_5, RANK_6, RANK_7, RANK_8
};

extern const char *square2str[];

const square_t SQUARE_NULL = -1;

inline int Column( square_t sq )
{
  return sq % MAXCOL;
}

inline int Row( square_t sq )
{
  return sq / MAXCOL;
}

inline square_t CR2Square( int col, int row )
{
  return row*MAXCOL + col;
}

inline int FlipRow( int row )
{
  return MAXROW - row - 1;
}

inline square_t FlipSquare( square_t sq )
{
  return CR2Square( Column( sq ), MAXROW - Row( sq ) - 1 );
}

inline int Distance( square_t a, square_t b )
{
  return std::max( std::abs( Column( a ) - Column( b ) ),
		   std::abs( Row( a ) - Row( b ) ) );
}

inline bool IsSquareOk( square_t sq )
{
  return sq >= 0 && sq < MAXPOS;
}

const int WHITE = 0;
const int BLACK = 1;
const int PLAYERS = 2;

inline int ChangeSide( int side )
{
  return side ^ 1;
}

///////////////////////////////////////////////
// Piece constants and types

typedef unsigned int piece_t;

const piece_t EMPTY = 0;
const piece_t PAWN = 1;
const piece_t KNIGHT = 2;
const piece_t BISHOP = 3;
const piece_t ROOK = 4;
const piece_t QUEEN = 5;
const piece_t KING = 6;
const int PIECES_NUM = 7;

const piece_t WHITE_PIECE = 0;
const piece_t BLACK_PIECE = 8;
const piece_t PLAYER_PIECE_MASK = 8;

const piece_t WHITE_PAWN = PAWN | WHITE_PIECE;
const piece_t WHITE_KNIGHT = KNIGHT | WHITE_PIECE;
const piece_t WHITE_BISHOP = BISHOP | WHITE_PIECE;
const piece_t WHITE_ROOK = ROOK | WHITE_PIECE;
const piece_t WHITE_QUEEN = QUEEN | WHITE_PIECE;
const piece_t WHITE_KING = KING | WHITE_PIECE;

const piece_t BLACK_PAWN = PAWN | BLACK_PIECE;
const piece_t BLACK_KNIGHT = KNIGHT | BLACK_PIECE;
const piece_t BLACK_BISHOP = BISHOP | BLACK_PIECE;
const piece_t BLACK_ROOK = ROOK | BLACK_PIECE;
const piece_t BLACK_QUEEN = QUEEN | BLACK_PIECE;
const piece_t BLACK_KING = KING | BLACK_PIECE;

extern const char *piece2str[];

inline piece_t GetPiece( piece_t p )
{
  return p & ~PLAYER_PIECE_MASK;
}

inline int GetPieceColor( piece_t p )
{
  return ( p & PLAYER_PIECE_MASK ) == WHITE_PIECE ? WHITE : BLACK;
}


/////////////////////////////////////////////
// Castling flags

const int NO_CASTLING = 0;           // castling unavailable
const int CASTLING_KING_SIDE = 1;    // short castling available
const int CASTLING_QUEEN_SIDE = 2;   // long castling available



// engine status structure
struct Status {
  bool debug;         // debug mode: we can output debug information to standard error
  bool post;          // post mode: we can output search information to standard output
  int  player;        // engine player color
  int max_depth;      // engine max depth
  bool quit;          // quit status
};



// F U N C T I O N S   D E C L A R A T I O N

void Initialize();
void Destroy();


// U T I L I T Y   F U N C T I O N S

inline int PERC( uint64_t curr, uint64_t total )
{
  return( int )( curr * 100 / std::max( total, 1ULL ) );
}

// P O R T A B I L I T Y   D E F I N E S

#if defined(_WIN32) || defined(_WIN64)
#define U64F "%I64u"
#else
#define U64F "%llu"
#endif


#endif
