/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#include "thread.h"
#include "search.h"
#include "smp.h"

#include <iostream>

using namespace std;

#if defined(USESMP)

#if defined(_WIN32) || defined(_WIN64)

// we don't have condition in Windows so spinloop is mandatory
#if !defined(SPINLOOP)
#define SPINLOOP
#endif

#endif

// G L O B A L S

static volatile int start_thread = 0;
volatile bool thread_search_terminate = false;
volatile score_t thread_search_score = SCORE_NULL;
static volatile bool thread_pause = false;
static event_t thread_pause_event;

#if !defined(SPINLOOP)
static pthread_cond_t start_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t start_cond_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t end_cond = PTHREAD_COND_INITIALIZER;
#endif

// Atomic number variables:
// 1. working threads number
// 2. initialized threads number
// 3. timeout (it's not needed to be atomic)
static volatile AtomicNum<int> working_threads( 0 );
static volatile AtomicNum<int> init_threads( 0 );
static volatile int thread_timeout = 0;


// F U N C T I O N   D E C L A R A T I O N S

static THREADRET ThreadStart( void *arg );


// F U N C T I O N   D E F I N I T I O N S

/*
  Start not working threads
*/
void StartThreads()
{
#if defined(SPINLOOP)
  start_thread = 1;
#else
  pthread_mutex_lock( &start_cond_mutex );
  start_thread = 1;
  pthread_mutex_unlock( &start_cond_mutex );
  pthread_cond_broadcast( &start_cond );
#endif
}

/*
  Initialize a thread
 */
void InitThread( ThreadData &td )
{
  NewThread( td.tid, ThreadStart, (void*)&td );
}

/*
  Wait initialization for num threads
*/
void WaitThreadsInitialization( int num )
{
  while( init_threads != num );
}

/*
  Wait threads finish them work
*/
void WaitThreads()
{
#if defined( SPINLOOP )
  start_thread = 0;
#else
  pthread_mutex_lock( &start_cond_mutex );
  start_thread = 0;
  pthread_mutex_unlock( &start_cond_mutex );
  pthread_cond_broadcast( &end_cond );
#endif
  while( working_threads != 0 );
}

/*
  Wait threads destruction
*/
void DestroyInactiveThreads()
{
  WaitThreads();
#if !defined(SPINLOOP)
  pthread_cond_broadcast( &start_cond );
#endif
}

/*
  Reset thread timeout
  this should be called only by main thread
*/
void ResetThreadsTimeout()
{
  thread_timeout = 0;
}

/*
  Raise thread timeout
*/
void ThreadsTimeout()
{
  thread_timeout = 1;
}

/*
  Check if thread timeout is raised
*/
bool IsThreadTimeout()
{
  return thread_timeout == 1;
}

void InitPauseThreads()
{
  thread_pause = false;
  EventInit( thread_pause_event );
}

void PauseThreads()
{
  if( !thread_pause ) {
    thread_pause = true;
  }
}

void ThreadsNotifyAll()
{
  if( thread_pause ) {
    thread_pause = false;
    EventNotifyAll( thread_pause_event );
    WaitThreads();
    EventReset( thread_pause_event );
  }
}

/*
  Thread main function
  Helper thread wait for a job from main thread, then
  help him
*/
THREADRET ThreadStart( void *arg )
{
  // thread data structure must be volatile:
  //   it could be modified by main thread
  volatile ThreadData *tdv = (volatile ThreadData *)arg;

  ++init_threads;

  do {
#if defined(SPINLOOP)
    while( !start_thread && tdv->active && !thread_pause ) ;
    ++working_threads;
#else
    // wait until start signal
    
    pthread_mutex_lock( &start_cond_mutex );
    while( start_thread == 0 && tdv->active ) {
      pthread_cond_wait( &start_cond, &start_cond_mutex );
    }
    ++working_threads;
    pthread_mutex_unlock( &start_cond_mutex );
#endif

    // if program exit...
    if( tdv->active == false ) {
      --working_threads;
      break;
    }

    // if thread stand by..
    if( thread_pause ) {
      do {
	EventWait( thread_pause_event );
      } while( thread_pause );
      --working_threads;
      continue;
    }

    if( start_thread ) {
      ThreadData *td = (ThreadData *)tdv;
      // help main thread to search bestmove
      int init_treeply = td->board.treeply;
      try {
	score_t g = -Search( *td, td->root_alpha, td->root_beta, td->root_depth, 1 );
	TerminateThreadSearch( g );
      } catch( ThreadInterrupt &ti ) {
	// restore start position
	while( init_treeply < td->board.treeply ) {
	  td->board.undoMove();
	}
      }
      assert( init_treeply == td->board.treeply );
    }
    
    --working_threads;

#if defined(SPINLOOP)
    while( start_thread );
#else
    pthread_mutex_lock( &start_cond_mutex );
    while( start_thread == 1 ) {
      pthread_cond_wait( &end_cond, &start_cond_mutex );
    }
    pthread_mutex_unlock( &start_cond_mutex );
#endif
  } while( true );

  --init_threads;

  return 0;
}

/*
  Terminate thread search with specified score
*/
void TerminateThreadSearch( score_t score )
{
  thread_search_score = score;
  thread_search_terminate = true;
}

/*
  Reset thread termination signal
*/
void ResetThreadTermination()
{
  thread_search_score = SCORE_NULL;
  thread_search_terminate = false;
}

#endif
