/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#ifndef _THREAD_H_
#define _THREAD_H_

/*
  Threads data

  Created in 11/04/2006
  Maurizio
*/

#include "smash.h"
#include "stat.h"
#include "killer.h"
#include "history.h"
#include "board.h"
#include "eval.h"
#include "smp.h"
#include <exception>


struct ThreadCommunicationData
{
  ThreadCommunicationData( const Board &b ) : board( b )
  {
    for( int i = 0; i < MAXPLY; ++i ) {
      bound[i].alpha = SCORE_MAX;
      bound[i].beta = SCORE_MIN;
      LockInit( bound[i].lock );
    }
  }
  
  struct {
    AtomicNum<score_t> alpha;
    AtomicNum<score_t> beta;
    lock_t lock;
    int depth;
    bool pvnode;
  } bound[MAXPLY];
  
  const Board &board;
};

struct ThreadData
{
  inline ThreadData();

  // thread info
  int thread_id;
  thread_t tid;
  bool active;

  // root info
  score_t root_alpha;
  score_t root_beta;
  int root_depth;
  score_t root_score;  
  
  // stats
  SearchStatistics mystat;

  // search data
  Board board;
  KillerMoves killers;
  KillerMate kmate;
  History history;

  // communication
  VOLATILE ThreadCommunicationData comm;

  // move gen data
  move_t moves[MAX_MOVES];
  move_t *moves_set[MAX_MOVE_SETS];

  // move ordering data
  score_t sort_scores[MAX_MOVES];
  score_t *scores_set[MAX_MOVE_SETS];
};

// initialization
ThreadData::ThreadData() : comm( board )
{
  moves_set[0] = moves_set[1] = moves;
  scores_set[0] = scores_set[1] = sort_scores;
}

#if defined(USESMP)

void InitThread( ThreadData &td );
void WaitThreadsInitialization( int num );
void StartThreads();
void WaitThreads();
void DestroyInactiveThreads();
void ResetThreadsTimeout();
void ThreadsTimeout();
bool IsThreadTimeout();
void TerminateThreadSearch( score_t score );
void ResetThreadTermination();
void InitPauseThreads();
void PauseThreads();
void ThreadsNotifyAll();

extern volatile bool thread_search_terminate;
extern volatile score_t thread_search_score;

#else

#define InitThread( td )
#define WaitThreadsInitialization( num )
#define StartThreads()
#define WaitThreads()
#define DestroyInactiveThreads()
#define ResetThreadsTimeout()
#define ThreadsTimeout()
#define TerminateThreadSearch( score )
#define ResetThreadTermination()
#define InitPauseThreads();
#define PauseThreads();
#define ThreadsNotifyAll();

inline bool IsThreadTimeout()
{
  return false;
}

const bool thread_search_terminate = false;
const score_t thread_search_score = SCORE_NULL;

#endif

class ThreadInterrupt : public std::exception
{
public:
  ThreadInterrupt() throw() {}
  virtual ~ThreadInterrupt() throw() {}

  virtual const char *what() const throw() {
    return "thread interrupt exception";
  }
};

class ThreadSearchTerminate : public ThreadInterrupt
{
 public:
  ThreadSearchTerminate() throw() {}
  virtual ~ThreadSearchTerminate() throw() {}

  virtual const char *what() const throw() {
    return "thread search terminate exception";
  }
};


#endif
