/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#include <iostream>
#include <cstdio>
#include "ttable.h"
#include "board.h"
#include "search.h"
#include "smp.h"

using namespace std;


#if defined USESMP
#define CLUSTER_SIZE sizeof(Entry) * cluster_size + sizeof(lock_t)
#define LockCluster( index ) Locker locker( locks[(index)] )
#else
#define CLUSTER_SIZE sizeof(Entry) * cluster_size
#define LockCluster( index )
#endif


// S T A T I C   F U N C T I O N   D E C L A R A T I O N S

static int FirstPrimeAfter( int num );


// T T A B L E

TTable::TTable() : size( m_size ), capacity( m_capacity )
{
#if defined(USESMP)
  locks = 0;
#endif
  table = 0;
  resize( 0 );
}

TTable::~TTable()
{
  delete [] table;
}


/*
  Change TT capacity.
    capacity -- bytes to allocate (approximately) for the TT
*/
void TTable::resize( int capacity )
{
  delete [] table;
    
  if( capacity > 0 ) {
    capacity /= CLUSTER_SIZE;
    capacity = FirstPrimeAfter( capacity );
    table = new Entry[capacity * cluster_size];

#if defined(USESMP)
    locks = new lock_t[capacity];
    for( int i = 0; i < capacity; ++i ) {
      LockInit( locks[i] );
    }
#endif
    
  } else {
    table = 0;
  }
  m_capacity = capacity;
  clean();
}


/*
  Clean TT entries
*/
void TTable::clean()
{
  m_size = 0;
  for( int i = 0; i < capacity * cluster_size; i++ ) {
    table[i].scoretype = SCORE_EMPTY;
    table[i].nproc = 0;
  }

  m_timestamp = 0;

  // reset statistics
  Hits = Probes = 0;
}

/*
  Increment current timestamp: "ageing" all TT entries
*/
void TTable::ageing()
{
  ++m_timestamp;
}

/*
  Store one entry in the transposition table
*/
void TTable::store( const Board &board, Entry &newentry ) VOLATILE
{
  assert( newentry.score != SCORE_ONEVALUATION );
  assert( newentry.score != SCORE_NULL );
  
  if( capacity == 0 ) {
    return;
  }

  
  hash_t hashkey = board.hashkey;
  newentry.hashkey = hashkey;
  newentry.timestamp = m_timestamp;

  LockCluster( hashkey%capacity );
  Entry *entry = &table[(hashkey%capacity) * cluster_size];

  Entry *replace_entry = 0;
  int replace_score = INT_MIN;

  // look for the entry in the hash
  bool found = false;
  for( int i = 0; i < cluster_size; ++i, ++entry ) {
    if( entry->timestamp != m_timestamp ) {
      entry->nproc = 0;
      if( entry->scoretype != SCORE_EMPTY 
	  && entry->score == SCORE_ONEVALUATION ) {
	entry->scoretype = SCORE_EMPTY;
	--m_size;
	assert( m_size >= 0 );
      }
    }
    if( entry->scoretype != SCORE_EMPTY ) {
      if( entry->hashkey == hashkey ) {
	replace_entry = entry;
	found = true && m_timestamp == entry->timestamp;
	break;
      } else {
	int score = ( m_timestamp - entry->timestamp ) * 256 - entry->depth - entry->nproc*4;
	if( score > replace_score ) {
	  replace_score = score;
	  replace_entry = entry;
	}
      }
    } else if( replace_score < INT_MAX ) {
      assert( entry->nproc == 0 );
      replace_entry = entry;
      replace_score = INT_MAX;
    }

    assert( replace_entry != 0 );
  }

  assert( replace_entry != 0 );
  newentry.nproc = 0;

  if( found ) {
    if( replace_entry->depth > newentry.depth ) {
      assert( replace_entry->nproc < NumThreads );
      return;
    } else if( replace_entry->depth == newentry.depth ) {
      newentry.nproc = replace_entry->nproc - 1;
      if( newentry.nproc > replace_entry->nproc ) {
	if( flags.debug ) {
	  cout << "info string WARNING: nproc < 0 after store. Depth=" 
	       << (int)newentry.depth << " hashkey=" << hex << replace_entry->hashkey << dec << endl;
	}
	newentry.nproc = 0;
      }
      // if entry in ttable is better...
      if( replace_entry->score != SCORE_ONEVALUATION && replace_entry->scoretype == newentry.scoretype ) {
	if( newentry.scoretype == SCORE_FAILHIGH && newentry.score < replace_entry->score ) {
	  newentry.score = replace_entry->score;
	  newentry.bestmove = replace_entry->bestmove;
	} else if( newentry.scoretype == SCORE_FAILLOW && newentry.score > replace_entry->score ) {
	  newentry.score = replace_entry->score;
	}
      }
    } else {
      if( flags.debug ) {
	cout << "info string WARNING: replacing with deeper score an entry in store. Depth=" 
	     << (int)newentry.depth << " hashkey=" << hex << replace_entry->hashkey << dec << endl;
      }
      newentry.nproc = 0;
    }
  } else {
    if( flags.debug ) {
      cout << "info string WARNING: replace an entry in store. Depth=" << (int)newentry.depth 
	   << " nproc=" << (int)replace_entry->nproc
	   << " hashkey=" << hex << replace_entry->hashkey << dec << endl;
    }
    newentry.nproc = 0;
    if( replace_entry->scoretype == SCORE_EMPTY ) {
      ++m_size;
      assert( m_size <= capacity*cluster_size );
    }
  }
  assert( newentry.nproc < NumThreads );
  *replace_entry = newentry;
}

/*
  Search for a board in the transposition table.
  If the position is stored, return true and set tofind argument,
  else return false
*/
bool TTable::probe( const Board &board, Entry &tofind, bool exclusive ) VOLATILE
{
  if( capacity == 0 ) {
    return false;
  }

  ++Probes;
  
  hash_t hashkey = board.hashkey;
  LockCluster( hashkey%capacity );
  Entry *entry = &table[(hashkey%capacity) * cluster_size];

  bool found = false;

  Entry *replace_entry = 0;
  int replace_score = INT_MIN;

  // look for the entry, if we are in exclusive mode
  // we need also to choose a lock entry
  for( int i = 0; i < cluster_size; ++i, ++entry ) {
    if( entry->timestamp != m_timestamp ) {
      entry->nproc = 0;
      if( entry->scoretype != SCORE_EMPTY
	  && entry->score == SCORE_ONEVALUATION ) {
	entry->scoretype = SCORE_EMPTY;
	--m_size;
	assert( m_size >= 0 );
      }
    }
    if( entry->scoretype != SCORE_EMPTY && entry->hashkey == hashkey ) {
      found = true;
      break;
    } else if( entry->scoretype == SCORE_EMPTY ) {
      assert( entry->nproc == 0 );
      replace_score = INT_MAX;
      replace_entry = entry;
    } else if( replace_score < INT_MAX ) {
      int score = ( m_timestamp - entry->timestamp ) * 256 - entry->depth - entry->nproc*4;
      if( score > replace_score ) {
	replace_score = score;
	replace_entry = entry;
      }
    }

    assert( replace_entry != 0 );
  }

  if( found ) {
    ++Hits;
    entry->timestamp = m_timestamp;
    int depth = tofind.depth;
    tofind = *entry;

    if( depth == 0 ) {
      return tofind.score != SCORE_ONEVALUATION;
    }
    
    // if we are looking deeper than others thread that are searching this
    // position, we replace them
    if( depth > entry->depth ) {
      entry->depth = depth;
      entry->nproc = 1;
      entry->score = SCORE_ONEVALUATION;
      if( tofind.score == SCORE_ONEVALUATION ) {
	found = false;
      }
    } else if( depth == entry->depth ) {
      if( tofind.nproc > 0 && exclusive ) {
	tofind.score = SCORE_ONEVALUATION;
      } else {
	++entry->nproc;
	assert( entry->nproc > 0 && entry->nproc <= NumThreads );
	assert( !exclusive || entry->nproc == 1 );
	assert( entry->nproc > 1 || entry->score != SCORE_ONEVALUATION );
      }
    } else {
      assert( entry->nproc < NumThreads );
    }

    assert( entry->nproc <= NumThreads );

    return found && ( exclusive || tofind.score != SCORE_ONEVALUATION );
  } else {
    assert( replace_entry != 0 );
    
    if( tofind.depth == 0 ) {
      return false;
    }
    
    // exclusive mode: we have to lock the position in replace_entry
    if( replace_entry->scoretype == SCORE_EMPTY ) {
      ++m_size;
      assert( m_size <= capacity*cluster_size );
    } else if( flags.debug && replace_entry->nproc > 0 ) {
      cout << "info string WARNING: replace an entry in probe with nproc > 0. hashkey=" 
	   << hex << replace_entry->hashkey << dec 
	   << " nproc=" << (int)replace_entry->nproc << endl;
    }
    replace_entry->hashkey = hashkey;
    replace_entry->depth = tofind.depth;
    replace_entry->timestamp = m_timestamp;
    replace_entry->nproc = 1;
    replace_entry->score = SCORE_ONEVALUATION;
    replace_entry->scoretype = SCORE_EXACT;
    replace_entry->bestmove = NULLMOVE;
    
    return false;
  }
}

/*
  Unlock a position for ABDADA
*/
void TTable::leave( TTEntry &oldentry, int depth ) VOLATILE
{
  if( capacity == 0 || depth == 0 ) {
    return;
  }

  hash_t hashkey = oldentry.hashkey;
  LockCluster( hashkey%capacity );
  Entry *entry = &table[(hashkey%capacity) * cluster_size];

  for( int i = 0; i < cluster_size; ++i, ++entry ) {
    if( entry->scoretype != SCORE_EMPTY && entry->hashkey == hashkey ) {
      if( entry->depth == depth ) {
	if( entry->nproc > 0 ) {
	  --entry->nproc;
	} else {
	  if( flags.debug ) {
	    cout << "info string WARNING: nproc < 0 after leave. Depth=" << depth 
		 << " hashkey=" << hex << hashkey << dec << endl;
	  }
	}
	assert( entry->nproc < NumThreads );

	if( entry->score == SCORE_ONEVALUATION ) {
	  // restore oldentry if needed
	  if( oldentry.score != SCORE_ONEVALUATION ) {
	    *entry = oldentry;
	    entry->nproc = 0;
	    assert( entry->scoretype != SCORE_EMPTY );
	    // destroy if no other thread is working on it
	  } else if( entry->nproc == 0 ) {
	    entry->scoretype = SCORE_EMPTY;
	    --m_size;
	    assert( m_size >= 0 );
	  }
	}
	if( entry->nproc == 0 && entry->score == SCORE_ONEVALUATION ) {
	  entry->scoretype = SCORE_EMPTY;
	  --m_size;
	  assert( m_size >= 0 );
	}
      } else {
	assert( entry->nproc > 0 || entry->score != SCORE_ONEVALUATION );
	assert( entry->nproc < NumThreads );
      }
      break;
    }
  }
}

// S T A T I S T I C S

void TTable::dumpInfo()
{
  cout << "info hashfull " << PERC( size.get(), capacity*cluster_size ) << endl;
  if( flags.debug ) {
    printf( "info string TTHits: " U64F "(%d%%)\n", Hits.get(), PERC( Hits.get(), Probes.get() ) );
  }
}

// U T I L I T Y   F U N C T I O N S

/*
  return first prime >= num
*/
static int FirstPrimeAfter( int num )
{
  if( num <= 6 ) {
    switch( num ) {
    case 0:
    case 1:
    case 2: return 2;
    case 3: return 3;
    case 4:
    case 5: return 5;
    case 6: return 7;
    }
  }

  bool prime;
  int p = ( num % 2 ) ? num + 1 : num;
  do {
    prime = true;
    for( int i = 3; i*i <= p; i += 2 ) {
      if( p % i == 0 ) {
	prime = false;
	break;
      }
    }
    num = p;
    p += 2;
  } while( !prime );

  return num;
}

