/*
  smash, a chess engine
  Copyright (c) 2003-2006  Maurizio Sambati
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  See LICENSE file for more info.

  written by Maurizio Sambati <sambati@cli.di.unipi.it>
  http://www.cli.di.unipi.it/~sambati
*/
#ifndef _TTABLE_H_
#define _TTABLE_H_

/*
  Transposition table module

  Created in 15/10/2005
  Maurizio
*/

#include <stdint.h>
#include "smash.h"
#include "zobrist.h"
#include "move.h"
#include "eval.h"
#include "smp.h"


// C O N S T A N T S

const uint8_t SCORE_EMPTY = 0;
const uint8_t SCORE_FAILLOW = 1;
const uint8_t SCORE_FAILHIGH = 2;
const uint8_t SCORE_EXACT = 3;

// C L A S S E S


// TTable class

class TTable
{
 public:
  TTable();
  ~TTable();

  void resize( int capacity );
  void clean();
  void ageing();
  void dumpInfo();

  struct Entry
  {
    move_t bestmove;
    score_t score;
    uint8_t scoretype;
    uint8_t depth;
    uint8_t timestamp;
    hash_t hashkey;
    uint8_t nproc;   // ABDADA: number of threads that are using this entry
  };

  // Entry locker:
  //   lock and release a TT entry
  class LockEntry
  {
  public:
    LockEntry( VOLATILE TTable &tt ) : ttable( tt ), depth( 0 ) {}
    ~LockEntry() { if( depth > 0 ) release(); }
    
    inline bool acquire( const Board &b, Entry &entry, bool exclusive = false );
    inline void store( const Board &b, Entry &entry );
    inline void release();

  private:
    VOLATILE TTable &ttable;
    int depth;
    hash_t hashkey;
    Entry oldentry;
  };

  void store( const Board &board, Entry &entry ) VOLATILE;
  bool probe( const Board &board, Entry &entry, bool exclusive = false ) VOLATILE;
  void leave( Entry &oldentry, int depth ) VOLATILE;

  const AtomicNum<int> &size;
  const int &capacity;

 private:
  static const int cluster_size = 4;
  
  Entry *table;
#if defined(USESMP)
  lock_t *locks;
#endif
  AtomicNum<uint64_t> Hits;
  AtomicNum<uint64_t> Probes;
  AtomicNum<int> m_size;
  int m_capacity;
  uint8_t m_timestamp;
};

typedef TTable::Entry TTEntry;
typedef TTable::LockEntry TTLockEntry;


// I N L I N E   F U N C T I O N S

bool TTLockEntry::acquire( const Board &b, TTEntry &entry, bool exclusive )
{
  hashkey = b.hashkey;
  depth = entry.depth;
  bool ret = ttable.probe( b, entry, exclusive );
  if( ret && exclusive && entry.score == SCORE_ONEVALUATION ) {
    depth = 0;
  } else if( ret ) {
    oldentry = entry;
  } else {
    oldentry.hashkey = hashkey;
    oldentry.score = SCORE_ONEVALUATION;
  }
  return ret;
}

void TTLockEntry::store( const Board &b, TTEntry &entry )
{
  assert( hashkey == b.hashkey );
  assert( depth == entry.depth );
  ttable.store( b, entry );
  depth = 0;
}

void TTLockEntry::release()
{
  assert( depth > 0 );
  ttable.leave( oldentry, depth );
  depth = 0;
}

#endif
